/*
 *  Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "cache/apartment-proxy.h"
#include "ds/catch/catch.h"
#include "ds/conf-mock.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "hal/hardware-abstraction-mock.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device-mock.h"
#include "model/modelconst.h"
#include "model/modelmaintenance-mock.h"
#include "model/zone.h"
#include "propertysystem_common_paths.h"
#include "util.h"

static ds::LogChannel dsLogChannel("apartmentStatus", ds::LogSeverity::DEBUG);
static const char* TAGS = "[apartmentStatus]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid l1{"0000000000000000000000000000000001"};
static const Dsuid l2{"0000000000000000000000000000000002"};
static const Dsuid l3{"0000000000000000000000000000000003"};
static const Dsuid l4{"0000000000000000000000000000000004"};
static const Dsuid l5{"0000000000000000000000000000000005"};
static const Dsuid l6{"0000000000000000000000000000000006"};
static const Dsuid l7{"0000000000000000000000000000000007"};

static const Dsuid s1{"0000000000000000000000000000000008"};
static const Dsuid s2{"0000000000000000000000000000000009"};
static const Dsuid s3{"000000000000000000000000000000000a"};
static const Dsuid s4{"000000000000000000000000000000000b"};
static const Dsuid s5{"000000000000000000000000000000000c"};
static const Dsuid s6{"000000000000000000000000000000000d"};
static const Dsuid s7{"000000000000000000000000000000000e"};
static const Dsuid s8{"000000000000000000000000000000000f"};

static const Dsuid a1{"0000000000000000000000000000000011"};
static const Dsuid a2{"0000000000000000000000000000000012"};
static const Dsuid a3{"0000000000000000000000000000000013"};
static const Dsuid a4{"0000000000000000000000000000000014"};
static const Dsuid a5{"0000000000000000000000000000000015"};
static const Dsuid a6{"0000000000000000000000000000000016"};
static const Dsuid a7{"0000000000000000000000000000000017"};

void createDevice(Apartment& apt, const Dsuid dsid, ApplicationType group = ApplicationType::lights) {
  DeviceSpec spec;
  if (group == ApplicationType::lights) {
    spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100::output, FID100keys::_1);
    spec.ProductID = (1 << 10) | 200;
  } else if (group == ApplicationType::blinds){
    spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_GR, FID100keys::_1);
    spec.ProductID = (3 << 10) | 200;
  } else if (group == ApplicationType::audio){
    spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_SW, FID100::output, FID100keys::_1);
    spec.ProductID = 0x0CC8;
  }
  std::shared_ptr<Device> dev = apt.allocateDevice(dsid);
  dev->setPartiallyFromSpec(spec);
  dev->setIsPresent(true);
  dev->setIsValid(true);
  dev->setIsConnected(true);
  if (group == ApplicationType::lights) {
    dev->addToGroup(GroupIDYellow);
  } else if (group == ApplicationType::blinds){
    dev->addToGroup(GroupIDGray);
  } else if (group == ApplicationType::audio){
    dev->addToGroup(GroupIDCyan);
  }
}

void addDeviceToZone(Apartment& apt, Zone& zone, const Dsuid dsid) {
  DeviceReference lightref(dsid, &apt);
  zone.addDevice(lightref);
}

SCENARIO("Apartment status light calculation", TAGS) {
  PropertySystem propSys;
  auto pApartment = makeApartment(nullptr, &propSys);
  auto& apt = *pApartment;
  ds::ConfMock conf;

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  auto& zone1 = *apt.allocateZone(1);
  auto& zone2 = *apt.allocateZone(2);
  apt.getZone(0)->addGroup(std::make_shared<Cluster>(16, apt));
  apt.getCluster(16)->setApplicationType(ApplicationType::lights);
  apt.getZone(0)->addGroup(std::make_shared<Cluster>(17, apt));
  apt.getCluster(17)->setApplicationType(ApplicationType::lights);

  createDevice(apt, l1);
  createDevice(apt, l2);
  createDevice(apt, l3);
  createDevice(apt, l4);
  createDevice(apt, l5);
  createDevice(apt, l6);
  createDevice(apt, l7);

  addDeviceToZone(apt, zone1, l1);
  addDeviceToZone(apt, zone1, l2);
  addDeviceToZone(apt, zone1, l3);
  addDeviceToZone(apt, zone1, l4);
  addDeviceToZone(apt, zone2, l5);
  addDeviceToZone(apt, zone2, l6);
  addDeviceToZone(apt, zone2, l7);

  apt.getDeviceByDSID(l1)->addToGroup(16);
  apt.getDeviceByDSID(l5)->addToGroup(16);
  apt.getDeviceByDSID(l5)->addToGroup(17);
  apt.getDeviceByDSID(l6)->addToGroup(17);
  apt.getDeviceByDSID(l7)->addToGroup(17);

  ds::sqlite3::Database db(":memory:");
  auto hal = HardwareAbstractionMock{};

  auto deviceCache = CacheDb<DeviceCache>(&db);
  REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  ALLOW_CALL(hal, areaMember(l1)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(l2)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(l3)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(l4)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(l5)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(l6)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(l7)).RETURN(std::vector<int>{});

  WHEN("l1 is on") {
    THEN("zone 1, area 1 and cluster 1 must be on") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}}};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 30.0, .targetValue = 30.0, .level = 0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds2);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds2);

      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG();

      REQUIRE(lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
    THEN("but locally priorized") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}},
                           .localPriority = true};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 30.0, .targetValue = 30.0, .level = 0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds2);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds2);


      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG();

      REQUIRE(lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::on);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::on);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
  }

  WHEN("l1 is dimming") {
    THEN("zone 1, area 1 and cluster 1 must be dimming") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::dimming,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}}};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds0);

      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG(lights);

      REQUIRE(lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
  }

  WHEN("l1 is moving") {
    THEN("zone 1, area 1 and cluster 1 must be dimming") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::brightness, {.value = 100.0, .targetValue = 100.0}}}};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::brightness, {.value = 0.0, .targetValue = 0.0}}}};
      REQUIRE_CALL(hal, getState(l1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(l2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l5)).TIMES(3).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l6)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(l7)).TIMES(2).RETURN(ds0);

      dsModel::LightsStatus lights;
      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone1.id()));

      DS_DEBUG(lights);

      REQUIRE(lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::dimming);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      REQUIRE_NOTHROW(lights = proxy->lightsStatus(zone2.id()));
      REQUIRE(lights.status == dsModel::LightsStatusStatus::off);
      REQUIRE(lights.area1 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area2 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area3 == dsModel::LightsStatusStatus::none);
      REQUIRE(lights.area4 == dsModel::LightsStatusStatus::none);

      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "16");
      REQUIRE(g[0].application.lights.status == dsModel::LightsStatusStatus::dimming);
      REQUIRE(g[1].id == "17");
      REQUIRE(g[1].application.lights.status == dsModel::LightsStatusStatus::off);
    }
  }
}

SCENARIO("Apartment status shade calculation", TAGS) {
  PropertySystem propSys;
  auto pApartment = makeApartment(nullptr, &propSys);
  auto& apt = *pApartment;
  ds::ConfMock conf;

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  auto& zone1 = *apt.allocateZone(1);
  auto& zone2 = *apt.allocateZone(2);
  auto& zone3 = *apt.allocateZone(3);

  apt.getZone(0)->addGroup(std::make_shared<Cluster>(18, apt));
  apt.getCluster(18)->setApplicationType(ApplicationType::blinds);
  apt.getZone(0)->addGroup(std::make_shared<Cluster>(19, apt));
  apt.getCluster(19)->setApplicationType(ApplicationType::blinds);

  createDevice(apt, s1, ApplicationType::blinds);
  createDevice(apt, s2, ApplicationType::blinds);
  createDevice(apt, s3, ApplicationType::blinds);
  createDevice(apt, s4, ApplicationType::blinds);
  createDevice(apt, s5, ApplicationType::blinds);
  createDevice(apt, s6, ApplicationType::blinds);
  createDevice(apt, s7, ApplicationType::blinds);
  createDevice(apt, s8, ApplicationType::blinds);

  addDeviceToZone(apt, zone1, s1);
  addDeviceToZone(apt, zone1, s2);
  addDeviceToZone(apt, zone1, s3);
  addDeviceToZone(apt, zone1, s4);
  addDeviceToZone(apt, zone2, s5);
  addDeviceToZone(apt, zone2, s6);
  addDeviceToZone(apt, zone2, s7);
  addDeviceToZone(apt, zone3, s8);

  apt.getDeviceByDSID(s1)->addToGroup(18);
  apt.getDeviceByDSID(s5)->addToGroup(18);
  apt.getDeviceByDSID(s5)->addToGroup(19);
  apt.getDeviceByDSID(s6)->addToGroup(19);
  apt.getDeviceByDSID(s7)->addToGroup(19);

  ds::sqlite3::Database db(":memory:");
  HardwareAbstractionMock hal{};

  auto deviceCache = CacheDb<DeviceCache>(&db);
  REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  ALLOW_CALL(hal, areaMember(s1)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(s2)).RETURN(std::vector<int>{6});
  ALLOW_CALL(hal, areaMember(s3)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(s4)).RETURN(std::vector<int>{7});
  ALLOW_CALL(hal, areaMember(s5)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(s6)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(s7)).RETURN(std::vector<int>{});
  ALLOW_CALL(hal, areaMember(s8)).RETURN(std::vector<int>{});

  WHEN("shade status is checked") {
    THEN("it must be correct") {
      ds::DateTime timeNow;
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 100.0, .targetValue = 100.0}}}};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 0.0, .targetValue = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = timeNow,
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 50.0, .targetValue = 60.0}}}};

      REQUIRE_CALL(hal, getState(s1)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s5)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s6)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s7)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s8)).TIMES(2).RETURN(ds2);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDGray).get(), 1, 1, 0, 255, 255));
#endif
      dsModel::ShadesStatus shades;
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone1.id()));

      DS_DEBUG(shades);

      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::closed);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
#endif
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone2.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone3.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
#endif
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(dsModel::ZoneId{"3"}));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(18).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(19).get(), 1));
#endif
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "18");
      REQUIRE(g[0].application.shades.status == dsModel::ShadesStatusStatus::open);
      REQUIRE(g[1].id == "19");
      REQUIRE(g[1].application.shades.status == dsModel::ShadesStatusStatus::mixed);

      dsModel::DSDeviceStatus status;
      REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
      REQUIRE_NOTHROW(status = proxy->dsDeviceStatus(dsModel::DSDeviceId{ds::str(s8)}));
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].initialValue == 50);
      REQUIRE(!status.attributes.functionBlocks[0].outputs[0].value);
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].startedAt == timeNow);

      REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
      REQUIRE_NOTHROW(status = proxy->dsDeviceStatus(dsModel::DSDeviceId{ds::str(s1)}));
      REQUIRE(!status.attributes.functionBlocks[0].outputs[0].initialValue);
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].value == 100);
      REQUIRE(status.attributes.functionBlocks[0].outputs[0].startedAt == ds::DateTime());
    }
    THEN("it must be correct also with local prio") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 100.0, .targetValue = 100.0}}},
                           .localPriority = true};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 0.0, .targetValue = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 50.0, .targetValue = 60.0}}}};

      REQUIRE_CALL(hal, getState(s1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s5)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s6)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s7)).TIMES(2).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s8)).TIMES(1).RETURN(ds2);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDGray).get(), 0, 1, 0, 255, 255));
#endif
      dsModel::ShadesStatus shades;
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone1.id()));

      DS_DEBUG(shades);

      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::closed);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDGray).get(), 0, 255, 255, 255, 255));
#endif
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone2.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::mixed);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone3.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
#endif
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(dsModel::ZoneId{"3"}));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(19).get(), 0));
#endif
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "18");
      REQUIRE(g[0].application.shades.status == dsModel::ShadesStatusStatus::open);
      REQUIRE(g[1].id == "19");
      REQUIRE(g[1].application.shades.status == dsModel::ShadesStatusStatus::mixed);
    }
  }

  WHEN("shade status with dimming") {
    THEN("it must be correct") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::dimming,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 100.0, .targetValue = 100.0}}}};
      hal::DeviceState ds0{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 0.0, .targetValue = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::moving,
                           .outputs = {{ChannelType::shadePositionOutside, {.value = 50.0, .targetValue = 60.0}}}};

      REQUIRE_CALL(hal, getState(s1)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s2)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s3)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s4)).TIMES(1).RETURN(ds0);
      REQUIRE_CALL(hal, getState(s5)).TIMES(3).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s6)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s7)).TIMES(2).RETURN(ds1);
      REQUIRE_CALL(hal, getState(s8)).TIMES(1).RETURN(ds2);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDGray).get(), 1, 1, 0, 255, 255));
#endif

      dsModel::ShadesStatus shades;
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone1.id()));

      DS_DEBUG(shades);

      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::closed);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone2.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
#endif
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone2.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone3.getGroup(GroupIDGray).get(), 1, 255, 255, 255, 255));
#endif
      REQUIRE_NOTHROW(shades = proxy->shadesStatus(zone3.id()));
      REQUIRE(shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(shades.area1 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area2 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area3 == dsModel::ShadesStatusStatus::none);
      REQUIRE(shades.area4 == dsModel::ShadesStatusStatus::none);

#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(18).get(), 1));
      REQUIRE_CALL(busInterface.action, setState(apt.getCluster(19).get(), 1));
#endif
      std::vector<dsModel::GroupStatus> g;
      REQUIRE_NOTHROW(g = proxy->groupsStatus());
      REQUIRE(g[0].id == "18");
      REQUIRE(g[0].application.shades.status == dsModel::ShadesStatusStatus::moving);
      REQUIRE(g[1].id == "19");
      REQUIRE(g[1].application.shades.status == dsModel::ShadesStatusStatus::moving);
    }
  }
}

SCENARIO("Apartment status audio calculation", TAGS) {
  PropertySystem propSys;
  auto pApartment = makeApartment(nullptr, &propSys);
  auto& apt = *pApartment;
  ds::ConfMock conf;

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  auto& zone1 = *apt.allocateZone(1);
  auto& zone2 = *apt.allocateZone(2);

  apt.getZone(0)->addGroup(std::make_shared<Cluster>(20, apt));
  apt.getCluster(18)->setApplicationType(ApplicationType::audio);
  apt.getZone(0)->addGroup(std::make_shared<Cluster>(21, apt));
  apt.getCluster(19)->setApplicationType(ApplicationType::audio);

  createDevice(apt, a1, ApplicationType::audio);
  createDevice(apt, a2, ApplicationType::audio);
  createDevice(apt, a3, ApplicationType::audio);
  createDevice(apt, a4, ApplicationType::audio);
  createDevice(apt, a5, ApplicationType::audio);
  createDevice(apt, a6, ApplicationType::audio);
  createDevice(apt, a7, ApplicationType::audio);

  addDeviceToZone(apt, zone1, a1);
  addDeviceToZone(apt, zone1, a2);
  addDeviceToZone(apt, zone1, a3);
  addDeviceToZone(apt, zone1, a4);
  addDeviceToZone(apt, zone2, a5);
  addDeviceToZone(apt, zone2, a6);
  addDeviceToZone(apt, zone2, a7);

  apt.getDeviceByDSID(a1)->addToGroup(20);
  apt.getDeviceByDSID(a5)->addToGroup(20);
  apt.getDeviceByDSID(a5)->addToGroup(21);
  apt.getDeviceByDSID(a6)->addToGroup(21);
  apt.getDeviceByDSID(a7)->addToGroup(21);

  ds::sqlite3::Database db(":memory:");
  HardwareAbstractionMock hal{};

  auto deviceCache = CacheDb<DeviceCache>(&db);
  REQUIRE_CALL(hal, getCache()).LR_RETURN(deviceCache);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  WHEN("Checking zone status with audio device in standby without sonos device") {
    THEN("it must be correct") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::powerState, {.value = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::standby,
                           .outputs = {{ChannelType::powerState, {.value = 100.0}}}};

      dsModel::ZoneApplicationsStatus zs1{};

      REQUIRE_CALL(hal, getState(a1)).TIMES(1).RETURN(ds2);
      REQUIRE_CALL(hal, getState(a2)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a3)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a4)).TIMES(1).RETURN(ds1);
#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDCyan).get(), 0, 255,255,255,255));
#endif

      REQUIRE_CALL(hal, getZoneState(1)).RETURN(zs1);
      dsModel::AudioStatus audio;
      REQUIRE_NOTHROW(audio = proxy->audioStatus(zone1.id()));

      REQUIRE(audio.status == dsModel::AudioStatusStatus::standby);
    }
  }

  WHEN("Checking zone status with audio device in on state without sonos device") {
    THEN("it must be correct") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::powerState, {.value = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::powerState, {.value = 100.0, .level = 2.0}}}};

      dsModel::ZoneApplicationsStatus zs1{};

      REQUIRE_CALL(hal, getState(a1)).TIMES(1).RETURN(ds2);
      REQUIRE_CALL(hal, getState(a2)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a3)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a4)).TIMES(1).RETURN(ds1);
#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDCyan).get(), 1, 255, 255, 255, 255));
#endif

      REQUIRE_CALL(hal, getZoneState(1)).RETURN(zs1);
      dsModel::AudioStatus audio;
      REQUIRE_NOTHROW(audio = proxy->audioStatus(zone1.id()));

      REQUIRE(audio.status == dsModel::AudioStatusStatus::playing);
    }
  }

  WHEN("Checking zone status with audio with sonos device") {
    THEN("It must be correct") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::powerState, {.value = 0.0}}}};
      hal::DeviceState ds2{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::standby,
                           .outputs = {{ChannelType::powerState, {.value = 100.0, .level = 2.0}}}};

      dsModel::ZoneApplicationsStatus zs1{.audio = {
                                            .status = dsModel::AudioStatusStatus::playing,
                                            .muted = true,
                                            .volume = 30.0
                                          }};

      REQUIRE_CALL(hal, getState(a1)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a2)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a3)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(a4)).TIMES(1).RETURN(ds2);
#ifdef EMPTYCLICK_NON_LIGHTS
      REQUIRE_CALL(busInterface.action, setState(zone1.getGroup(GroupIDCyan).get(), 1, 255, 255, 255, 255));
#endif

      REQUIRE_CALL(hal, getZoneState(1)).RETURN(zs1);
      dsModel::AudioStatus audio;
      REQUIRE_NOTHROW(audio = proxy->audioStatus(zone1.id()));

      REQUIRE(audio.status == dsModel::AudioStatusStatus::playing);
      REQUIRE(audio.muted == true);
      REQUIRE(audio.volume == 30.0);
    }
  }
}

} // namespace
