/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "http-server.h"

#include <ds-api/rest-mock.h>
#include <ds/asio/promise.h>
#include <ds/asio/test.h>
#include <ds/clock-mock.h>
#include <ds/log.h>

#include <boost/asio/connect.hpp>
#include <boost/beast/core.hpp>
#include <boost/beast/http.hpp>
#include <boost/beast/version.hpp>

using namespace std::literals;
using namespace trompeloeil;
using boost::system::error_code;
using tcp = boost::asio::ip::tcp;    // from <boost/asio/ip/tcp.hpp>
namespace http = boost::beast::http; // from <boost/beast/http.hpp>

namespace {

static const char* TAGS = "[dsApiHttpServer][dsApi]";

SCENARIO("dsApiHttpServer", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  auto& io = loop.io;

  dsApi::AsyncCalleeMock asyncCalle;
  auto server = dsApi::makeHttpServer(loop, {.acceptor = {.endpoint = {.address = "127.0.0.1"}}}, io, asyncCalle);
  WHEN("client connects") {
    tcp::resolver resolver{io};
    tcp::socket socket{io};
    boost::asio::connect(socket, resolver.resolve(server->acceptor().localAsioEndpoint()));

    THEN("request is delivered as promise") {
      struct X {
        http::verb beastVerb;
        dsApi::HttpMethod method;
      };
      const auto& x = GENERATE(
          values(std::initializer_list<X>{{.beastVerb = http::verb::get, .method = dsApi::HttpMethod::GET},
                                          {.beastVerb = http::verb::put, .method = dsApi::HttpMethod::PUT},
                                          {.beastVerb = http::verb::post, .method = dsApi::HttpMethod::POST},
                                          {.beastVerb = http::verb::delete_, .method = dsApi::HttpMethod::DELETE},
                                          {.beastVerb = http::verb::patch, .method = dsApi::HttpMethod::PATCH}}));

      auto promise = std::unique_ptr<ds::Promise<dsApi::HttpResult>>{};
      {
        auto request = http::request<http::string_body>{};
        request.version(11);
        request.method(x.beastVerb);
        request.target("/path?key=value");
        request.body() = R"("requestContent")";
        request.set("x-header1", "a");
        request.insert("x-header2", "b");
        request.insert("x-header2", "c");
        request.prepare_payload();
        http::async_write(socket, request, [&](error_code ec, std::size_t) { loop.stop(); });
        REQUIRE_CALL(asyncCalle,
                     asyncInvoke(dsApi::HttpRequest{.path = "/path",
                                                    .httpArgs = {.method = x.method,
                                                                 .content = ds::toJsonFragment("requestContent"),
                                                                 .query = "key=value",
                                                                 .headers =
                                                                     {
                                                                         {"Content-Type", {"application/json"}},
                                                                         {"x-header1", {"a"}},
                                                                         {"x-header2", {"b", "c"}},
                                                                     }}},
                                 _))
            .LR_SIDE_EFFECT(promise = std::move(_2); loop.stop());
        loop.run();
        loop.run();
      }

      THEN("disconnected") {
        socket.close();
        SUCCEED();
      }

      THEN("response is delivered as http response") {
        promise->value = dsApi::okDocumentResult("responseContent");
        promise->value->headers["x-header1"].push_back("aa");
        promise->value->headers["x-header2"].push_back("bb");
        promise->value->headers["x-header2"].push_back("cc");
        promise.reset();

        auto buffer = boost::beast::flat_buffer{};
        auto response = http::response<http::string_body>{};
        http::async_read(socket, buffer, response, [&](error_code ec, std::size_t) {
          CHECK(response.version() == 11);
          CHECK(response.result() == http::status{200});
          CHECK(response[http::field::content_type] == "application/json");
          CHECK(response["x-header1"] == "aa");
          {
            auto range = response.equal_range("x-header2");
            CHECK(std::next(range.first, 0)->value() == "bb");
            CHECK(std::next(range.first, 1)->value() == "cc");
            CHECK(std::next(range.first, 2) == range.second);
          }
          CHECK(response.body() == R"({"data":"responseContent"})");
          loop.stop();
        });
        loop.run();
        SUCCEED();
      }
    }

    THEN("Unknown verb returns error") {
      auto request = http::request<http::string_body>{};
      request.version(11);
      request.method_string("unknown");
      request.target("/");
      request.prepare_payload();
      http::async_write(socket, request, [&](error_code ec, std::size_t) { loop.stop(); });
      loop.run();

      auto buffer = boost::beast::flat_buffer{};
      auto response = http::response<http::string_body>{};
      http::async_read(socket, buffer, response, [&](error_code ec, std::size_t) {
        CHECK(response.version() == 11);
        CHECK(response.result() == http::status{500});
        CHECK(response[http::field::content_type] == "application/json");
        CHECK(response.body().find("Unknown ::dsApi::HttpMethod enum option") != std::string::npos);
        loop.stop();
      });
      loop.run();
      SUCCEED();
    }
  }
}

} // namespace
