#include "notifications.h"

#include <ds-net/buffered-message-writer.h>
#include <ds-net/session-base.h>
#include <ds-net/ssl.h>
#include <ds-net/websocket.h>
#include <ds/clock.h>
#include <ds/json.h>

#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/intrusive/list.hpp>

static ds::LogChannel dsLogChannel("dsApiNotifications");

namespace intrusive = boost::intrusive;
using boost::system::error_code;
namespace asio = boost::asio;
using namespace std::literals;

namespace dsApi {

namespace notifications {
struct SubscribeArgs {
  std::string type;
  ds::Clock::duration timeout;
};
} // namespace notifications

using namespace notifications;

namespace {

struct NotificationsSessionImpl final : dsNet::Session<NotificationsSessionImpl>,
                                        dsNet::SessionsHook,
                                        NotificationsSession,
                                        dsNet::SessionStateBase {
  using This = NotificationsSessionImpl;
  dsNet::ClosePtr<dsNet::SignalrSession> m_signalr;
  dsApi::EventHub& m_eventHub;
  std::unique_ptr<dsApi::EventSink> m_eventSink{};
  std::shared_ptr<void> m_owningPtr; // holds session this alive until close()

  NotificationsSessionImpl(dsNet::Sessions& sessions, dsNet::ClosePtr<dsNet::SignalrSession> signalr,
                           dsApi::EventHub& eventHub)
      : dsNet::Session<NotificationsSessionImpl>{sessions, std::string{signalr->id()}},
        dsNet::SessionsHook{sessions},
        m_signalr(std::move(signalr)),
        m_eventHub{eventHub} {}

  void asyncRun(NotificationsSessionHooks&& hooks) override {
    setOpening();
    m_signalr->asyncRun({.opened = weakCallback(&This::onSignalrOpened),
                         .closed = weakCallback(&This::onSignalrClosed),
                         .invokeNoReply = weakCallback(&This::onSignalrInvokeNoReply),
                         .bufferedMessageWriterConfig = std::move(hooks.bufferedMessageWriterConfig)});
    m_owningPtr = shared_from_this();
  }

  void close() override {
    DS_DEBUG_ENTER(*this, state());
    assertSynchronized();
    if (isClosed()) {
      return;
    }
    setClosed();
    m_owningPtr.reset();
    m_signalr.reset();
    m_eventSink.reset();
  }

  void onSignalrOpened() {
    DS_INFO_ENTER(*this);
    try {
      setOpen();
      m_eventSink =
          dsApi::makeEventSink(*this, m_eventHub, {.event = weakCallback(&NotificationsSessionImpl::onEvent)});
      m_eventSink->subscribe(
          {.type = dsApi::EventType{"apartmentStructureChanged"}, .expiresAt = ds::Clock::time_point::max()});
      m_eventSink->subscribe(
          {.type = dsApi::EventType{"apartmentStatusChanged"}, .expiresAt = ds::Clock::time_point::max()});
      DS_NOTICE("Notifications session started", *this);
    } catch (const std::exception& e) {
      DS_ERROR("Opening sessios failed.", e, *this);
      close();
    }
  }

  void onSignalrClosed(error_code ec) {
    DS_INFO_ENTER(*this, ec);
    close();
  }

  void onSignalrInvokeNoReply(std::string&& target, std::vector<ds::JsonFragment>&& arguments) {
    DS_INFO_ENTER(target, arguments);
    requireOpen();
    if (target == "subscribe") {
      DS_REQUIRE(arguments.size() > 0, "Subscribe needs arguments");
      subscribe(ds::fromJson<SubscribeArgs>(arguments.at(0)));
    }
  }

  void subscribe(SubscribeArgs&& args) {
    DS_INFO_ENTER(args);
    requireOpen();
    m_eventSink->subscribe(
        {.type = dsApi::EventType{std::move(args.type)}, .expiresAt = ds::Clock::now() + args.timeout});
  }

  void onEvent(ds::Box<dsApi::Event>&& event) {
    DS_INFO_ENTER(event);
    if (isClosed()) {
      return;
    }
    try {
      requireOpen();
      m_signalr->invokeNoReply("event", {ds::toJsonFragment(*event)});
    } catch (std::exception& e) {
      DS_ERROR("Failed to write event", e);
      close();
    }
  }
};

struct NotificationsImpl final : ds::asio::Resource<NotificationsConfig>, Notifications {
  std::unique_ptr<dsNet::SignalrAcceptor> m_acceptor;
  dsApi::EventHub& m_eventHub;
  dsNet::Sessions m_sessions{*this};

  NotificationsImpl(ds::asio::ResourceBase& parent, std::unique_ptr<dsNet::SignalrAcceptor> acceptor,
                    NotificationsConfig&& config, dsApi::EventHub& eventHub)
      : ds::asio::Resource<NotificationsConfig>{parent, "impl", std::move(config)},
        m_acceptor{std::move(acceptor)},
        m_eventHub{eventHub} {
    asyncRun();
  }

  void asyncRun() {
    m_acceptor->asyncRun({.accept = [this](dsNet::ClosePtr<dsNet::SignalrSession>&& signalr) {
      DS_INFO("accept", *this, *signalr);
      std::make_shared<NotificationsSessionImpl>(m_sessions, std::move(signalr), m_eventHub)
          ->asyncRun({.bufferedMessageWriterConfig = config().sessionWriteBuffer});
    }});
  }
};

} // namespace

dsNet::ClosePtr<NotificationsSession> makeNotificationsSession(dsNet::Sessions& sessions,
                                                               dsNet::ClosePtr<dsNet::SignalrSession>&& signalr,
                                                               dsApi::EventHub& eventHub) {
  return std::make_shared<NotificationsSessionImpl>(sessions, std::move(signalr), eventHub);
}

std::unique_ptr<Notifications> makeNotifications(ds::asio::ResourceBase& parent,
                                                 std::unique_ptr<dsNet::SignalrAcceptor> acceptor,
                                                 dsApi::EventHub& eventHub, NotificationsConfig&& config) {
  return std::make_unique<NotificationsImpl>(parent, std::move(acceptor), std::move(config), eventHub);
}

} // namespace dsApi

DS_JSON_DECLARE_STRUCT((dsApi)(notifications), SubscribeArgs, (type)(timeout));

DS_JSON_DEFINE_STRUCT((dsApi)(notifications), SubscribeArgs);
DS_JSON_DEFINE_STRUCT((dsApi), NotificationsConfig);
