/*
    Copyright (c) 2016 digitalSTROM AG, Zurich, Switzerland

    This file is part of vdSM.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vdSM. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <avahi-client/lookup.h>
#include <ds/asio/timer.h>
#include <ds/collection.h>
#include <ds/common.h>

#include <boost/noncopyable.hpp>
#include <memory>
#include <string>
#include <vector>

#include "client.h"

namespace dsAvahi {

/// Information available about an unresolved service
struct BrowsedService {
  AvahiIfIndex interface{};
  AvahiProtocol protocol{};
  std::string name{};
  std::string type{};
  std::string domain{};
};

/// Browses service type and
/// provides BrowsedService collection notifications
class ServicesBrowser : private Client::Hook {
public:
  typedef ::ds::collection::Event<BrowsedService>::Type Event;
  typedef std::function<void(Event)> OnEvent;
  ServicesBrowser(Client &client, std::string type, OnEvent onEvent);
  ~ServicesBrowser();

  Client &client() { return m_client; }
  const std::string &type() { return m_type; }
  AvahiLookupFlags flags() const { return m_flags; }

private:
  Client &m_client;
  struct Deleter {
    void operator()(AvahiServiceBrowser *x) { avahi_service_browser_free(x); }
  };
  std::unique_ptr<AvahiServiceBrowser, Deleter> m_ptr;
  std::string m_type;
  AvahiLookupFlags m_flags;
  OnEvent m_onEvent;
  bool m_isPendingReset; // could be removed if we could ask timer whether it is active.
  ds::asio::Timer m_pendingResetTimer;

  void emitPendingReset();
  void onRunning() override;
  void freed() override;
  std::string createServiceId(AvahiIfIndex interface, AvahiProtocol protocol, const char *name, const char *type,
                              const char *domain);
  static void sCallback(AvahiServiceBrowser *, AvahiIfIndex interface, AvahiProtocol protocol, AvahiBrowserEvent event,
                        const char *name, const char *type, const char *domain, AvahiLookupResultFlags flags,
                        void *data);
  void callback(AvahiIfIndex interface, AvahiProtocol protocol, AvahiBrowserEvent event, const char *name,
                const char *type, const char *domain, AvahiLookupResultFlags flags);

  const std::string &typePrefix() const { return m_client.typePrefix(); }
};

/// Information available about a resolved service
struct ResolvedService {
  AvahiIfIndex interface{};
  AvahiProtocol protocol{};
  std::string name{};
  std::string type{};
  std::string domain{};
  std::string address{};
  uint16_t port{};
  std::vector<std::string> txts{};
  AvahiLookupResultFlags flags{};
};

/// Resolves one service and calls callback when resolved.
/// Resolved service information may change over time,
/// then the callback will be called again.
class ServiceResolver : private Client::Hook {
public:
  // Either error or a value.
  typedef std::variant<std::string, ResolvedService> Reply;
  typedef std::function<void(Reply)> Replied;

  /// Requires client to be in running state.
  /// The browsed service is invalidated in !running state anyway.
  ServiceResolver(Client &client, const BrowsedService &service, AvahiLookupFlags flags, Replied replied);
  ~ServiceResolver();

  AvahiLookupFlags flags() const { return m_flags; }

private:
  Client &m_client;
  AvahiLookupFlags m_flags;
  Replied m_replied;
  struct Deleter {
    void operator()(AvahiServiceResolver *x) { avahi_service_resolver_free(x); }
  };
  std::unique_ptr<AvahiServiceResolver, Deleter> m_ptr;

  void freed() override;
  static void sCallback(AvahiServiceResolver *, AvahiIfIndex interface, AvahiProtocol protocol,
                        AvahiResolverEvent event, const char *name, const char *type, const char *domain,
                        const char *hostName, const AvahiAddress *address, uint16_t port, AvahiStringList *txt,
                        AvahiLookupResultFlags flags, void *data);
  void callback(AvahiIfIndex interface, AvahiProtocol protocol, AvahiResolverEvent event, const char *name,
                const char *type, const char *domain, const char *hostName, const AvahiAddress *address, uint16_t port,
                AvahiStringList *txtList, AvahiLookupResultFlags flags);

  const std::string &typePrefix() const { return m_client.typePrefix(); }
};

/// Browses and resolves service type and
/// provides ResolvedService collection notifications.
/// Hides unresolved services.
///
/// Internally maintains ServicesBrowser
/// and ServiceResolver for each browsed service.
class ServicesResolver {
public:
  typedef ::ds::collection::Event<ResolvedService>::Type Event;
  typedef std::function<void(Event)> OnEvent;

  ServicesResolver(Client &client, std::string type, OnEvent onEvent);
  ~ServicesResolver();

  Client &client() { return m_browser.client(); }
  const std::string &type() { return m_browser.type(); }
  AvahiLookupFlags flags() { return m_browser.flags(); }

private:
  ServicesBrowser m_browser;
  OnEvent m_onEvent;
  struct Entry {
    std::unique_ptr<ServiceResolver> resolver;
    bool resolved = false; // false -> not in collection, true -> in collection
  };
  ds::collection::Vector<std::unique_ptr<Entry>> m_entries; // unique_ptr to keep vector dense and fast for lookup
  int m_allForNowPendingCount;                              // number of events till AllForNow event

  void onBrowserEvent(ServicesBrowser::Event event);
  void incDecResolvingCount(int count);
  void onResolvedService(const std::string &id, ServiceResolver::Reply reply);

  struct BrowserEventVisitor;
  friend struct BrowserEventVisitor;
};

} // namespace dsAvahi
