/*
    Copyright (c) 2016 digitalSTROM.org," " Zurich," " Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/
#include "measurements-source.h"

#include <ds-model/model.h>
#include <ds/asio/ds-signal-mock.h>
#include <ds/asio/test.h>
#include <event.h>
#include <event/event_create.h>
#include <model/apartment-mock.h>
#include <model/zone.h>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsBackendMeasurementsSourceTest");
static const char* TAGS = "[dss][dsBackendMeasurementsSource]";

namespace {

DS_SLOT_MOCK1(MeasurementSlotMock, dsModel::Measurement);

TEST_CASE("dsBackendMeasurementsSource", TAGS) {
  ds::asio::TestLoopResource loop;
  dss::EventInterpreter eventInterpreter{nullptr};
  dss::EventQueue queue{&eventInterpreter};
  eventInterpreter.initialize();
  eventInterpreter.setEventQueue(&queue);
  auto source = dsBackend::makeMeasurementsSource(loop, eventInterpreter);
  auto nextMeasurement = MeasurementSlotMock{loop, source->next()};

  dss::ApartmentMock apartment{};
  auto& zone0 = *apartment.allocateZone(0);
  auto& zone1 = *apartment.allocateZone(1);
  const auto lastChanged = loop.system.now() - 1min;

  SECTION("Zone measurements are emitted") {
    struct X {
      dss::SensorType dssType;
      dsModel::MeasurementType dsType;
      double value = 20;
    };
    const auto& x = GENERATE(values(std::initializer_list<X>{
        {.dssType = dss::SensorType::TemperatureIndoors, .dsType = dsModel::MeasurementType::temperature},
        {.dssType = dss::SensorType::HumidityIndoors, .dsType = dsModel::MeasurementType::humidity},
        {.dssType = dss::SensorType::CO2Concentration, .dsType = dsModel::MeasurementType::carbonDioxide},
        {.dssType = dss::SensorType::BrightnessIndoors, .dsType = dsModel::MeasurementType::brightness},
        {.dssType = dss::SensorType::RoomTemperatureSetpoint, .dsType = dsModel::MeasurementType::temperatureSetpoint},
        {.dssType = dss::SensorType::RoomTemperatureControlVariable,
         .dsType = dsModel::MeasurementType::temperatureControlValue},
    }));
    INFO(x.dssType);
    INFO(x.dsType);
    INFO(x.value);
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("a"));
    REQUIRE_CALL(nextMeasurement,
                 call(dsModel::Measurement{
                     .id = dsModel::MeasurementId{"a"},
                     .type = "measurement",
                     .lastChanged = lastChanged,
                     .attributes = {.type = x.dsType,
                                    .origin = dsModel::MeasurementOriginZone{.type = "zone", .id = zone1.id()},
                                    .value = dss::sensorValueToDouble(x.dssType,
                                                                      dss::doubleToSensorValue(x.dssType, x.value))}}))
        .LR_SIDE_EFFECT(loop.stop());
    queue.pushEvent([&] {
      auto result = dss::createZoneSensorValueEvent(zone1.getGroup(0), x.dssType,
                                                    dss::doubleToSensorValue(x.dssType, x.value), dss::Dsuid{});
      result->setTimestamp(dss::DateTime{lastChanged});
      return result;
    }());
    eventInterpreter.executePendingEvent();
    loop.run();
  }

  SECTION("Outdoor measurements are emitted") {
    struct X {
      dss::SensorType dssType;
      dsModel::MeasurementType dsType;
      double value = 20;
    };
    const auto& x = GENERATE(values(std::initializer_list<X>{
        {.dssType = dss::SensorType::TemperatureOutdoors, .dsType = dsModel::MeasurementType::temperature},
        {.dssType = dss::SensorType::HumidityOutdoors, .dsType = dsModel::MeasurementType::humidity},
        {.dssType = dss::SensorType::BrightnessOutdoors, .dsType = dsModel::MeasurementType::brightness},
    }));
    INFO(x.dssType);
    INFO(x.dsType);
    INFO(x.value);
    REQUIRE_CALL(loop.system, randomUuidString()).RETURN(ds::str("a"));
    REQUIRE_CALL(nextMeasurement, call(dsModel::Measurement{
                                      .id = dsModel::MeasurementId{"a"},
                                      .type = "measurement",
                                      .lastChanged = lastChanged,
                                      .attributes = {.type = x.dsType,
                                                     .origin = dsModel::MeasurementOriginOutdoor{.type = "outdoor"},
                                                     .value = dss::sensorValueToDouble(
                                                         x.dssType, dss::doubleToSensorValue(x.dssType, x.value))}}))
        .LR_SIDE_EFFECT(loop.stop());
    queue.pushEvent([&] {
      auto result = dss::createZoneSensorValueEvent(zone0.getGroup(0), x.dssType,
                                                    dss::doubleToSensorValue(x.dssType, x.value), dss::Dsuid{});
      result->setTimestamp(dss::DateTime{lastChanged});
      return result;
    }());
    eventInterpreter.executePendingEvent();
    loop.run();
  }
}

} // namespace
