/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "module.h"

#include <ds-api/http-server.h>
#include <ds-model/model-fwd.h>
#include <ds-net/websocket.h>
#include <ds/asio/loop.h>
#include <ds/asio/resource.h>

#include "ds-backend/connector.h"
#include "measurements-source.h"
#include "signals-source.h"
#include "timed-measurements-source.h"
#include "uploader.h"
#include "wss-client.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsBackendModule");

namespace dsBackend {

class ModuleImpl : public Module, ds::asio::Resource<> {
public:
  ModuleImpl(ds::asio::ResourceBase& parent, dsBackend::ConnectorHook& busHook, dss::EventInterpreter& eventInterpreter,
             dss::Apartment& apartment, dsApi::EventHub& eventHub)
      : ds::asio::Resource<>{parent, "backend"},
        m_busHook{busHook},
        m_eventInterpreter{eventInterpreter},
        m_apartment(apartment),
        m_eventHub{eventHub} {
    // unconditionally subscribed events
    auto& eventSink = m_bus->eventSink();
    eventSink.subscribe({.type = dsModel::apartmentStructureChanged, .expiresAt = ds::Clock::time_point::max()});
    eventSink.subscribe({.type = dsModel::apartmentStatusChanged, .expiresAt = ds::Clock::time_point::max()});
  }

  dsBackend::Connector& bus() override { return *m_bus; }

private:
  dsBackend::ConnectorHook& m_busHook;
  dss::EventInterpreter& m_eventInterpreter;
  dss::Apartment& m_apartment;
  dsApi::EventHub& m_eventHub;

  boost::asio::ssl::context m_sslContext{dsNet::systemClientSslContext()};
  std::unique_ptr<dsNet::MessageStreamClient> m_streamClient = dsBackend::makeWSSClient(*this, m_sslContext);
  std::unique_ptr<dsBackend::Connector> m_bus =
      dsBackend::makeConnector(*this, "connector", *m_streamClient, m_busHook, m_eventHub);
  std::unique_ptr<dsApi::HttpServer> m_httpServer =
      dsApi::makeHttpServer(*this, {.acceptor = {.endpoint = {.address = "127.0.0.1", .port = 8089}}}, io(), *m_bus);

  std::unique_ptr<Uploader> m_uploader = makeUploader(*this, *m_bus);
  std::unique_ptr<MeasurementsSource> m_measurementsSource = makeMeasurementsSource(*this, m_eventInterpreter);
  ds::Slot<dsModel::Measurement> DS_UNIQUE_NAME(slot){
      *this, m_measurementsSource->next(), [this](auto&& x) {
        m_uploader->asyncUpload(
            dsApi::HttpRequest{.path = "/backend/v1/apartment/measurements",
                               .httpArgs = {.method = dsApi::HttpMethod::POST,
                                            .content = ds::toJsonFragment(dsApi::document(DS_FWD(x)))}});
      }};

  // Timed measurements perform blocking calls, isolate to separate thread.
  // Querying rrd can take minutes inside `syncCachedDBValues` DSD-7782
  ds::asio::Loop timedMeasurementsLoop{conf(), "dsBackendTimedMeasurementsLoop", {.warningLatency = 15min}};
  std::unique_ptr<TimedMeasurementsSource> m_timedMeasurementsSource = timedMeasurementsLoop.invoke(
      [&] { return makeTimedMeasurementsSource(timedMeasurementsLoop, m_eventInterpreter, m_apartment); });
  ds::Slot<dsModel::Measurement> DS_UNIQUE_NAME(slot){
      *this, m_timedMeasurementsSource->next(), [this](auto&& x) {
        m_uploader->asyncUpload(
            dsApi::HttpRequest{.path = "/backend/v1/apartment/measurements",
                               .httpArgs = {.method = dsApi::HttpMethod::POST,
                                            .content = ds::toJsonFragment(dsApi::document(DS_FWD(x)))}});
      }};
  std::unique_ptr<SignalsSource> m_signalsSource = makeSignalsSource(*this, m_eventInterpreter);
  ds::Slot<dsModel::Signal> DS_UNIQUE_NAME(slot){
      *this, m_signalsSource->next(), [this](auto&& x) {
        m_uploader->asyncUpload(
            dsApi::HttpRequest{.path = "/backend/v1/apartment/signals",
                               .httpArgs = {.method = dsApi::HttpMethod::POST,
                                            .content = ds::toJsonFragment(dsApi::document(DS_FWD(x)))}});
      }};
};

std::unique_ptr<Module> makeModule(ds::asio::ResourceBase& parent, dsBackend::ConnectorHook& busHook,
                                   dss::EventInterpreter& eventInterpreter, dss::Apartment& apartment,
                                   dsApi::EventHub& eventHub) {
  return std::make_unique<ModuleImpl>(parent, busHook, eventInterpreter, apartment, eventHub);
}

} // namespace dsBackend
