/*
    Copyright (c) 2016 digitalSTROM.org," " Zurich," " Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation," " either version 3 of the License," " or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful," "
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not," " see <http://www.gnu.org/licenses/>.
*/
#include "uploader.h"

#include <ds-backend/connector-mock.h>
#include <ds-model/model.h>
#include <ds/asio/promise.h>
#include <ds/asio/test.h>
#include <ds/clock-mock.h>

using namespace std::literals;
using namespace trompeloeil;

static ds::LogChannel dsLogChannel("dsBackendUploaderTest");
static const char* TAGS = "[dss][dsBackendUploader]";

namespace {

class HooksMock : public trompeloeil::mock_interface<dsBackend::UploaderHooks> {
public:
  IMPLEMENT_MOCK0(nothingToDo);
  IMPLEMENT_MOCK0(blockedOnDisconnected);
  IMPLEMENT_MOCK1(succeeded);
  IMPLEMENT_MOCK1(dropped);
};

TEST_CASE("dsBackendUploader", TAGS) {
  ds::ClockMocker clock;
  ds::asio::TestLoopResource loop;
  dsBackend::ConnectorMock bus;
  auto busIsConnected = false;
  ALLOW_CALL(bus, isConnected()).LR_RETURN(busIsConnected);
  auto hooks = HooksMock{};
  auto uploader = dsBackend::makeUploader(loop, bus, hooks);

  auto request = [&](int id) {
    return dsApi::HttpRequest{
        .path = "/backend/v1/apartment/measurements",
        .httpArgs = {.method = dsApi::HttpMethod::POST, .content = ds::toJsonFragment(ds::str(id))}};
  };

  const auto success = [] { return dsApi::HttpResult{.status = 200}; };

  GIVEN("bus is connected") {
    {
      REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
      busIsConnected = true;
      bus.emitIsConnectedChanged(true);
      loop.run();
    }

    SECTION("Single record is uploaded -> success") {
      REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(1)); _2->value = success(););
      REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
      REQUIRE_CALL(hooks, succeeded(request(1)));
      uploader->asyncUpload(request(1));
      loop.run();
    }

    SECTION("2 records are uploaded -> 2xsuccess") {
      REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(2)); _2->value = success(););
      REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(1)); _2->value = success(););
      REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
      REQUIRE_CALL(hooks, succeeded(request(2)));
      REQUIRE_CALL(hooks, succeeded(request(1)));
      uploader->asyncUpload(request(1));
      uploader->asyncUpload(request(2));
      loop.run();
    }

    SECTION("Oldest record is dropped when capacity is full") {
      std::unique_ptr<ds::Promise<dsApi::HttpResult>> asynInvokePromise;
      {
        REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(0)); asynInvokePromise = std::move(_2));
        uploader->asyncUpload(request(0));
      }
      const auto capacity = uploader->config().capacity;
      for (size_t i = 0; i < capacity; i++) {
        uploader->asyncUpload(request((i + 1)));
      }
      {
        REQUIRE_CALL(hooks, dropped(request(1)));
        uploader->asyncUpload(request((capacity + 1)));
      }

      SECTION("Oldest record is uploaded on upload success") {
        REQUIRE_CALL(hooks, succeeded(request(0)));
        REQUIRE_CALL(bus, asyncInvoke(_, _))
            .LR_SIDE_EFFECT(CHECK(_1 == request(2)); _2->value = success())
            .LR_SIDE_EFFECT(loop.stop());
        asynInvokePromise->value = success();
        asynInvokePromise.reset();
        loop.run();
      }
    }

    SECTION("Single record is uploaded with error on disconnected -> back off") {
      {
        REQUIRE_CALL(hooks, blockedOnDisconnected()).LR_SIDE_EFFECT(loop.stop());
        REQUIRE_CALL(bus, asyncInvoke(_, _))
            .LR_SIDE_EFFECT(CHECK(_1 == request(1)); bus.emitIsConnectedChanged(false););
        uploader->asyncUpload(request(1));
        loop.run();
      }

      SECTION("Connection goes up -> retry with success") {
        bus.emitIsConnectedChanged(false);

        REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
        REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(1)); _2->value = success());
        REQUIRE_CALL(hooks, succeeded(request(1)));
        bus.emitIsConnectedChanged(true);
        loop.run();
      }
    }

    SECTION("Single record is uploaded with error -> dropped") {
      REQUIRE_CALL(bus, asyncInvoke(_, _))
          .LR_SIDE_EFFECT(CHECK(_1 == request(1)); _2->value = dsApi::HttpResult{.status = 500};);
      REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
      REQUIRE_CALL(hooks, dropped(request(1)));
      uploader->asyncUpload(request(1));
      loop.run();
    }

    SECTION("Single record is uploaded with cancel error -> dropped") {
      REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(1)));
      REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
      REQUIRE_CALL(hooks, dropped(request(1)));
      uploader->asyncUpload(request(1));
      loop.run();
    }
  }

  SECTION("Single record is uploaded -> nothingToDo") {
    {
      REQUIRE_CALL(hooks, blockedOnDisconnected()).LR_SIDE_EFFECT(loop.stop());
      loop.io.post([&] { uploader->asyncUpload(request(1)); });
      loop.run();
    }

    SECTION("bus is connected -> success") {
      REQUIRE_CALL(bus, asyncInvoke(_, _)).LR_SIDE_EFFECT(CHECK(_1 == request(1)); _2->value = success(););
      REQUIRE_CALL(hooks, nothingToDo()).LR_SIDE_EFFECT(loop.stop());
      REQUIRE_CALL(hooks, succeeded(request(1)));
      bus.emitIsConnectedChanged(true);
      loop.run();
    }
  }
}

} // namespace
