/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "media-library-api.h"

#include <ds-api/service.h>
#include <ds-model/media-library.h>
#include <ds/catch/catch.h>
#include <ds/system-mock.h>
#include <ds/system.h>

#include <boost/algorithm/string/predicate.hpp>

#include "media-library-api-mock.h"

using namespace trompeloeil;
using namespace std::literals;

static const char *TAGS = "[dsMediaLibrary]";

namespace {

SCENARIO("dsMediaLibrary", TAGS) {
  const auto clipId0Path = "/usr/share/media-library/clips/clipId0.json";
  const auto clipId1Path = "/usr/share/media-library/clips/clipId1.json";
  const auto clipId2Path = "/usr/share/media-library/clips/clipId2.json";
  const auto clipId0JsonStr =
      R"({"id":"clipId0","type":"clip","lastChanged":"2020-08-06T07:18:00.000Z","attributes":{"name":{"domain":"vdc-gtin-7640156793710","message":"doorbell","context":"media-library-clip"},"category":"doorbell","resources":[{"uri":"\/clips\/doorbell.mp3","duration":5.0}]}})";
  const auto clipId1JsonStr =
      R"({"id":"clipId1","type":"clip","lastChanged":"2020-08-06T07:18:00.000Z","attributes":{"name":{"domain":"vdc-gtin-7640156793710","message":"Fire alarm with announcement","context":"media-library-clip"},"category":"alarm","resources":[{"uri":"\/clips\/fire-alarm-en.mp3","duration":25.0,"language":"en-US"},{"uri":"\/clips\/fire-alarm-de.mp3","duration":25.0,"language":"de-DE"}]}})";
  const auto clipId2JsonStr =
      R"({"id":"clipId2","type":"clip","lastChanged":"2020-08-06T07:18:00.000Z","attributes":{"name":{"domain":"vdc-gtin-7640156793710","message":"Hail warning","context":"media-library-clip"},"category":"alarm","resources":[{"uri":"\/clips\/hail-warining-en.mp3","duration":6.0,"language":"en-US"}]}})";

  const auto userClipId0Path = "/usr/share/media-library/clips/user/clipIdu0.json";
  const auto userClipId1Path = "/usr/share/media-library/clips/user/clipIdu1.json";
  const auto userClipId0JsonStr =
      R"({"id":"clipIdu0","type":"clip","lastChanged":"2020-08-06T07:18:00.000Z","attributes":{"name":{"message":"userclip0"},"category":"user","resources":[{"uri":"\/clips\/user\/userclip0.mp3"}]}})";
  const auto userClipId1JsonStr =
      R"({"id":"clipIdu1","type":"clip","lastChanged":"2020-08-06T07:18:00.000Z","attributes":{"name":{"message":"userclip1"},"category":"user","resources":[{"uri":"\/clips\/user\/userclip1.mp3"}]}})";

  auto clipId0{ds::fromJson<dsModel::Clip>(clipId0JsonStr)};
  auto clipId1{ds::fromJson<dsModel::Clip>(clipId1JsonStr)};
  auto clipId2{ds::fromJson<dsModel::Clip>(clipId2JsonStr)};
  auto clipIdu0{ds::fromJson<dsModel::Clip>(userClipId0JsonStr)};
  auto clipIdu1{ds::fromJson<dsModel::Clip>(userClipId1JsonStr)};
  dsModel::Clips clips{{clipId0, clipId1, clipId2, clipIdu0, clipIdu1}};

  SECTION("initialize") {
    auto system = ds::SystemMock();
    sequence seq;
    REQUIRE_CALL(system, lsRegularFiles(_, _))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == "/usr/share/ds-media-library/clips/"); CHECK(_2 == ".json"))
        .LR_RETURN(std::vector<std::string>{clipId0Path, clipId1Path, clipId2Path});

    REQUIRE_CALL(system, readFile(_))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == clipId0Path))
        .LR_RETURN(clipId0JsonStr);
    REQUIRE_CALL(system, readFile(_))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == clipId1Path))
        .LR_RETURN(clipId1JsonStr);
    REQUIRE_CALL(system, readFile(_))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == clipId2Path))
        .LR_RETURN(clipId2JsonStr);

    REQUIRE_CALL(system, lsRegularFiles(_, _))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == "/usr/share/ds-media-library/clips/user"); CHECK(_2 == ".json"))
        .LR_RETURN(std::vector<std::string>{userClipId0Path, userClipId1JsonStr});

    REQUIRE_CALL(system, readFile(_))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == userClipId0Path))
        .LR_RETURN(userClipId0JsonStr);
    REQUIRE_CALL(system, readFile(_))
        .IN_SEQUENCE(seq)
        .LR_SIDE_EFFECT(CHECK(_1 == userClipId1JsonStr))
        .LR_RETURN(userClipId1JsonStr);

    auto mediaLibraryApi = dsMediaLibrary::makeMediaLibraryApi(system);
    auto mediaLibraryService = dsMediaLibrary::makeMediaLibraryService(*mediaLibraryApi);

    CHECK(mediaLibraryApi->clips().clips == clips.clips);
    CHECK(mediaLibraryApi->clip(dsModel::ClipId{"clipId0"}) == clipId0);
    CHECK(mediaLibraryApi->clip(dsModel::ClipId{"clipId1"}) == clipId1);
    CHECK(mediaLibraryApi->clip(dsModel::ClipId{"clipId2"}) == clipId2);
    CHECK(mediaLibraryApi->clip(dsModel::ClipId{"clipIdu0"}) == clipIdu0);
    CHECK(mediaLibraryApi->clip(dsModel::ClipId{"clipIdu1"}) == clipIdu1);
  }

  SECTION("get clips") {
    auto mediaLibraryApi = std::make_unique<dsMediaLibrary::MediaLibraryApiMock>();
    auto mediaLibraryService = dsMediaLibrary::makeMediaLibraryService(*mediaLibraryApi);

    REQUIRE_CALL(*mediaLibraryApi, clips()).LR_RETURN(clips);
    const auto result = mediaLibraryService->invoke(
        dsApi::HttpRequest{.path = "/clips", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    CHECK(result.status == 200);
  }

  SECTION("get clip") {
    auto mediaLibraryApi = std::make_unique<dsMediaLibrary::MediaLibraryApiMock>();
    auto mediaLibraryService = dsMediaLibrary::makeMediaLibraryService(*mediaLibraryApi);

    const auto args = dsModel::ClipId{"clipId0"};
    REQUIRE_CALL(*mediaLibraryApi, clip(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(clipId0);
    const auto result = mediaLibraryService->invoke(
        dsApi::HttpRequest{.path = "/clips/clipId0", .httpArgs = {.method = dsApi::HttpMethod::GET}});
    CHECK(result.status == 200);
  }

  SECTION("create clip") {
    auto mediaLibraryApi = std::make_unique<dsMediaLibrary::MediaLibraryApiMock>();
    auto mediaLibraryService = dsMediaLibrary::makeMediaLibraryService(*mediaLibraryApi);

    const auto clipId0 = dsModel::ClipId{"clipId0"};
    const auto args = dsModel::Clip{.attributes{.name = {.message = "clip name 0"},
                                                .category = dsModel::ClipCategoryId::user,
                                                .resources = {dsModel::ClipResource{.uri = "/tmp/a6836fa160c5.mp3"}}}};
    REQUIRE_CALL(*mediaLibraryApi, addClip(_)).LR_SIDE_EFFECT(CHECK(_1 == args);).LR_RETURN(clipId0);
    const auto result = mediaLibraryService->invoke(dsApi::HttpRequest{
        .path = "/clips",
        .httpArgs =
            {.method = dsApi::HttpMethod::POST,
             .content = ds::fromJson<ds::JsonFragment>(
                 R"({"data":{"attributes":{"name":{"message":"clip name 0"},"category":"user","resources":[{"uri":"/tmp/a6836fa160c5.mp3"}]}}})")},
    });
    CHECK(result.status == 201);
  }

  SECTION("delete clip") {
    auto mediaLibraryApi = std::make_unique<dsMediaLibrary::MediaLibraryApiMock>();
    auto mediaLibraryService = dsMediaLibrary::makeMediaLibraryService(*mediaLibraryApi);

    const auto args = dsModel::ClipId{"clipId0"};
    REQUIRE_CALL(*mediaLibraryApi, deleteClip(_)).LR_SIDE_EFFECT(CHECK(_1 == args););
    const auto result = mediaLibraryService->invoke(
        dsApi::HttpRequest{.path = "/clips/clipId0", .httpArgs = {.method = dsApi::HttpMethod::DELETE}});
    CHECK(result.status == 204);
  }

  SECTION("update clip") {
    auto mediaLibraryApi = std::make_unique<dsMediaLibrary::MediaLibraryApiMock>();
    auto mediaLibraryService = dsMediaLibrary::makeMediaLibraryService(*mediaLibraryApi);

    dsApi::PatchOperations patchOperations{};
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = "/name", .value = ds::toJsonFragment("new clip name")});
    patchOperations.push_back({.op = dsApi::PatchOperationType::replace,
                               .path = "/uri",
                               .value = ds::toJsonFragment("/tmp/a6836fa160c5.mp3")});
    patchOperations.push_back(
        {.op = dsApi::PatchOperationType::replace, .path = "/duration", .value = ds::toJsonFragment(10)});

    REQUIRE_CALL(*mediaLibraryApi, setClipName(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == dsModel::ClipId{"clipId0"}); CHECK(_2 == "new clip name"));
    REQUIRE_CALL(*mediaLibraryApi, setClipUri(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == dsModel::ClipId{"clipId0"}); CHECK(_2 == "/tmp/a6836fa160c5.mp3"));
    REQUIRE_CALL(*mediaLibraryApi, setClipDuration(_, _))
        .LR_SIDE_EFFECT(CHECK(_1 == dsModel::ClipId{"clipId0"}); CHECK(_2 == 10s));
    const auto result = mediaLibraryService->invoke(dsApi::HttpRequest{
        .path = "/clips/clipId0",
        .httpArgs = {.method = dsApi::HttpMethod::PATCH, .content = ds::toJsonFragment(patchOperations)}});
    CHECK(result.status == 204);
  }
}

} // namespace
