/*
    Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/common.h>
#include <ds/log.h>

#include <boost/asio/io_service.hpp>
#include <boost/intrusive/list.hpp>
#include <functional>

#include "resource.h"

/// Classes ds::Signal and ds::Slot provide variant of observer pattern.
/// See https://en.wikipedia.org/wiki/Signals_and_slots.
///
/// Design goals:
/// * exeption thrown in slot is not propagated to signal caller
/// * slots are executed asynchronously,
///     * not blocking the signal caller
///     * not blocking other slots
/// * slots are automatically disconnected when they go out of scope.
/// * slots never calls callback after it is destroyed
///
/// There are quite a few existing implementations.
/// But all investigated ones require extra wrapper
/// to fulfill the desired design goals.
/// Implementation based on std::function
/// and on boost::intrusive::list<auto_unlink> is
/// simple enough and avoids extra layering and dependencies.
///
/// Known problems:
/// - connect, disconnect and emit calls are _not_ thread safe.
/// - deleting slot when callback is posted leads to crash
/// This problems must be fixed once slots are dynamic
///
/// Usage:
///
/// class Publisher {
/// public:
///   ds::Slots<std::string>& next() { return m_next; }
/// private:
///   ds::Signal<std::string> m_next;
///   void emit(const std::string& x) { m_next.emit(x); }
/// }
///
/// class Observer {
/// private:
///   ds::Slot<std::string> DS_UNIQUE_NAME(slot){io, publisher.next(), [this](std::string&& x) { onNext(std::move(x));
///   }};
/// }

namespace ds {

template <class... Args>
class Slots;

/// Slot connectable to Signal.
template <class... Args>
class Slot : public boost::intrusive::list_base_hook<boost::intrusive::link_mode<boost::intrusive::auto_unlink>> {
public:
  Slot() = default;

  /// Connect callback posted to signal.
  ///
  /// Callback is posted to io_service for execution.
  ///
  /// Link is automatically removed when slot is destroyed.
  Slot(ds::asio::ResourceBase& resource, Slots<Args...>& slots, std::function<void(Args...)>&& callback);

  void post(Args... args) const {
    // TODO(c++20): avoid std::tuple, std::apply ... [args=std::forward<Args...>(args)...]
    m_resource.io().post(
        [session = std::weak_ptr{m_session}, this, tuple = std::make_tuple(std::forward<Args...>(args)...)]() mutable {
          if (session.lock()) { // discard invocation if slot was destroyed
            std::apply(m_callback, std::move(tuple));
          }
        });
  }

private:
  ds::asio::ResourceBase& m_resource;
  std::function<void(Args...)> m_callback;
  std::shared_ptr<void> m_session = std::make_shared<bool>(true);
};

template <class... Args>
class Slots : public boost::intrusive::list<Slot<Args...>, boost::intrusive::constant_time_size<false>> {};

/// Signal invoking all connected slots.
///
/// `Slots&` allows list linking and unlinking of slots - for observers.
/// `Signal &` is for publishers.
template <class... Args>
class Signal : boost::noncopyable {
public:
  /// Invoke connected slots.
  void emit(const Args&... args) const;

  /// implicit cast to Slots interface that allows only linking and unlinking of slots.
  operator Slots<Args...>&() { return m_slots; }
  Slots<Args...>& slots() { return m_slots; }

private:
  Slots<Args...> m_slots;
};

template <class... Args>
Slot<Args...>::Slot(ds::asio::ResourceBase& resource, Slots<Args...>& slots, std::function<void(Args...)>&& callback)
    : m_resource{resource}, m_callback{std::move(callback)} {
  slots.push_back(*this);
}

template <class... Args>
void Signal<Args...>::emit(const Args&... args) const {
  for (auto& slot : m_slots) {
    slot.post(args...);
  }
}

} // namespace ds
