/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "latency-watchdog.h"
#include <ds/asio/loop.h>
#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/clock-mock.h>
#include <ds/conf-mock.h>
#include <ds/log.h>
#include <trompeloeil.hpp>

static ds::LogChannel dsLogChannel("dsAsioLatencyWatchdogTest");

static const char* TAGS = "[dsAsioLatencyWatchdog][dsAsio][ds]";

using namespace std::literals;
using namespace trompeloeil;

struct Hooks {
public:
  MAKE_MOCK0(fire, void());
  MAKE_MOCK0(pinged, void());
  MAKE_MOCK0(ponged, void());
};

TEST_CASE("dsAsioLatencyWatchdog", TAGS) {
  auto clock = ds::ClockMocker{};
  auto watchedLoop = ds::asio::TestLoop{};
  auto watchdogLoop = ds::asio::TestLoop{};

  const auto latency = 1000ms;
  auto hooks = Hooks{};
  auto latencyWatchdog = ds::asio::LatencyWatchdog{watchedLoop.io,
                                                   watchdogLoop.io,
                                                   latency,
                                                   {.fire = [&](auto&&) { hooks.fire(); },
                                                    .pinged = [&]() { hooks.pinged(); },
                                                    .ponged = [&]() { hooks.ponged(); }}};

  SECTION("can be created and destroyed") {}

  SECTION("does not fire when watchedIo is executing handlers") {
    for (auto i = 0; i < 5; i++) {
      clock.advance(latency * 2);
      REQUIRE_CALL(hooks, pinged()).LR_SIDE_EFFECT(watchdogLoop.stop());
      watchdogLoop.run();

      clock.advance(latency * 2);
      REQUIRE_CALL(hooks, ponged()).LR_SIDE_EFFECT(watchedLoop.stop());
      watchedLoop.run();
    }

    SECTION("fires when watchedIo does not execute handlers for 2*latency.") {
      clock.advance(latency * 2);
      REQUIRE_CALL(hooks, pinged()).LR_SIDE_EFFECT(watchdogLoop.stop());
      watchdogLoop.run();

      clock.advance(latency * 2);
      REQUIRE_CALL(hooks, fire()).LR_SIDE_EFFECT(watchdogLoop.stop());
      watchdogLoop.run();
    }
  }
}

// real assertion. Not part of test suite
TEST_CASE("dsAsioLatencyWatchdog assert", "[!hide]") {
  auto conf = ds::ConfMock{};
  const auto config = ds::asio::LatencyWatchdogConfig{.warningLatency = 1s, .assertLatency = 3s};
  auto loop = ds::asio::Loop{conf, "testLoop", ds::asio::LatencyWatchdogConfig{config}};

  DS_NOTICE("Event loop running.");
  std::this_thread::sleep_for(config.warningLatency * 2);

  DS_NOTICE("Event loop blocked.");
  loop.io().post([&] { std::this_thread::sleep_for(config.assertLatency * 2); });
  std::this_thread::sleep_for(config.assertLatency * 3);
}
