/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "enum.h"
#include <ds/catch/catch.h>
#include <ds/str.h>
#include <array>

static const char* TAGS = "[dsEnum][ds]";

DS_REFLECTED_ENUM((ds)(enumTest), Transport, (car)(plane))
DS_REFLECTED_ENUM_WITH_INITIALIZERS((ds)(enumTest), TransportWithInitializers, ((car0)(0))((plane2)(2)))

DS_REFLECTED_ENUM((ds)(enumTest), AB, (a)(b));
DS_REFLECTED_ENUM((ds)(enumTest), BC, (b)(c));

template <>
struct DSToEnum<ds::enumTest::BC, ds::enumTest::AB, void> {
  static std::optional<ds::enumTest::BC> tryTo(ds::enumTest::AB x) {
    switch (x) {
      case ds::enumTest::AB::a:
        return {};
      case ds::enumTest::AB::b:
        return ds::enumTest::BC::b;
    }
    return {};
  }
};

DS_REFLECTED_ANONYMOUS_ENUM(GlobalTransport, (car)(plane))

namespace {

using ds::enumTest::Transport;
using ds::enumTest::TransportWithInitializers;

TEST_CASE("For reflected enum", TAGS) {
  SECTION("tryFromEnum<const char*> returns matching string") {
    CHECK(ds::tryFromEnum<const char*>(Transport::car).value() == std::string("car"));
    CHECK(ds::tryFromEnum<const char*>(Transport::plane).value() == std::string("plane"));
  }

  SECTION("tryFromEnum<const char*> returns none on non-matching option") {
    CHECK(bool(ds::tryFromEnum<const char*>(static_cast<Transport>(100))) == false);
  }

  SECTION("tryFromEnum<std::string> returns matching string") {
    CHECK(ds::tryFromEnum<std::string>(Transport::car).value() == "car");
    CHECK(ds::tryFromEnum<std::string>(Transport::plane).value() == "plane");
  }

  SECTION("tryFromEnum<std::string> returns none on non-matching option") {
    CHECK(bool(ds::tryFromEnum<std::string>(static_cast<Transport>(100))) == false);
  }

  SECTION("fromEnum<std::string> returns matching string") {
    CHECK(ds::fromEnum<std::string>(Transport::car) == "car");
    CHECK(ds::fromEnum<std::string>(Transport::plane) == "plane");
  }

  SECTION("fromEnum<std::string> returns default constructed string") {
    CHECK(ds::fromEnum<std::string>(static_cast<Transport>(100)) == std::string{});
    CHECK(ds::fromEnum<std::string>(static_cast<Transport>(-100)) == std::string{});
  }

  SECTION("tryToEnum returns matching value") {
    CHECK(ds::tryToEnum<Transport>("car").value() == Transport::car);
    CHECK(ds::tryToEnum<Transport>("plane").value() == Transport::plane);
  }

  SECTION("tryToEnum returns none on non-matching option") {
    CHECK(bool(ds::tryToEnum<Transport>("notAMatchingOption")) == false);
  }

  SECTION("tryToEnum accepts std::string") {
    CHECK(ds::tryToEnum<Transport>(std::string("car")).value() == Transport::car);
  }

  SECTION("toEnum returns matching value") {
    CHECK(ds::toEnum<Transport>("car") == Transport::car);
    CHECK(ds::toEnum<Transport>("plane") == Transport::plane);
    CHECK(ds::toEnum<Transport>(std::string("plane")) == Transport::plane);
  }

  SECTION("toEnum returns default value for unknown options") {
    CHECK(ds::toEnum<Transport>("unknown") == Transport{});
    CHECK(ds::toEnum<Transport>("5") == Transport{});
  }

  SECTION("requireEnum return enum for known options") {
    CHECK(ds::requireEnum<Transport>("car") == Transport::car);
    CHECK(ds::requireEnum<Transport>("plane") == Transport::plane);
    CHECK(ds::requireEnum<Transport>(std::string("car")) == Transport::car);
  }

  SECTION("requireEnum throws for unknown options") {
    CHECK_THROWS_FIND(ds::requireEnum<Transport>("notAMatchingOption"),
                      "Unknown ::ds::enumTest::Transport enum option. value:notAMatchingOption");
    CHECK_THROWS_FIND(ds::requireEnum<Transport>(std::string("notAMatchingOption")),
                      "Unknown ::ds::enumTest::Transport enum option. value:notAMatchingOption");
  }

  SECTION("tryToEnum accepts std::optional<std::string>") {
    CHECK(ds::tryToEnum<Transport>(std::optional<std::string>("car")).value() == Transport::car);
    CHECK(ds::tryToEnum<Transport>(std::optional<std::string>("plane")).value() == Transport::plane);
    CHECK(ds::tryToEnum<Transport>(std::optional<std::string>("unknown")) == std::nullopt);
    CHECK(ds::tryToEnum<Transport>(std::optional<std::string>()) == std::nullopt);
  }

  SECTION("tryToEnum accepts int") {
    CHECK(ds::tryToEnum<Transport>(0).value() == Transport::car);
    CHECK(ds::tryToEnum<Transport>(1).value() == Transport::plane);
    CHECK(ds::tryToEnum<Transport>(2) == std::nullopt);
    CHECK(ds::tryToEnum<Transport>(-1) == std::nullopt);
  }

  SECTION("tryToEnum accepts std::optional<int>") {
    CHECK(ds::tryToEnum<Transport>(std::optional<int>(0)).value() == Transport::car);
    CHECK(ds::tryToEnum<Transport>(std::optional<int>(1)).value() == Transport::plane);
    CHECK(ds::tryToEnum<Transport>(std::optional<int>(2)) == std::nullopt);
    CHECK(ds::tryToEnum<Transport>(std::optional<int>(-1)) == std::nullopt);
    CHECK(ds::tryToEnum<Transport>(std::optional<int>()) == std::nullopt);
  }

  SECTION("ostream << writes matching values as string") {
    CHECK(ds::str(Transport::car) == "car");
    CHECK(ds::str(Transport::plane) == "plane");
    CHECK(ds::str(static_cast<Transport>(100)) == "");
  }

  SECTION("isValidEnum == true for matching value ") { CHECK(ds::isValidEnum(Transport::car)); }

  SECTION("isValidEnum == false for non-matching value ") { CHECK(!ds::isValidEnum(static_cast<Transport>(100))); }
}

TEST_CASE("For reflected global enum", TAGS) {
  SECTION("smoke test") {
    CHECK(ds::tryFromEnum<const char*>(GlobalTransport::car).value() == std::string("car"));
    CHECK(ds::tryToEnum<GlobalTransport>(std::optional<int>(0)).value() == GlobalTransport::car);
    CHECK(ds::str(GlobalTransport::car) == "car");
  }
}

TEST_CASE("For reflected enum with initializers", TAGS) {
  SECTION("smoke test") {
    CHECK(ds::tryFromEnum<const char*>(TransportWithInitializers::car0).value() == std::string("car0"));
    CHECK(ds::tryFromEnum<const char*>(TransportWithInitializers::plane2).value() == std::string("plane2"));
    CHECK(ds::tryToEnum<TransportWithInitializers>(std::optional<int>(0)).value() == TransportWithInitializers::car0);
    CHECK(ds::tryToEnum<TransportWithInitializers>(std::optional<int>(2)).value() == TransportWithInitializers::plane2);
    CHECK(ds::str(TransportWithInitializers::car0) == "car0");
  }
}

SCENARIO("Conversion between enums", TAGS) {
  GIVEN("DSToEnum<BC, AB>::tryTo)") {
    THEN("tryToEnum<> converts AB -> BC") {
      CHECK(ds::tryToEnum<ds::enumTest::BC>(ds::enumTest::AB::a) == std::nullopt);
      CHECK(ds::tryToEnum<ds::enumTest::BC>(ds::enumTest::AB::b) == ds::enumTest::BC::b);
    }

    THEN("toEnum<> converts AB -> BC") {
      CHECK(ds::toEnum<ds::enumTest::BC>(ds::enumTest::AB::a) == ds::enumTest::BC{});
      CHECK(ds::toEnum<ds::enumTest::BC>(ds::enumTest::AB::b) == ds::enumTest::BC::b);
    }

    THEN("requireEnum<> converts AB -> BC") {
      CHECK_THROWS_FIND(ds::requireEnum<ds::enumTest::BC>(ds::enumTest::AB::a),
                        "Unknown ::ds::enumTest::BC enum option. value:a");
      CHECK(ds::requireEnum<ds::enumTest::BC>(ds::enumTest::AB::b) == ds::enumTest::BC::b);
    }
  }
}
} // namespace
