/*
 *  Copyright (c) 2015 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "event_create.h"

#include <ds/log.h>
#include <vdcapi.pb.h>

#include "event/event_fields.h"
#include "ds485types.h"
#include "model/state.h"
#include "util.h"

namespace dss {

namespace EventName {
  const std::string CallScene = "callScene";
  const std::string CallSceneBus = "callSceneBus";
  const std::string DeviceSensorValue = "deviceSensorValue";
  const std::string DeviceSensorEvent = "deviceSensorEvent";
  const std::string DeviceStatus = "deviceStatusEvent";
  const std::string DeviceInvalidSensor = "deviceInvalidSensor";
  const std::string DeviceBinaryInputEvent = "deviceBinaryInputEvent";
  const std::string DeviceCustomActionChangedEvent = "deviceCustomActionChangedEvent";
  const std::string DeviceActionEvent = "deviceActionEvent";
  const std::string DeviceStateEvent = "deviceStateEvent";
  const std::string DeviceEventEvent = "deviceEventEvent";
  const std::string DeviceButtonClick = "buttonClick";
  const std::string IdentifyBlink = "blink";
  const std::string ExecutionDenied = "executionDenied";
  const std::string Running = "running";
  const std::string ModelReady = "model_ready";
  const std::string UndoScene = "undoScene";
  const std::string ZoneSensorValue = "zoneSensorValue";
  const std::string ZoneSensorError = "zoneSensorError";
  const std::string StateChange = "stateChange";
  const std::string AddonStateChange = "addonStateChange";
  const std::string HeatingEnabled = "HeatingEnabled";
  const std::string HeatingSystemCapability ="HeatingSystemCapability";
  const std::string HeatingZoneCapability ="HeatingZoneCapability";
  const std::string HeatingControllerSetup = "HeatingControllerSetup";
  const std::string HeatingControllerValue = "HeatingControllerValue";
  const std::string HeatingControllerState = "HeatingControllerState";
  const std::string OldStateChange = "oldStateChange";
  const std::string HighLevelEvent = "highlevelevent";
  const std::string AddonToCloud = "AddonToCloud";
  const std::string HeatingValveProtection = "execute_valve_protection";
  const std::string DeviceHeatingTypeChanged = "DeviceHeatingTypeChanged";
  const std::string UpdateAutoselect = "UpdateAutoselect";
  const std::string DebugMonitorUpdate = "debugMonitorUpdate";
  const std::string Sunshine = "sunshine";
  const std::string FrostProtection = "frostprotection";
  const std::string HeatingModeSwitch = "heating_mode_switch";
  const std::string BuildingService = "building_service";
  const std::string OperationLock = "operation_lock";
  const std::string ClusterConfigLock = "cluster_config_lock";
  const std::string CheckSensorValues = "check_sensor_values";
  const std::string DeviceEvent = "DeviceEvent";
  const std::string TemperatureControllersElection = "temperatureControllersElection";
  const std::string Signal = "SIGNAL";
  const std::string WebSessionCleanup = "webSessionCleanup";
  const std::string SendMail = "sendmail";
  const std::string ButtonClickBus = "buttonClickBus";
  const std::string DSMeterReady = "dsMeter_ready";
  const std::string DSMeterChanged = "dsMeterChanged";
  const std::string ZoneChanged = "zoneChanged";
  const std::string ExecutionDeniedDigestCheck = "execution_denied_digest_check";
  const std::string DevicesFirstSeen = "devices_first_seen";
  const std::string DatabaseImported = "database_imported";
  const std::string ButtonDeviceAction = "buttonDeviceAction";

  const std::string SceneNameChanged = "scene_name_changed";

  const std::string ReexportTimings = "reexport_timings";
  const std::string AutoClusterUpdate = "autocluster_update";
  const std::string LockClusterUpdate = "lockClusterUpdate";
  const std::string LockClusterRefresh = "lockClusterRefresh";
  const std::string ClusterCheckWind = "clusterCheckWind";
  const std::string JSPropertyStore = "js_property_store";
  const std::string AddonReset = "addon-reset";
  const std::string IdentifyDevice = "identifyDevice";
  const std::string UpdateShadeStopTimings = "update_shade_stop_timings";
  const std::string UpdateShadeStopTimingsDelayed = "update_shade_stop_timings_delayed";

  const std::string DeviceBulkSetConfig = "deviceBulkSetConfig";
  const std::string DeviceBulkSetConfigFailed = "deviceBulkSetConfigFailed";
  const std::string DeviceBulkSetConfigCancelled = "deviceBulkSetConfigCancelled";
  const std::string RetryRegistration = "retry_registration";
  const std::string CacheUpdateFinished = "cache_update_finished";
  const std::string ApartmentProxyStateChanged = "apartmentProxyStateChanged";
  const std::string ApartmentProxyDeviceTimeout = "apartmentProxyDeviceTimeout";
  const std::string ApartmentProxyDeviceDimmingTimeout = "apartmentProxyDeviceDimmingTimeout";
  const std::string SunPositionSensorUpdate = "sun_position_sensor_update";
  const std::string StopDsmTimer = "stop_dsm_timer";
}

std::shared_ptr<Event>
createDeviceStatusEvent(std::shared_ptr<DeviceReference> _devRef,
                        int _index, int _value)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceStatus, _devRef);
  event->setProperty("statusIndex", intToString(_index));
  event->setProperty("statusValue", intToString(_value));
  return event;
}

std::shared_ptr<Event>
createDeviceBinaryInputEvent(std::shared_ptr<DeviceReference> _devRef,
                             int _index, BinaryInputType _type, BinaryInputStateValue _state)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceBinaryInputEvent, _devRef);
  event->setProperty("inputIndex", intToString(_index));
  event->setProperty("inputType", _type);
  event->setProperty("inputState", _state);
  return event;
}

std::shared_ptr<Event>
createDeviceCustomActionChangedEvent(std::shared_ptr<DeviceReference> _devRef,
    const std::string& name, const std::string& action, const std::string& title,
    const vdcapi::PropertyElement& params)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceCustomActionChangedEvent, _devRef);

  Properties actionParams;
  actionParams.set("customActionId", name);
  actionParams.set("actionId", action);
  actionParams.set("customActionTitle", title);

  for (int n = 0; n < params.elements_size(); n++) {
    const vdcapi::PropertyElement& pelement = params.elements(n);
    if (!pelement.has_name() || !pelement.has_value()) {
      continue;
    }
    actionParams.set("params." + pelement.name(), propertyValue2String(pelement.value()));
  }

  event->setProperties(actionParams);
  return event;
}

std::shared_ptr<Event>
createDeviceActionChangedEvent(std::shared_ptr<DeviceReference> _devRef,
    const std::string& name, const std::string& title)
{
  auto event = std::make_shared<Event>(EventName::DeviceCustomActionChangedEvent, _devRef);

  Properties actionParams;
  actionParams.set("dynamicActionId", name);
  actionParams.set("dynamicActionTitle", title);
  event->setProperties(actionParams);
  return event;
}

std::shared_ptr<Event>
createDeviceActionEvent(std::shared_ptr<DeviceReference> _devRef, const std::string& name,
    const vdcapi::PropertyElement& params)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceActionEvent, _devRef);

  Properties actionParams;
  actionParams.set("actionId", name);
  for (int n = 0; n < params.elements_size(); n++) {
    const vdcapi::PropertyElement& pelement = params.elements(n);
    if (!pelement.has_name() || !pelement.has_value()) {
      continue;
    }
    actionParams.set("params." + pelement.name(), propertyValue2String(pelement.value()));
  }
  event->setProperties(actionParams);

  return event;
}

std::shared_ptr<Event>
createDeviceStateEvent(std::shared_ptr<DeviceReference> _devRef,
                             const std::string& name, const std::string& value)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceStateEvent, _devRef);
  event->setProperty("stateId", name);
  event->setProperty("value", value);
  return event;
}

std::shared_ptr<Event>
createDeviceEventEvent(std::shared_ptr<DeviceReference> _devRef, const std::string& name)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceEventEvent, _devRef);
  event->setProperty("eventId", name);
  return event;
}

std::shared_ptr<Event>
createDeviceSensorValueEvent(std::shared_ptr<DeviceReference> _devRef,
                             int _index, SensorType _type, int _value)
{
  std::shared_ptr<Event> event;
  // TODO ensure sensorType valid or implement fallback
  double floatValue = sensorValueToDouble(_type, _value);

  event = std::make_shared<Event>(EventName::DeviceSensorValue, _devRef);
  event->setProperty(ef_sensorIndex, intToString(_index));
  event->setProperty("sensorType", _type);
  event->setProperty("sensorValue", intToString(_value));
  event->setProperty("sensorValueFloat", doubleToString(floatValue));
  return event;
}

std::shared_ptr<Event>
createDeviceSensorValueExEvent(std::shared_ptr<DeviceReference> _devRef,
                             int _index, SensorType _type, double _value, uint32_t _age,
                             uint32_t _contextId, const std::string& _contextMsg)
{
  std::shared_ptr<Event> event;
  DateTime now;

  event = std::make_shared<Event>(EventName::DeviceSensorValue, _devRef);
  event->setProperty(ef_sensorIndex, intToString(_index));
  event->setProperty("sensorType", _type);
  event->setProperty("sensorValueFloat", doubleToString(_value));
  event->setProperty("contextId", intToString(_contextId));
  event->setProperty("contextMsg", _contextMsg);
  if (_age > 0) {
    now = now.addMilliSeconds((int) _age * (-1));
    event->setProperty("contextTimestamp", now.toISO8601_ms());
  }
  return event;
}

std::shared_ptr<Event>
createDeviceInvalidSensorEvent(std::shared_ptr<DeviceReference> _devRef,
                               int _index, SensorType _type, const DateTime& _ts)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::DeviceInvalidSensor, _devRef);
  event->setProperty(ef_sensorIndex, intToString(_index));
  event->setProperty("sensorType", _type);
  event->setProperty("lastValueTS", _ts.toISO8601_ms());
  return event;
}

std::shared_ptr<Event>
createZoneSensorValueEvent(std::shared_ptr<Group> _group, SensorType _type,
                           int _value, const dsuid_t &_sourceDevice)
{
  std::shared_ptr<Event> event;
  // TODO ensure sensorType valid or implement fallback
  double floatValue = sensorValueToDouble(_type, _value);

  event = std::make_shared<Event>(EventName::ZoneSensorValue, _group);
  event->setProperty("sensorType", _type);
  event->setProperty("sensorValue", intToString(_value));
  event->setProperty("sensorValueFloat", doubleToString(floatValue));
  event->setProperty("originDSID", dsuid2str(_sourceDevice));
  return event;
}

std::shared_ptr<Event>
createZoneSensorErrorEvent(std::shared_ptr<Group> _group, SensorType _type,
                      const DateTime& _ts)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::ZoneSensorError, _group);
  event->setProperty("sensorType", _type);
  event->setProperty("lastValueTS", _ts.toISO8601_ms());
  return event;
}

std::shared_ptr<Event>
createGroupCallSceneEvent(std::shared_ptr<Group> _group, int _sceneID,
                          int _groupID, int _zoneID,
                          const CallOrigin& _callOrigin,
                          const dsuid_t& _originDSUID,
                          const std::string& _originToken,
                          bool _forced)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::CallScene, _group);
  event->setProperty(ef_sceneID, intToString(_sceneID));
  event->setProperty("groupID", intToString(_groupID));
  event->setProperty("zoneID", intToString(_zoneID));
  event->setProperty(ef_originDSUID, dsuid2str(_originDSUID));
  event->setProperty(ef_callOrigin, _callOrigin);
  event->setProperty("originToken", _originToken);
  if (_forced) {
    event->setProperty(ef_forced, "true");
  }
  return event;
}

std::shared_ptr<Event>
createGroupUndoSceneEvent(std::shared_ptr<Group> _group, int _sceneID,
                          int _groupID, int _zoneID,
                          const CallOrigin& _callOrigin,
                          const dsuid_t& _originDSUID,
                          const std::string& _originToken)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::UndoScene, _group);
  event->setProperty(ef_sceneID, intToString(_sceneID));
  event->setProperty("groupID", intToString(_groupID));
  event->setProperty("zoneID", intToString(_zoneID));
  event->setProperty(ef_originDSUID, dsuid2str(_originDSUID));
  event->setProperty(ef_callOrigin, _callOrigin);
  event->setProperty("originToken", _originToken);
  return event;
}

std::shared_ptr<Event>
createStateChangeEvent(std::shared_ptr<State> _state, int _oldstate,
                       CallOrigin _callOrigin)
{
  std::shared_ptr<Event> event;

  if (_state->getType() == StateType_Script) {
    event = std::make_shared<Event>(EventName::AddonStateChange, _state);
    event->setProperty("scriptID", _state->getProviderService());
  } else {
    event = std::make_shared<Event>(EventName::StateChange, _state);
  }

  event->setProperty("statename", _state->getName());
  event->setProperty("state", _state->toString());
  event->setProperty("value", intToString(_state->getState()));
  event->setProperty("oldvalue", intToString(_oldstate));
  event->setProperty(ef_callOrigin, _callOrigin);
  return event;
}

std::shared_ptr<Event>
createActionDenied(const std::string &_type, const std::string &_name,
                   const std::string &_source, const std::string &_reason)
{
  std::shared_ptr<Event> event;

  event = std::make_shared<Event>(EventName::ExecutionDenied);
  event->setProperty("action-type", _type);
  event->setProperty("action-name", _name);
  event->setProperty("source-name", _source);
  event->setProperty("reason", _reason);
  return event;
}

/**
 * this function is used only by unit tests
 * the real event is created from java script
 * TODO: - add meta-event description of required fileds
 *       - upon raise event verify event matches description
 */
std::shared_ptr<Event>
createHeatingEnabled(bool _heatingEnabled, bool _coolingEnabled)
{
  std::shared_ptr<Event> event;

  event = std::make_shared<Event>(EventName::HeatingEnabled);
  event->setProperty("HeatingEnabled", _heatingEnabled ? "true" : "false");
  event->setProperty("CoolingEnabled", _coolingEnabled ? "true" : "false");
  return event;
}

std::shared_ptr<Event>
  createHeatingSystemCapability(bool _heatingSupported, bool _coolingSupported)
{
  std::shared_ptr<Event> event;

  event = std::make_shared<Event>(EventName::HeatingSystemCapability);
  event->setProperty("HeatingSupported", _heatingSupported ? "true" : "false");
  event->setProperty("CoolingSupported", _coolingSupported ? "true" : "false");
  return event;
}

std::shared_ptr<Event>
createHeatingControllerConfig(int _zoneID, const dsuid_t &_ctrlDsuid,
                              const ZoneHeatingConfigSpec_t &_config)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::HeatingControllerSetup);

  event->setProperty("ZoneID", intToString(_zoneID));
  event->setProperty("ControlDSUID", dsuid2str(_ctrlDsuid));
  event->setProperty("ControlMode", intToString(static_cast<uint8_t>(_config.mode)));
  event->setProperty("EmergencyValue", intToString(_config.EmergencyValue - 100));
  switch (_config.mode) {
    case HeatingControlMode::OFF:
      break;
    case HeatingControlMode::PID:
      event->setProperty("CtrlKp", doubleToString((double)_config.Kp * 0.025));
      event->setProperty("CtrlTs", intToString(_config.Ts));
      event->setProperty("CtrlTi", intToString(_config.Ti));
      event->setProperty("CtrlKd", intToString(_config.Kd));
      event->setProperty("CtrlImin", doubleToString((double)_config.Imin * 0.025));
      event->setProperty("CtrlImax", doubleToString((double)_config.Imax * 0.025));
      event->setProperty("CtrlYmin", intToString(_config.Ymin - 100));
      event->setProperty("CtrlYmax", intToString(_config.Ymax - 100));
      event->setProperty("CtrlAntiWindUp", (_config.AntiWindUp > 0) ? "true" : "false");
      event->setProperty("CtrlKeepFloorWarm", (_config.KeepFloorWarm > 0) ? "true" : "false");
      break;
    case HeatingControlMode::ZONE_FOLLOWER:
      event->setProperty("ReferenceZone", intToString(_config.SourceZoneId));
      event->setProperty("CtrlOffset", intToString(_config.Offset));
      break;
    case HeatingControlMode::FIXED:
      break;
    case HeatingControlMode::MANUAL:
      event->setProperty("ManualValue", intToString(_config.ManualValue - 100));
      break;
  }
  return event;
}

std::shared_ptr<Event> createHeatingControllerValue(int _zoneID, const dsuid_t& _ctrlDsuid,
    const TemperatureConfig& _properties, const ZoneHeatingOperationModeSpec_t& _mode) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::HeatingControllerValue);

  event->setProperty("ZoneID", intToString(_zoneID));
  event->setProperty("ControlDSUID", dsuid2str(_ctrlDsuid));
  switch (_properties.m_mode) {
    case HeatingControlMode::OFF:
      break;
    case HeatingControlMode::PID:
      event->setProperty("NominalTemperature_Off",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[0])));
      event->setProperty("NominalTemperature_Comfort",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[1])));
      event->setProperty("NominalTemperature_Economy",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[2])));
      event->setProperty("NominalTemperature_NotUsed",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[3])));
      event->setProperty("NominalTemperature_Night",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[4])));
      event->setProperty("NominalTemperature_Holiday",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[5])));
      event->setProperty("NominalTemperature_Cooling",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[6])));
      event->setProperty("NominalTemperature_CoolingOff",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[7])));
      event->setProperty("NominalTemperature_CoolingEconomy",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[8])));
      event->setProperty("NominalTemperature_CoolingNotUsed",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[9])));
      event->setProperty("NominalTemperature_CoolingNight",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[10])));
      event->setProperty("NominalTemperature_CoolingHoliday",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureSetpoint, _mode.opModes[11])));
      break;
    case HeatingControlMode::ZONE_FOLLOWER:
      break;
    case HeatingControlMode::FIXED:
      event->setProperty("ControlValue_Off",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[0])));
      event->setProperty("ControlValue_Comfort",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[1])));
      event->setProperty("ControlValue_Economy",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[2])));
      event->setProperty("ControlValue_NotUsed",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[3])));
      event->setProperty("ControlValue_Night",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[4])));
      event->setProperty("ControlValue_Holiday",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[5])));
      event->setProperty("ControlValue_Cooling",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[6])));
      event->setProperty("ControlValue_CoolingOff",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[7])));
      event->setProperty("ControlValue_CoolingEconomy",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[8])));
      event->setProperty("ControlValue_CoolingNotUsed",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[9])));
      event->setProperty("ControlValue_CoolingNight",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[10])));
      event->setProperty("ControlValue_CoolingHoliday",
          doubleToString(sensorValueToDouble(SensorType::RoomTemperatureControlVariable, _mode.opModes[11])));
      break;
    case HeatingControlMode::MANUAL:
      break;
  }
  return event;
}

std::shared_ptr<Event>
  createHeatingControllerState(int _zoneID, const dsuid_t &_ctrlDsuid, int _ctrlState)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::HeatingControllerState);

  event->setProperty("ZoneID", intToString(_zoneID));
  event->setProperty("ControlDSUID", dsuid2str(_ctrlDsuid));
  event->setProperty("ControlState", intToString(_ctrlState));
  return event;
}

std::shared_ptr<Event>
  createOldStateChange(const std::string &_scriptId, const std::string &_name,
                       const std::string &_value, CallOrigin _origin)
{
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::OldStateChange);

  event->setProperty("scriptID", _scriptId);
  event->setProperty("statename", _name);
  event->setProperty("state", _value);
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event> createGenericSignalSunshine(const uint8_t &_value,
    const CardinalDirection_t &_direction, CallOrigin _origin) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::Sunshine);

  // value: {active=1, inactive=2}
  DS_ASSERT(_value == 1 || _value == 2, _value);
  DS_ASSERT(ds::isValidEnum(_origin), static_cast<int>(_origin));
  DS_ASSERT(valid(_direction), _direction);
  event->setProperty("value", intToString(_value));
  event->setProperty("direction", toString(_direction));
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event>
  createGenericSignalFrostProtection(const uint8_t &_value, CallOrigin _origin) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::FrostProtection);

  // value: {active=1, inactive=2}
  DS_ASSERT(_value == 1 || _value == 2, _value);
  DS_ASSERT(ds::isValidEnum(_origin), static_cast<int>(_origin));
  event->setProperty("value", intToString(_value));
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event>
  createGenericSignalHeatingModeSwitch(HeatingModeSwitchValue value, CallOrigin _origin) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::HeatingModeSwitch);

  DS_ASSERT(ds::isValidEnum(_origin), static_cast<int>(_origin));
  event->setProperty("value", intToString(static_cast<int>(value)));
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event>
  createGenericSignalBuildingService(const uint8_t &_value, CallOrigin _origin) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::BuildingService);

  // value: {active=1, inactive=2}
  DS_ASSERT(_value == 1 || _value == 2, _value);
  DS_ASSERT(ds::isValidEnum(_origin), static_cast<int>(_origin));
  event->setProperty("value", intToString(_value));
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event>
createOperationLockEvent(std::shared_ptr<Group> _group, const int _zoneID, const int _groupID,
    const bool _lock, CallOrigin _origin) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::OperationLock, _group);
  event->setProperty("zoneID", intToString(_zoneID));
  event->setProperty("groupID", intToString(_groupID));
  event->setProperty("lock", intToString(_lock));
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event>
createClusterConfigLockEvent(std::shared_ptr<Group> _group, const int _zoneID, const int _groupID,
    const bool _lock, CallOrigin _origin) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::ClusterConfigLock, _group);
  event->setProperty("zoneID", intToString(_zoneID));
  event->setProperty("groupID", intToString(_groupID));
  event->setProperty("lock", intToString(_lock));
  event->setProperty(ef_callOrigin, _origin);
  return event;
}

std::shared_ptr<Event>
createClusterCheckWindEvent(const int _clusterID, const int _time) {
  std::shared_ptr<Event> event;
  event = std::make_shared<Event>(EventName::ClusterCheckWind);
  event->setProperty(EventProperty::Time, "+" + intToString(_time));
  event->setProperty("clusterId", intToString(_clusterID));
  return event;
}

}
