/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <chrono>
#include <cmath>

#include "base.h"
#include "dss.h"
#include "hal/hal-device.h"
#include "model/apartment.h"
#include "model/device.h"
#include "util.h"

DS_REFLECT_DEFINE_STRUCT((dss), DevicePropertyId);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceSceneId);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceScenePropertyId);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceSceneChannelId);
DS_REFLECT_DEFINE_STRUCT((dss), DeviceSceneChannelPropertyId);

DS_JSON_DEFINE_STRUCT((dss)(hal), DeviceInternalState);
DS_JSON_DEFINE_STRUCT((dss)(hal), DeviceInternalOutputState);
DS_JSON_DEFINE_STRUCT((dss)(hal), DeviceChannelValues);
DS_JSON_DEFINE_STRUCT((dss)(hal), DeviceState);
DS_JSON_DEFINE_STRUCT((dss)(hal), DeviceCacheState);

using namespace dss::hal;

namespace dss {

static ds::LogChannel dsLogChannel("halHardwareAbstraction");

std::string toPath(const DevicePropertyId& id) {
  return std::string("/property/") + ds::fromEnum<std::string>(id.property); // dsuid not part of path
}

std::string toPath(const DeviceSceneId& id) {
  return std::string("/scene/") + std::to_string(id.sceneId);
}

std::string toPath(const DeviceScenePropertyId& id) {
  return toPath(id.scene) + "/" + ds::fromEnum<std::string>(id.property);
}

std::string toPath(const struct DeviceSceneChannelId& id) {
  return toPath(id.scene) + "/" + ds::fromEnum<std::string>(id.type);
}

std::string toPath(const struct DeviceSceneChannelPropertyId& id) {
  return toPath(id.channel) + "/" + ds::fromEnum<std::string>(id.property);
}

struct HardwareAbstractionImpl : boost::noncopyable, public HardwareAbstraction {
  HardwareAbstractionImpl(Apartment* apt, ds::sqlite3::Database* db);

  void write(const Dsuid& dsuid, const std::string& parameter, double value) const override;
  void write(const Dsuid& dsuid, const std::string& parameter, bool value) const override;
  void write(const Dsuid& dsuid, const std::string& parameter, int value) const override;
  void write(const Dsuid& dsuid, const std::string& parameter, const std::string& value) const override;
  void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const override;

  double readDouble(const Dsuid& dsuid, const std::string& parameter) const override;
  bool readBool(const DevicePropertyId& id) const override;
  bool readBool(const DeviceScenePropertyId& id) const override;
  bool readBool(const DeviceSceneChannelPropertyId& id) const override;
  std::string readString(const Dsuid& dsuid, const std::string& parameter) const override;

  bool callScene(const Dsuid& dsuid, int sceneId, bool forced, ds::DateTime now) const override;
  bool undoScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const override;
  void saveScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const override;
  bool setOutputValue(const Dsuid& dsuid, uint8_t value, ds::DateTime now) const override;
  bool setOutputValue(const Dsuid& dsuid, ChannelType channel, double value, ds::DateTime now) const override;
  bool setSceneLocalPrio(const Dsuid& dsuid, int sceneId) const override;
  bool callSceneMin(const Dsuid& dsuid, int sceneId, ds::DateTime now) const override;
  void setOperationLock(const Dsuid& device, bool lock) const override;
  bool stopDimming(const Dsuid& dsuid, int sceneId, ds::DateTime now) const override;
  bool completeMoving(const Dsuid& dsuid, ds::DateTime now) const override;

  hal::DeviceState getState(const Dsuid& dsuid) const override;
  dsModel::ZoneApplicationsStatus getZoneState(int zoneId) const override;

  void populateCache(const Dsuid& dsuid) const override; // TODO unit tests probably fails
  hal::DeviceCacheState cacheStatus(const Dsuid& dsuid) const override;

  bool deviceCacheable(const Dsuid& dsuid) const override;
  CacheDb<DeviceCache>& getCache() const override { return m_sceneCacheDb; }
  CacheDb<hal::DeviceInternalState>& getStateCache() const override { return m_stateCache; }

  std::vector<int> areaMember(const Dsuid& dsuid) const override;

  std::vector<double> getDeviceThresholds(const Dsuid& dsuid) const override;
  std::vector<int> getPlayerScenarios(int zoneId) const override;

  void publishToPropertyTree(const Dsuid& dsuid) const override; // ugly hack in dss::Device
  bool supportedBasicFunction(const Device& device) const;
  bool supportedDimmingFunction(const Device& device) const;

  const hal::HalDevice& getHalDevice(const Device& device) const;

  mutable CacheDb<DeviceCache> m_sceneCacheDb;
  mutable CacheDb<hal::DeviceInternalState> m_stateCache;
  Apartment* m_apartment;
  hal::HalLightDevice m_halLight;
  hal::HalShadeDevice m_halShade;
  hal::HalJokerDevice m_halJoker;
  hal::HalAudioDevice m_halAudio;
  hal::HalClimateDevice m_halClimate;
};

HardwareAbstractionImpl::HardwareAbstractionImpl(Apartment* apt, ds::sqlite3::Database* db)
    : m_sceneCacheDb(db),
      m_stateCache(db),
      m_apartment(apt),
      m_halLight(m_sceneCacheDb, m_stateCache),
      m_halShade(m_sceneCacheDb, m_stateCache),
      m_halJoker(m_sceneCacheDb, m_stateCache),
      m_halAudio(m_sceneCacheDb, m_stateCache, apt),
      m_halClimate(m_sceneCacheDb, m_stateCache) {}

const HalDevice& HardwareAbstractionImpl::getHalDevice(const Device& device) const {
  switch (device.getDeviceClass()) {
    case DeviceClasses_t::DEVICE_CLASS_GE: return m_halLight;
    case DeviceClasses_t::DEVICE_CLASS_GR: return m_halShade;
    case DeviceClasses_t::DEVICE_CLASS_SW: return m_halJoker;
    case DeviceClasses_t::DEVICE_CLASS_BL: return m_halClimate;
    case DeviceClasses_t::DEVICE_CLASS_TK:
    case DeviceClasses_t::DEVICE_CLASS_MG:
    case DeviceClasses_t::DEVICE_CLASS_RT:
    case DeviceClasses_t::DEVICE_CLASS_GN:
    case DeviceClasses_t::DEVICE_CLASS_WE:
    case DeviceClasses_t::DEVICE_CLASS_INVALID: DS_FAIL_REQUIRE("unhandled", static_cast<int>(device.getDeviceClass()));
  }
  DS_FAIL_REQUIRE("unhandled", static_cast<int>(device.getDeviceClass()));
}

bool HardwareAbstractionImpl::deviceCacheable(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if ((device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE && device->hasOutput()) ||
      (device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR && device->hasExtendendSceneTable() && device->hasOutput()) ||
      (device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW && device->hasOutput()) ||
      (device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL && device->isBLKM300()) ||
      (device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_MG)) {
    return true;
  }
  return false;
}

bool HardwareAbstractionImpl::supportedBasicFunction(const Device& device) const {
  return ((device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE && device.hasOutput()) ||
          (device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR && device.hasExtendendSceneTable() && device.hasOutput()) ||
          (device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW && device.hasOutput()) ||
          (device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL && device.isBLKM300()));
}

bool HardwareAbstractionImpl::supportedDimmingFunction(const Device& device) const {
  return ((device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GE && device.hasOutput()) ||
          (device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_GR && device.hasExtendendSceneTable() && device.hasOutput()) ||
          (device.getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW && device.hasOutput()));
}

void HardwareAbstractionImpl::saveScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).saveScene(*device, sceneId, now);
}

void HardwareAbstractionImpl::write(const Dsuid& dsuid, const std::string& parameter, double value) const {
  DS_INFO("Writing double to device", dsuid, parameter, value);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).write(*device, parameter, value);
}

void HardwareAbstractionImpl::write(const Dsuid& dsuid, const std::string& parameter, bool value) const {
  DS_INFO("Writing bool to device", dsuid, parameter, value);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).write(*device, parameter, value);
}

void HardwareAbstractionImpl::write(const Dsuid& dsuid, const std::string& parameter, const std::string& value) const {
  DS_INFO("Writing string to device", dsuid, parameter, value);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).write(*device, parameter, value);
}

void HardwareAbstractionImpl::write(const Dsuid& dsuid, const std::string& parameter, int value) const {
  return write(dsuid, parameter, static_cast<double>(value));
}

double HardwareAbstractionImpl::readDouble(const Dsuid& dsuid, const std::string& parameter) const {
  DS_INFO("Reading double from device", dsuid, parameter);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).readDouble(*device, parameter);
}

bool HardwareAbstractionImpl::readBool(const DeviceScenePropertyId& id) const {
  DS_INFO("readBool", id);
  auto device = m_apartment->getDeviceByDSID(id.scene.dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", id);
  return getHalDevice(*device).readBool(*device, toPath(id));
}

bool HardwareAbstractionImpl::readBool(const DevicePropertyId& id) const {
  DS_INFO("readBool", id);
  auto device = m_apartment->getDeviceByDSID(id.dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", id);
  return getHalDevice(*device).readBool(*device, toPath(id));
}

bool HardwareAbstractionImpl::readBool(const DeviceSceneChannelPropertyId& id) const {
  DS_INFO("readBool", id);
  auto device = m_apartment->getDeviceByDSID(id.channel.scene.dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", id);
  return getHalDevice(*device).readBool(*device, toPath(id));
}

std::string HardwareAbstractionImpl::readString(const Dsuid& dsuid, const std::string& parameter) const {
  DS_INFO("Reading string from device", dsuid, parameter);
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).readString(*device, parameter);
}

void HardwareAbstractionImpl::updateDeviceConfig(const Dsuid& dsuid, int cls, int index, int value) const {
  DS_INFO("Updating device config", dsuid, cls, index, value);
  std::shared_ptr<Device> device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    return; // ignore call
  }
  return getHalDevice(*device).updateDeviceConfig(*device, cls, index, value);
}

void HardwareAbstractionImpl::populateCache(const Dsuid& dsuid) const {
  DS_INFO("populateCache", dsuid);
  std::shared_ptr<Device> device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).populateCache(*device);
}

DeviceCacheState HardwareAbstractionImpl::cacheStatus(const Dsuid& dsuid) const {
  DS_INFO("cacheStatus", dsuid);
  std::shared_ptr<Device> device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).cacheStatus(*device);
}

bool HardwareAbstractionImpl::callScene(const Dsuid& dsuid, int sceneId, bool forced, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    return false; // no change without implementation
  }
  return getHalDevice(*device).callScene(*device, sceneId, forced, now);
}

bool HardwareAbstractionImpl::undoScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    return false; // no change without implementation
  }
  return getHalDevice(*device).undoScene(*device, sceneId, now);
}

bool HardwareAbstractionImpl::setOutputValue(const Dsuid& dsuid, uint8_t value, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    return false; // no change without implementation
  }
  return getHalDevice(*device).setOutputValue(*device, value, now);
}

bool HardwareAbstractionImpl::setOutputValue(const Dsuid& dsuid, ChannelType channel, double value,
                                             ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    return false; // no change without implementation
  }
  return getHalDevice(*device).setOutputValue(*device, channel, value, now);
}

bool HardwareAbstractionImpl::setSceneLocalPrio(const Dsuid& dsuid, int sceneId) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedDimmingFunction(*device)) {
    return false; // no change without implementation
  }
  return getHalDevice(*device).setSceneLocalPrio(*device, sceneId);
}

bool HardwareAbstractionImpl::callSceneMin(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedDimmingFunction(*device)) {
    return false; // no change without implementation
  }
  return getHalDevice(*device).callSceneMin(*device, sceneId, now);
}

DeviceState HardwareAbstractionImpl::getState(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    return {}; // empty without implementation
  }
  return getHalDevice(*device).getState(*device);
}

dsModel::ZoneApplicationsStatus HardwareAbstractionImpl::getZoneState(int zoneId) const {
  dsModel::ZoneApplicationsStatus zoneState;
  zoneState.audio = m_halAudio.getZoneState(zoneId);
  return zoneState;
}

void HardwareAbstractionImpl::publishToPropertyTree(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedBasicFunction(*device)) {
    DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  }
  return getHalDevice(*device).publishToPropertyTree(*device);
}

void HardwareAbstractionImpl::setOperationLock(const Dsuid& dsuid, bool lock) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedDimmingFunction(*device)) {
    DS_FAIL_REQUIRE("No HAL implementation for this type of device", dsuid);
  }
  return getHalDevice(*device).setOperationLock(*device, lock);
}

std::vector<int> HardwareAbstractionImpl::areaMember(const Dsuid& dsuid) const {
  std::vector<int> ret;
  DeviceCache cacheEntry = m_sceneCacheDb.tryLookup(dsuid);
  // device is in area if don't care flag for that area is not set
  for (auto i : {SceneA11, SceneA21, SceneA31, SceneA41}) {
    if (cacheEntry.scene.sceneTable[i].flags.dontCare && (*cacheEntry.scene.sceneTable[i].flags.dontCare == false)) {
      ret.push_back(i);
    }
  }
  return ret;
}

bool HardwareAbstractionImpl::stopDimming(const Dsuid& dsuid, int sceneId, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedDimmingFunction(*device)) {
    return {}; // empty without implementation
  }
  return getHalDevice(*device).stopDimming(*device, sceneId, now);
}

bool HardwareAbstractionImpl::completeMoving(const Dsuid& dsuid, ds::DateTime now) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  if (!supportedDimmingFunction(*device)) {
    return {}; // empty without implementation
  }
  return getHalDevice(*device).completeMoving(*device, now);
}

std::vector<double> HardwareAbstractionImpl::getDeviceThresholds(const Dsuid& dsuid) const {
  auto device = m_apartment->getDeviceByDSID(dsuid);
  DS_REQUIRE(supportedBasicFunction(*device), "No HAL implementation for this type of device", dsuid);
  return getHalDevice(*device).getDeviceThresholds(*device);
}

std::vector<int> HardwareAbstractionImpl::getPlayerScenarios(int zoneId) const {
  return m_halAudio.playerZoneScenarios(zoneId);
}

std::unique_ptr<HardwareAbstraction> makeHardwareAbstraction(Apartment* apt, ds::sqlite3::Database* db) {
  return std::make_unique<HardwareAbstractionImpl>(apt, db);
}

} // namespace dss
