/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#include "hardware-abstraction.h"

#include <trompeloeil.hpp>
#include "businterface-mock.h"
#include "ds/catch/catch.h"
#include "ds/json.h"
#include "ds/log.h"
#include "event-mock.h"
#include "model/apartment.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"

static ds::LogChannel dsLogChannel("lightDeviceMultiChannel", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][lightDeviceMultiChannel]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {


SCENARIO("Color light devices", TAGS) {
  auto pApartment = makeApartment(nullptr);
  auto& apt = *pApartment;
  BusInterfaceMock busInterface;
  EventSinkMock eventSink;
  ModelMaintenanceMock modelMaintenance;

  apt.setBusInterface(&busInterface);
  apt.setEventSink(eventSink);
  apt.setModelMaintenance(&modelMaintenance);

  auto dev1 = createDevice(apt, {
                                    .hardwareType = HardwareType::GE_KM200,
                                });
  auto dev2 = createDevice(apt, {
                                    .hardwareType = HardwareType::GE_KM200,
                                });
  dev2->setOutputChannels({ChannelType::brightness, ChannelType::hue, ChannelType::saturation, ChannelType::colortemp,
                           ChannelType::x, ChannelType::y});

  ds::sqlite3::Database d("cache.db");
  auto halPtr = makeHardwareAbstraction(&apt, &d);
  auto& hal = *halPtr;
  hal.getCache().empty();

  WHEN("a not implemented device") {
    THEN("hal access fails") {
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/hue", 50.0));
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/saturation", 50.0));
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/colortemp", 50.0));
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/x", 50.0));
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/y", 50.0));
    }
  }

  WHEN("writing a value") {
    THEN("lastChanged is updated") {
      ALLOW_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, _, _, _, _));
      ALLOW_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, _, _)).RETURN(0xAF);

      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/hue", 50.0));
      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.lastChanged);
      REQUIRE(devCache.scene.sceneTable[5].lastChanged);
      REQUIRE(*devCache.scene.sceneTable[5].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/6/saturation", 50.0));
      devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.lastChanged);
      REQUIRE(devCache.scene.sceneTable[6].lastChanged);
      REQUIRE(*devCache.scene.sceneTable[6].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/7/colortemp", 150.0));
      devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.lastChanged);
      REQUIRE(devCache.scene.sceneTable[7].lastChanged);
      REQUIRE(*devCache.scene.sceneTable[7].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/8/x", 0.500));
      devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.lastChanged);
      REQUIRE(devCache.scene.sceneTable[8].lastChanged);
      REQUIRE(*devCache.scene.sceneTable[8].lastChanged == *devCache.lastChanged);

      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/9/y", 0.500));
      devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.lastChanged);
      REQUIRE(devCache.scene.sceneTable[9].lastChanged);
      REQUIRE(*devCache.scene.sceneTable[9].lastChanged == *devCache.lastChanged);
    }
  }

  WHEN("Writing a hue value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 1, 8, 5, 35));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/hue", 50.0));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::hue].value == 50.0);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 1, 8, 5, 0));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/hue", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 1, 8, 5, 255));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/hue", 360));
    }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/hue", 360.1)); }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/hue", -1)); }
    THEN("fails if device does not have channel") { REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/hue", 0)); }
  }

  WHEN("reading hue channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 1, 5)).RETURN(158);
      REQUIRE(roundDigits(hal.readDouble(dev2->getDSID(), "/scene/5/hue"), 1) == 223.1);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::hue].value, 1) ==
              223.1);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 1, 5)).RETURN(0);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/hue") == 0);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::hue].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 1, 5)).RETURN(255);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/hue") == 360);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::hue].value == 360);
    }
    THEN("fails if device does not have relevant channel") {
      REQUIRE_THROWS(hal.readDouble(dev1->getDSID(), "/scene/5/hue"));
    }
  }

  WHEN("Writing a saturation value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 2, 8, 5, 128));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/saturation", 50.0));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::saturation].value == 50.0);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 2, 8, 5, 0));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/saturation", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 2, 8, 5, 255));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/saturation", 100));
    }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/saturation", 100.1)); }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/saturation", -1)); }
    THEN("fails if device does not have channel") {
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/saturation", 0));
    }
  }

  WHEN("reading saturation channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 2, 5)).RETURN(136);
      REQUIRE(roundDigits(hal.readDouble(dev2->getDSID(), "/scene/5/saturation"), 1) == 53.3);
      REQUIRE(
          roundDigits(
              *hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::saturation].value,
              1) == 53.3);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 2, 5)).RETURN(0);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/saturation") == 0);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::saturation].value ==
              0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 2, 5)).RETURN(255);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/saturation") == 100);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::saturation].value ==
              100);
    }
    THEN("fails if device does not have relevant channel") {
      REQUIRE_THROWS(hal.readDouble(dev1->getDSID(), "/scene/5/saturation"));
    }
  }

  WHEN("Writing a colortemp value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 3, 8, 5, 109));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/colortemp", 485.0));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::colortemp].value == 485.0);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 3, 8, 5, 0));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/colortemp", 100));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 3, 8, 5, 255));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/colortemp", 1000));
    }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/colortemp", 1000.1)); }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/colortemp", 99)); }
    THEN("fails if device does not have channel") {
      REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/colortemp", 0));
    }
  }

  WHEN("reading colortemp channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 3, 5)).RETURN(78);
      REQUIRE(roundDigits(hal.readDouble(dev2->getDSID(), "/scene/5/colortemp"), 1) == 375.3);
      REQUIRE(roundDigits(
                  *hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::colortemp].value,
                  1) == 375.3);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 3, 5)).RETURN(0);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/colortemp") == 100);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::colortemp].value ==
              100);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 3, 5)).RETURN(255);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/colortemp") == 1000);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::colortemp].value ==
              1000);
    }
    THEN("fails if device does not have relevant channel") {
      REQUIRE_THROWS(hal.readDouble(dev1->getDSID(), "/scene/5/colortemp"));
    }
  }

  WHEN("Writing a x value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 4, 8, 5, 128));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/x", 0.5));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::x].value == 0.5);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 4, 8, 5, 0));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/x", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 4, 8, 5, 255));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/x", 1));
    }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/x", 1.1)); }
    WHEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/x", -1)); }
    THEN("fails if device does not have channel") { REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/x", 0)); }
  }

  WHEN("reading x channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 4, 5)).RETURN(136);
      REQUIRE(roundDigits(hal.readDouble(dev2->getDSID(), "/scene/5/x"), 3) == 0.533);
      REQUIRE(roundDigits(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::x].value,
                          3) == 0.533);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 4, 5)).RETURN(0);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/x") == 0);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::x].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 4, 5)).RETURN(255);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/x") == 1);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::x].value == 1);
    }
    THEN("fails if device does not have relevant channel") {
      REQUIRE_THROWS(hal.readDouble(dev1->getDSID(), "/scene/5/x"));
    }
  }

  WHEN("Writing a y value") {
    THEN("it writes to correct bank/offset") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 5, 8, 5, 128));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/y", 0.500));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::y].value == 0.500);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 5, 8, 5, 0));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/y", 0));
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 5, 8, 5, 255));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/y", 1));
    }
    THEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/y", 1.1)); }
    THEN("Writing a illegal value fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/y", -1)); }
    THEN("fails if device does not have channel") { REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/y", 0)); }
  }

  WHEN("reading y channel") {
    THEN("device is read") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 5, 5)).RETURN(136);
      REQUIRE(roundDigits(hal.readDouble(dev2->getDSID(), "/scene/5/y"), 3) == 0.533);
      REQUIRE(roundDigits(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::y].value,
                          3) == 0.533);
    }
    THEN("minimum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 5, 5)).RETURN(0);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/y") == 0);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::y].value == 0);
    }
    THEN("maximum") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneValue(_, 5, 5)).RETURN(255);
      REQUIRE(hal.readDouble(dev2->getDSID(), "/scene/5/y") == 1);
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::y].value == 1);
    }
    THEN("fails if device does not have relevant channel") {
      REQUIRE_THROWS(hal.readDouble(dev1->getDSID(), "/scene/5/y"));
    }
  }

  WHEN("Writing flags") {
    THEN("that are not cached yet requires read-modify-write") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x50);
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x51));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("that are not cached yet requires read-modify-write") {
      REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x51);
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x21));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/ledConfig", std::string("ledCon1")));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon1);
    }
  }

  WHEN("Writing flags with cache") {
    REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x150);
    REQUIRE(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);

    THEN("can be written directly") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x151));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/dontCare", true));

      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(*devCache.scene.sceneTable[5].flags.dontCare == true);
    }
    THEN("flags need to be boolean") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/dontCare", 25.0)); }
    THEN("Write ignoreLocalPrio") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x152));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/ignoreLocalPrio", true));
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.ignoreLocalPrio == true);
    }
    THEN("Write specialHandling") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x154));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/specialHandling", true));
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.specialHandling == true);
    }
    THEN("Write blink") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x150));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/blink", true));
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.blink == true);

      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x050));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/blink", false));
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.blink == false);
    }
    THEN("Write blink as number fails") { REQUIRE_THROWS(hal.write(dev1->getDSID(), "/scene/5/blink", 3.5)); }
    THEN("Write ledConfig") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneConfig(_, 5, 0x120));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/ledConfig", std::string("ledCon1")));
      REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon1);
    }
    THEN("Write ledConfig as bool fails") { REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/ledConfig", true)); }
    THEN("Write ledConfig out of range") {
      REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/ledConfig", std::string("ledCon8")));
      REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/ledConfig", std::string("1")));
    }
    THEN("Write dimTime fails") {
      REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/dimTime", std::string("dimTime1")));
      REQUIRE_THROWS(hal.write(dev2->getDSID(), "/scene/5/dimTime", true));
    }

    WHEN("Reading flags") {
      THEN("can be read directly") {
        REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x150);
        REQUIRE(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::dontCare}) == false);
        REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.dontCare == false);
      }
      THEN("flags need to be boolean") { REQUIRE_THROWS(hal.readDouble(dev2->getDSID(), "/scene/5/dontCare") == 25.0); }
      THEN("Read ignoreLocalPrio") {
        REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x152);
        REQUIRE(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::ignoreLocalPrio}) == true);
        REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.ignoreLocalPrio == true);
      }
      THEN("Read specialHandling") {
        REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x156);
        REQUIRE(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::specialHandling}) == true);
        REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.specialHandling == true);
      }
      THEN("Read blink") {
        REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x156);
        REQUIRE(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::blink}) == true);
        REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.blink == true);

        REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x056);
        REQUIRE(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::blink}) == false);
        REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.blink == false);
      }
      THEN("Read blink as number fails") { REQUIRE_THROWS(hal.readDouble(dev1->getDSID(), "/scene/5/blink") == 3.5); }
      THEN("Read ledConfig") {
        REQUIRE_CALL(busInterface.device, getDeviceOutputChannelSceneConfig(_, 5)).RETURN(0x1BE);
        REQUIRE(hal.readString(dev2->getDSID(), "/scene/5/ledConfig") == "ledCon2");
        REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].flags.ledConfig == LedConfig::ledCon2);
      }
      THEN("Read ledConfig as bool fails") {
        REQUIRE_THROWS(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::ledConfig}) == true);
      }
      THEN("Read dimTime") {
        REQUIRE_THROWS(hal.readString(dev2->getDSID(), "/scene/5/dimTime") == "dimTime10");
        REQUIRE_THROWS(hal.readBool({{.dsuid = dev2->getDSID(), .sceneId = 5}, SceneProperty::dimTime}) == true);
      }
    }
  }
  WHEN("Reading channel dontCare flag") {
    REQUIRE_CALL(busInterface.device, getDeviceOutputChannelDontCareFlags(_, 5)).RETURN(0x3F);
    REQUIRE(hal.readBool({{{.dsuid = dev2->getDSID(), .sceneId = 5}, ChannelType::brightness},
                          ChannelProperty::dontCare}) == true);
    REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::brightness].dontCare ==
            true);
    REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::hue].dontCare == true);
    REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::saturation].dontCare ==
            true);
    REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::x].dontCare == true);
    REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::y].dontCare == true);
  }
  WHEN("Writing channel dontCare flag") {
    REQUIRE_CALL(busInterface.device, getDeviceOutputChannelDontCareFlags(_, 5)).RETURN(0x3F);
    REQUIRE_CALL(busInterface.device, setDeviceOutputChannelDontCareFlags(_, 5, 0x3E));
    REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/brightness/dontCare", false));
    REQUIRE(*hal.getCache().tryLookup(dev2->getDSID()).scene.sceneTable[5].channels[ChannelType::brightness].dontCare ==
            false);
  }

  WHEN("device has hue/saturation channels set") {
    REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 1, 8, 5, _));
    REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/hue", 50.0));
    REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 2, 8, 5, _));
    REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/saturation", 50.0));
    auto devCache = hal.getCache().tryLookup(dev2->getDSID());
    REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::hue].value == 50.0);
    REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::saturation].value == 50.0);

    THEN("writing colortemp channel invalidates hue/saturation") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 3, 8, 5, _));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/colortemp", 150));
      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.scene.sceneTable[5].channels[ChannelType::hue].value.has_value() == false);
      REQUIRE(devCache.scene.sceneTable[5].channels[ChannelType::saturation].value.has_value() == false);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::colortemp].value == 150.0);
    }

    THEN("writing x channel invalidates hue/saturation") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 4, 8, 5, _));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/x", 0.5));
      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.scene.sceneTable[5].channels[ChannelType::hue].value.has_value() == false);
      REQUIRE(devCache.scene.sceneTable[5].channels[ChannelType::saturation].value.has_value() == false);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::x].value == 0.5);
    }

    THEN("writing y channel invalidates hue/saturation") {
      REQUIRE_CALL(busInterface.device, setDeviceOutputChannelSceneValue(_, 5, 8, 5, _));
      REQUIRE_NOTHROW(hal.write(dev2->getDSID(), "/scene/5/y", 0.5));
      auto devCache = hal.getCache().tryLookup(dev2->getDSID());
      REQUIRE(devCache.scene.sceneTable[5].channels[ChannelType::hue].value.has_value() == false);
      REQUIRE(devCache.scene.sceneTable[5].channels[ChannelType::saturation].value.has_value() == false);
      REQUIRE(*devCache.scene.sceneTable[5].channels[ChannelType::y].value == 0.5);
    }
  }
  WHEN("device has unknown channel types") {
    dev2->setOutputChannels({ChannelType::brightness, ChannelType::hue, ChannelType::saturation, ChannelType::colortemp,
                             ChannelType::x, ChannelType::y, static_cast<ChannelType>(192),
                             static_cast<ChannelType>(193)});
    THEN("the unknown channels are skipped") {
      hal.completeMoving(dev2->getDSID(), std::chrono::system_clock::now());
      auto statusCache = hal.getStateCache().tryLookup(dev2->getDSID());
      REQUIRE(statusCache.channels.size() == 6);
    }
  }
}
} // namespace
