/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "archives.h"

#include <ds/catch/catch.h>
#include <ds/system-mock.h>

#include <trompeloeil.hpp>

#include "ds-model.h"
#include "rrd-interface.h"

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

using system_clock = std::chrono::system_clock;

static ds::LogChannel dsLogChannel("dssMeteringArchivesTest");

namespace {

static const char* TAGS = "[dSS][metering][archives]";

struct RRDInterfaceMock : public trompeloeil::mock_interface<dss::RRDInterface> {
  IMPLEMENT_MOCK4(create);
  IMPLEMENT_MOCK1(info);
  IMPLEMENT_MOCK1(tune);
  IMPLEMENT_MOCK1(update);
  IMPLEMENT_MOCK1(flush);
  IMPLEMENT_MOCK1(lastValue);
  IMPLEMENT_MOCK1(lastTimestamp);
  IMPLEMENT_MOCK1(xport);
};

auto createRRDInfo() {
  return std::vector{RRDInfo{.key = "ds[power].index"s, .type = RrdInfoType::counter, .value = counter_t{0}},
                     RRDInfo{.key = "ds[energy].index"s, .type = RrdInfoType::counter, .value = counter_t{1}},
                     RRDInfo{.key = "rra[0].rows"s, .type = RrdInfoType::counter, .value = counter_t{600}},
                     RRDInfo{.key = "rra[1].rows"s, .type = RrdInfoType::counter, .value = counter_t{720}},
                     RRDInfo{.key = "rra[2].rows"s, .type = RrdInfoType::counter, .value = counter_t{2976}},
                     RRDInfo{.key = "rra[3].rows"s, .type = RrdInfoType::counter, .value = counter_t{370}},
                     RRDInfo{.key = "rra[4].rows"s, .type = RrdInfoType::counter, .value = counter_t{260}},
                     RRDInfo{.key = "rra[5].rows"s, .type = RrdInfoType::counter, .value = counter_t{60}}};
}

SCENARIO("file manipulation", TAGS) {
  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto rrdInterface = std::make_shared<RRDInterfaceMock>();
  auto system = ds::SystemMock();
  auto archives = makeArchives(rrdInterface, system);

  SECTION("filename") {
    CHECK(archives->filename({Dsuid{controllerId1}, Category::consumption}) ==
          "1111111111111111111111111111111100.rrd"s);
    CHECK(archives->filename({Dsuid{controllerId1}, Category::production}) ==
          "1111111111111111111111111111111100-production.rrd"s);
  }
  SECTION("hasArchive") {
    GIVEN("archive exists") {
      REQUIRE_CALL(system, existsFile(_)).RETURN(true);
      CHECK(archives->hasArchive({Dsuid{controllerId1}, Category::consumption}));
    }
    GIVEN("archive does not exists") {
      REQUIRE_CALL(system, existsFile(_)).RETURN(false);
      CHECK(!archives->hasArchive({Dsuid{controllerId1}, Category::consumption}));
    }
  }
}

SCENARIO("rrd-interface", TAGS) {
  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto rrdInterface = std::make_shared<RRDInterfaceMock>();
  auto system = ds::SystemMock();
  auto archives = makeArchives(rrdInterface, system);

  auto toFilename = [archives, controllerId1](Category category) {
    return archives->filename({Dsuid{controllerId1}, category});
  };

  WHEN("create consumption") {
    REQUIRE_CALL(*rrdInterface, create(_, 1, _,
                                       std::vector{"DS:power:GAUGE:5:0:8000"s, "DS:energy:DERIVE:5:0:U"s,
                                                   "RRA:AVERAGE:0.5:1:600"s, "RRA:AVERAGE:0.5:60:720"s,
                                                   "RRA:AVERAGE:0.5:900:2976"s, "RRA:AVERAGE:0.5:86400:370"s,
                                                   "RRA:AVERAGE:0.5:604800:260"s, "RRA:AVERAGE:0.5:2592000:60"s}));
    archives->create({Dsuid{controllerId1}, Category::consumption});
  }

  WHEN("create production") {
    REQUIRE_CALL(*rrdInterface, create(_, 1, _,
                                       std::vector{"DS:power:GAUGE:5:0:50000"s, "DS:energy:DERIVE:5:0:U"s,
                                                   "RRA:AVERAGE:0.5:1:600"s, "RRA:AVERAGE:0.5:60:720"s,
                                                   "RRA:AVERAGE:0.5:900:2976"s, "RRA:AVERAGE:0.5:86400:370"s,
                                                   "RRA:AVERAGE:0.5:604800:260"s, "RRA:AVERAGE:0.5:2592000:60"s}));
    archives->create({Dsuid{controllerId1}, Category::production});
  }

  GIVEN("invalid rrd meta") {
    THEN("validateDBSeries fails") {
      REQUIRE_CALL(*rrdInterface, info("1111111111111111111111111111111100.rrd"s)).RETURN(std::vector<RRDInfo>{});
      bool updatePowerSetting = false;
      CHECK(!archives->validateDBSeries(ArchiveId{controllerId1, Category::consumption}, updatePowerSetting));
    }
  }

  GIVEN("rrd meta match spec") {
    THEN("validateDBSeries succeeds") {
      REQUIRE_CALL(*rrdInterface, info("1111111111111111111111111111111100.rrd"s)).RETURN(createRRDInfo());
      bool updatePowerSetting = false;
      CHECK(archives->validateDBSeries(ArchiveId{controllerId1, Category::consumption}, updatePowerSetting));
      CHECK(!updatePowerSetting);
    }
    WHEN("power.max == legacy limit value") {
      auto y = GENERATE(Category::consumption, Category::production);
      auto x = GENERATE(4.5e3, // value used in first implementation, only for consumption
                        4e4    // 36978e9a08, no reason given
      );
      THEN("database power.max needs updating") {
        INFO(ds::str(x));
        REQUIRE_CALL(*rrdInterface, info(toFilename(y))).RETURN([x]() {
          auto info = createRRDInfo();
          info.emplace_back(
              RRDInfo{.key = "ds[power].max"s, .type = RrdInfoType::value, .value = gsl::narrow<rrd_value_t>(x)});
          return info;
        }());

        bool updatePowerSetting = false;
        CHECK(archives->validateDBSeries(ArchiveId{controllerId1, y}, updatePowerSetting));
        CHECK(updatePowerSetting);
      }
    }
    WHEN("power.max is correct") {
      struct X {
        Category category;
        int limit;
      };
      const auto& x = GENERATE(values(std::initializer_list<X>{
          {Category::consumption, 8000},
          {Category::production, 50'000},
      }));

      REQUIRE_CALL(*rrdInterface, info(toFilename(x.category))).RETURN([x]() {
        auto info = createRRDInfo();
        info.emplace_back(
            RRDInfo{.key = "ds[power].max"s, .type = RrdInfoType::value, .value = gsl::narrow<rrd_value_t>(x.limit)});
        return info;
      }());

      bool updatePowerSetting = false;
      CHECK(archives->validateDBSeries(ArchiveId{controllerId1, x.category}, updatePowerSetting));
      CHECK(!updatePowerSetting);
    }
  }

  SECTION("tune max power") {
    WHEN("consumption archive") {
      REQUIRE_CALL(*rrdInterface,
                   tune(std::vector{"tune"s, "1111111111111111111111111111111100.rrd"s, "--maximum"s, "power:8000"s}));
      archives->tuneDBPowerSettings(ArchiveId{controllerId1, Category::consumption});
    }
    WHEN("production archive") {
      REQUIRE_CALL(*rrdInterface, tune(std::vector{"tune"s, "1111111111111111111111111111111100-production.rrd"s,
                                                   "--maximum"s, "power:50000"s}));
      archives->tuneDBPowerSettings(ArchiveId{controllerId1, Category::production});
    }
  }

  GIVEN("rrd updated regularly") {
    THEN("checkDBFlush is no-op") {
      REQUIRE_CALL(*rrdInterface, lastTimestamp("1111111111111111111111111111111100.rrd"s)).RETURN((time_t)10);
      archives->checkDBFlush(::DateTime(time_t(5)), {Dsuid{controllerId1}, Category::consumption});
    }
  }

  WHEN("lastValue") {
    REQUIRE_CALL(*rrdInterface, lastValue(_)).RETURN(std::make_pair(0, std::vector{std::make_pair("energy"s, 0.0)}));
    archives->lastValue({Dsuid{controllerId1}, Category::consumption});
  }

  GIVEN("flush") {
    WHEN("rrdcached path is set") {
      archives->setRRDCachedPath("unix:cache-path"s);
      THEN("flush is executed") {
        REQUIRE_CALL(*rrdInterface, flush(std::vector{"flushcached"s, "--daemon"s, "unix:cache-path"s,
                                                      "1111111111111111111111111111111100.rrd"s}));
        archives->flush(std::vector{ArchiveId{controllerId1, Category::consumption}});
      }
    }
    WHEN("rrdcached path is empty") {
      archives->setRRDCachedPath(""s);
      THEN("flush is no-op") { archives->flush(std::vector{ArchiveId{controllerId1, Category::consumption}}); }
    }
  }

  SECTION("xport per category") {
    auto category = GENERATE(Category::consumption, Category::production);
    WHEN("power") {
      WHEN("single dsm") {
        REQUIRE_CALL(*rrdInterface,
                     xport(std::vector{"xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s,
                                       "3000"s, ds::str("DEF:raw0=", toFilename(category), ":power:AVERAGE"s),
                                       "CDEF:data0=raw0,UN,0,raw0,IF"s, "CDEF:sum=data0"s, "XPORT:sum"s}))
            .RETURN([]() -> ValueRange {
              return {1, 601, 1, std::vector<Value>{600, Value{1, 1}}};
            }());
        auto range = archives->xport(1, 601, 1, {ArchiveId{controllerId1, category}}, SeriesTypes::power, false);
        CHECK(range.values.size() == 600);
      }
      WHEN("2 dsm queried together") {
        REQUIRE_CALL(*rrdInterface,
                     xport(std::vector{"xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s,
                                       "3000"s, ds::str("DEF:raw0=", toFilename(category), ":power:AVERAGE"s),
                                       "CDEF:data0=raw0,UN,0,raw0,IF"s,
                                       ds::str("DEF:raw1=", toFilename(category), ":power:AVERAGE"s),
                                       "CDEF:data1=raw1,UN,0,raw1,IF"s, "CDEF:sum=data0,data1,+"s, "XPORT:sum"s}))
            .RETURN([]() -> ValueRange {
              return {1, 601, 1, std::vector<Value>{600, Value{}}};
            }());

        auto range =
            archives->xport(1, 601, 1, {ArchiveId{controllerId1, category}, ArchiveId{controllerId1, category}},
                            SeriesTypes::power, false);
        CHECK(range.values.size() == 600); // results are not modified
      }
    }
    WHEN("energy") {
      WHEN("single dsm") {
        REQUIRE_CALL(*rrdInterface,
                     xport(std::vector{"xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s,
                                       "3000"s, ds::str("DEF:raw0=", toFilename(category), ":energy:AVERAGE"s),
                                       "CDEF:data0=raw0,UN,0,raw0,IF"s, "CDEF:sum=data0"s,
                                       ds::str("CDEF:ratePerStep=sum,1,*"), "XPORT:ratePerStep"s}))
            .RETURN(ValueRange{});
        auto range = archives->xport(1, 601, 1, {ArchiveId{controllerId1, category}}, SeriesTypes::energy, false);
      }
      WHEN("2 dsm queried together") {
        REQUIRE_CALL(
            *rrdInterface,
            xport(std::vector{
                "xport"s, "--start"s, "1"s, "--end"s, "601"s, "--step"s, "1"s, "--maxrows"s, "3000"s,
                ds::str("DEF:raw0=", toFilename(category), ":energy:AVERAGE"s), "CDEF:data0=raw0,UN,0,raw0,IF"s,
                ds::str("DEF:raw1=", toFilename(category), ":energy:AVERAGE"s), "CDEF:data1=raw1,UN,0,raw1,IF"s,
                "CDEF:sum=data0,data1,+"s, ds::str("CDEF:ratePerStep=sum,1,*"), "XPORT:ratePerStep"s}))
            .RETURN([]() -> ValueRange {
              return {1, 601, 1, std::vector<Value>{600, Value{}}};
            }());

        auto range =
            archives->xport(1, 601, 1, {ArchiveId{controllerId1, category}, ArchiveId{controllerId1, category}},
                            SeriesTypes::energy, false);
        CHECK(range.values.size() == 600); // results are not modified
      }
    }
  }
  SECTION("xport") {
    WHEN("non-fininte floating point returned") {
      const auto nonfinite = GENERATE(std::nan(""), std::numeric_limits<double>::infinity(), -INFINITY);
      REQUIRE_CALL(*rrdInterface, xport(_)).RETURN(ValueRange{1, 61, 60, std::vector<Value>{{nonfinite, 60}}});
      auto timeseries =
          archives->xport(1, 61, 60, {ArchiveId{controllerId1, Category::consumption}}, SeriesTypes::energy, false);
      THEN("values are mapped to 0") {
        CHECK(std::isfinite(timeseries.values[0].value));
        CHECK(timeseries.values[0].value == 0);
      }
    }
  }

  WHEN("update") {
    REQUIRE_CALL(*rrdInterface,
                 update(std::vector{"update"s, "1111111111111111111111111111111100.rrd"s, "100:75:300"s}));
    archives->addValue({Dsuid{controllerId1}, Category::consumption},
                       {.power = 75.0, .energy = 300.0, .lastChanged = ds::DateTime(100s)});

    WHEN("using floating points") {
      THEN("energy values are rounded up (std::ceil)") {
        REQUIRE_CALL(*rrdInterface,
                     update(std::vector{"update"s, "1111111111111111111111111111111100.rrd"s, "100:75.1:301"s}));
        archives->addValue({Dsuid{controllerId1}, Category::consumption},
                           {.power = 75.1, .energy = 300.3, .lastChanged = ds::DateTime(100s)});
      }
    }
    WHEN("when limitPower of the db is exceeded") {
      THEN("value is passed to the db") {
        REQUIRE_CALL(*rrdInterface,
                     update(std::vector{"update"s, "1111111111111111111111111111111100.rrd"s, "100:80000:301"s}));
        archives->addValue({Dsuid{controllerId1}, Category::consumption},
                           {.power = 80000, .energy = 300.3, .lastChanged = ds::DateTime(100s)});
      }
    }
  }
}

SCENARIO("sanity checks", TAGS) {
  auto rrdInterface = std::make_shared<RRDInterfaceMock>();
  auto system = ds::SystemMock();
  auto archives = makeArchives(rrdInterface, system);

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};

  SECTION("files no longer receiving sensor values") {
    auto now = system_clock::now();
    GIVEN("file updates a little behind, but still within limits") {
      REQUIRE_CALL(*rrdInterface, lastTimestamp(_)).RETURN(system_clock::to_time_t(now - 10s));
      CHECK(!archives->withoutUpdates(ArchiveId{controllerId1, Category::consumption}));
    }
    GIVEN("outdated file") {
      // MAX_INTERPOLATION_INTERVAL
      REQUIRE_CALL(*rrdInterface, lastTimestamp(_)).RETURN(system_clock::to_time_t(now - 7 * 24h - 1s));
      CHECK(archives->withoutUpdates({Dsuid{controllerId1}, Category::consumption}));
    }
    GIVEN("lastTimestamp fails") {
      THEN("file is kept") {
        // rely on file format checks
        REQUIRE_CALL(*rrdInterface, lastTimestamp(_)).RETURN(std::nullopt);
        CHECK(!archives->withoutUpdates({Dsuid{controllerId1}, Category::consumption}));
      }
    }
    GIVEN("lastTimestamp is a little in the future") {
      THEN("file is still good") {
        REQUIRE_CALL(*rrdInterface, lastTimestamp(_)).RETURN(system_clock::to_time_t(now + 24h));
        CHECK(!archives->withoutUpdates({Dsuid{controllerId1}, Category::consumption}));
      }
    }
    GIVEN("lastTimestamp is far in the future") {
      THEN("file is still good") {
        REQUIRE_CALL(*rrdInterface, lastTimestamp(_)).RETURN(system_clock::to_time_t(now + 7 * 24h + 1s));
        CHECK(archives->withoutUpdates({Dsuid{controllerId1}, Category::consumption}));
      }
    }

    WHEN("energy is UINT_MAX") {
      THEN("energy is added to db") {
        REQUIRE_CALL(*rrdInterface,
                     update(std::vector{"update"s, "1111111111111111111111111111111100.rrd"s, "100:75.1:4294967295"s}));
        archives->addValue({Dsuid{controllerId1}, Category::consumption},
                           {.power = 75.1, .energy = static_cast<double>(UINT_MAX), .lastChanged = ds::DateTime(100s)});
      }
      WHEN("energy exceeds UINT_MAX + 1") {
        THEN("no limit is applied to power") {
          REQUIRE_CALL(*rrdInterface, update(std::vector{"update"s, "1111111111111111111111111111111100.rrd"s,
                                                         "100:75.1:4294967296"s}));
          archives->addValue(
              {Dsuid{controllerId1}, Category::consumption},
              {.power = 75.1, .energy = static_cast<double>(UINT_MAX) + 1, .lastChanged = ds::DateTime(100s)});
        }
      }
    }
  }
}

} // namespace
