
#include "power-states.h"

#include <ds/catch/catch.h>
#include <metering/metering-mock.h>
#include <model/apartment-mock.h>
#include <model/state.h>

#include <trompeloeil.hpp>

using namespace trompeloeil;
using namespace std::literals;

using namespace dss;
using namespace dss::metering;

static ds::LogChannel dsLogChannel("dssMeteringPowerStatesTest");

namespace {

constexpr auto TAGS = "[dSS][metering]";

SCENARIO("powerStates", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto powerStates = makePowerStates(apartment);
  auto config = PowerStateConfig{.category = Category::production, .setThreshold = 5, .resetThreshold = 7};

  WHEN("createState") {
    powerStates->createState(PowerStateConfig(config));
    THEN("state is created") {
      CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) != nullptr);
    }

    WHEN("removeState") {
      powerStates->removeState(config);
      THEN("state disappears") {
        CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) == nullptr);
      }
    }

    WHEN("2x createState") {
      powerStates->createState(PowerStateConfig(config));

      WHEN("removeState") {
        powerStates->removeState(config);
        THEN("state still exists") {
          CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) != nullptr);
        }
        WHEN("2x removeState") {
          powerStates->removeState(config);
          THEN("state disappears") {
            CHECK(apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s) == nullptr);
          }
        }
      }
    }
  }
}

SCENARIO("pollStates", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto metering = MeteringMock{};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto powerStates = makePowerStates(apartment);

  WHEN("pollState no config") { powerStates->trigger(TotalPower{.production = 100, .consumption = 0}); }

  WHEN("createPowerState production/on:5/off:7") {
    powerStates->createState({.category = Category::production, .setThreshold = 5, .resetThreshold = 7});
    auto state1 = apartment.tryGetState(StateType_Circuit, "apartment-power-production-on5-off7"s);
    CHECK(state1 != nullptr);

    WHEN("state not active") {
      WHEN("production < setThreshold") {
        THEN("state is remains unchanged") {
          DS_NOTICE("start");
          powerStates->trigger(TotalPower{.production = 2, .consumption = 0});
          CHECK(state1->getState() == State_Inactive);
        }
      }

      WHEN("production > setThreshold") {
        THEN("state is active") {
          CHECK(state1->getState() == State_Inactive);
          powerStates->trigger(TotalPower{.production = 10, .consumption = 0});
          CHECK(state1->getState() == State_Active);
        }
      }
    }
  }
}

} // namespace
