#include "power-states.h"

#include <numeric>
#include <tuple>

#include "metering.h"
#include "model/dsmeter.h"
#include "model/state.h"

static ds::LogChannel dsLogChannel("dssMeteringPowerStates");

using namespace dss;
using namespace dss::metering;

namespace {

static auto createStateName = [](const PowerStateConfig& config) {
  return ds::str("apartment-power-", config.category, "-on", config.setThreshold, "-off", config.resetThreshold);
};

class PowerStatesImpl : public PowerStates {
  std::vector<std::tuple<std::shared_ptr<State>, PowerStateConfig, int>> m_states{};
  Apartment& m_apartment;

public:
  PowerStatesImpl(Apartment& apartment) : m_apartment(apartment) {}
  std::shared_ptr<State> createState(PowerStateConfig&& powerStateConfig) override;
  void removeState(const PowerStateConfig& powerStateConfig) override;
  void trigger(TotalPower&& totalPower) override;
};

std::shared_ptr<State> PowerStatesImpl::createState(PowerStateConfig&& powerStateConfig) {
  auto it = std::find_if(std::begin(m_states), std::end(m_states), [config = powerStateConfig](const auto& element) {
    return std::get<PowerStateConfig>(element) == config;
  });
  if (it != std::end(m_states)) {
    auto& [state, config, references] = *it;
    // in clang variables from structured binding are only aliases they are not
    // variables, thus you cannot capture it in DS_REQUIRE lambda by reference
    auto scopeIssue = references++;
    DS_REQUIRE(scopeIssue > 0);
    return state;
  }

  auto state = m_apartment.allocateState(StateType_Circuit, createStateName(powerStateConfig), {});
  m_states.push_back(std::make_tuple(state, std::move(powerStateConfig), 1));
  return state;
}

void PowerStatesImpl::removeState(const PowerStateConfig& powerStateConfig) {
  auto it = std::find_if(std::begin(m_states), std::end(m_states), [config = powerStateConfig](const auto& element) {
    return std::get<PowerStateConfig>(element) == config;
  });
  DS_REQUIRE(it != std::end(m_states));

  auto& [state, config, references] = *it;
  auto scopeIssue = references--;
  DS_REQUIRE(scopeIssue > 0);

  if (references == 0) {
    DS_NOTICE("removeState", config);
    m_states.erase(it);
    m_apartment.removeState(state);
  }
}

void PowerStatesImpl::trigger(TotalPower&& totalPower) {
  DS_DEBUG(totalPower);

  for (auto [state, config, _] : m_states) {
    auto value = state->getState();

    if (value != State_Active) {
      if (totalPower.production > config.setThreshold) {
        state->setState(CallOrigin::system, State_Active);
      }
    } else {
      if (totalPower.production < config.resetThreshold) {
        state->setState(CallOrigin::system, State_Inactive);
      }
    }
  }
}

} // namespace

namespace dss::metering {

std::shared_ptr<PowerStates> makePowerStates(Apartment& apartment) {
  return std::make_shared<PowerStatesImpl>(apartment);
}
} // namespace dss::metering
