/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "rrd-interface.h"

#include <ds/catch/catch.h>
#include <ds/clock-mock.h>

#include <boost/filesystem.hpp>
#include <trompeloeil.hpp>

using namespace trompeloeil;
using namespace std::literals;
using trompeloeil::_;

using namespace dss;
using namespace dss::metering;

static ds::LogChannel dsLogChannel("dssMeteringRrdInterfaceTest");

namespace {

static const char* TAGS = "[dSS][metering]";

SCENARIO("Creating/adding/reading back", TAGS) {
  auto filename = "test.rrd";
  DS_DEFER(boost::filesystem::remove(filename));

  auto rrdInterface = makeRRDInterface();

  SECTION("GAUGE") {
    std::vector<std::string> lines;
    lines.push_back("DS:power:GAUGE:5:0:40000");
    lines.push_back(ds::str("RRA:AVERAGE:0.5:", 60, ":", 720));
    rrdInterface->create(filename, 1, std::chrono::system_clock::to_time_t(ds::DateTime{301s}), lines);

    WHEN("adding float values") {
      std::vector<std::string> lines;
      lines.push_back("update");
      lines.push_back(filename);
      lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{302s}), ":", 1.2));
      rrdInterface->update(lines);

      THEN("float values are read back") {
        auto [lastChanged, data] = rrdInterface->lastValue(filename);
        CHECK(lastChanged == 302);
        CHECK(!data.empty());
        CHECK(data[0] == std::make_pair("power"s, 1.2));
      }
    }
    WHEN("exceeding db limit") {
      CHECK_THROWS(rrdInterface->lastValue(filename));

      std::vector<std::string> lines;
      lines.push_back("update");
      lines.push_back(filename);
      lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{302s}), ":", 800001));
      rrdInterface->update(lines);

      THEN("value is still read back, ingoring upper limit") {
        auto [lastChanged, data] = rrdInterface->lastValue(filename);
        CHECK(lastChanged == 302);
        CHECK(!data.empty());
        CHECK(data[0] == std::make_pair("power"s, 800001.0));
      }
    }
  }
  SECTION("DERIVE") {
    std::vector<std::string> lines;
    lines.push_back("DS:energy:DERIVE:5:0:4000");
    lines.push_back(ds::str("RRA:AVERAGE:0.5:", 60, ":", 720));
    rrdInterface->create(filename, 1, std::chrono::system_clock::to_time_t(ds::DateTime{301s}), lines);

    THEN("adding float values fails") {
      std::vector<std::string> lines;
      lines.push_back("update");
      lines.push_back(filename);
      lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{302s}), ":", 1.2));
      CHECK_THROWS(rrdInterface->update(lines));
    }

    WHEN("adding integer values") {
      std::vector<std::string> lines;
      lines.push_back("update");
      lines.push_back(filename);
      lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{302s}), ":", 7));
      rrdInterface->update(lines);

      THEN("values are read back") {
        auto [lastChanged, data] = rrdInterface->lastValue(filename);
        CHECK(lastChanged == 302);
        CHECK(!data.empty());
        CHECK(data[0] == std::make_pair("energy"s, 7.0));
      }

      WHEN("adding a smaller value") {
        std::vector<std::string> lines;
        lines.push_back("update");
        lines.push_back(filename);
        lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{303s}), ":", 6));
        rrdInterface->update(lines);

        THEN("smaller value is read back") {
          auto [lastChanged, data] = rrdInterface->lastValue(filename);
          CHECK(data[0] == std::make_pair("energy"s, 6.0));
        }
      }
    }
    WHEN("exceeding db limit") {
      CHECK_THROWS(rrdInterface->lastValue(filename));

      std::vector<std::string> lines;
      lines.push_back("update");
      lines.push_back(filename);
      lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{302s}), ":", 800001));
      rrdInterface->update(lines);

      THEN("value is still read back, ingoring upper limit") {
        auto [lastChanged, data] = rrdInterface->lastValue(filename);
        CHECK(lastChanged == 302);
        CHECK(!data.empty());
        CHECK(data[0] == std::make_pair("energy"s, 800001.0));
      }
    }
  }
  WHEN("DDERIVE") {
    std::vector<std::string> lines;
    lines.push_back("DS:energy:DDERIVE:5:0:U");
    lines.push_back(ds::str("RRA:AVERAGE:0.5:", 60, ":", 720));
    rrdInterface->create(filename, 1, std::chrono::system_clock::to_time_t(ds::DateTime{301s}), lines);

    WHEN("adding float values") {
      std::vector<std::string> lines;
      lines.push_back("update");
      lines.push_back(filename);
      lines.push_back(ds::str(std::chrono::system_clock::to_time_t(ds::DateTime{302s}), ":", 1.2));
      rrdInterface->update(lines);

      THEN("float values are read back") {
        auto [lastChanged, data] = rrdInterface->lastValue(filename);
        CHECK(lastChanged == 302);
        CHECK(!data.empty());
        CHECK(data[0] == std::make_pair("energy"s, 1.2));
      }
    }
  }

  SECTION("tune db") {
    std::vector<std::string> lines;
    lines.push_back("DS:power:GAUGE:5:0:100");
    lines.push_back(ds::str("RRA:AVERAGE:0.5:", 60, ":", 720));
    rrdInterface->create(filename, 1, std::chrono::system_clock::to_time_t(ds::DateTime{301s}), lines);

    THEN("max power is 100") {
      auto limit{0.0};
      for (const auto& entry : rrdInterface->info(filename)) {
        DS_NOTICE(entry);
        if (entry.key == "ds[power].max"s) {
          CHECK(entry.type == RrdInfoType::value);
          limit = std::get<rrd_value_t>(entry.value);
        }
      }
      CHECK(limit == 100);
    }
    THEN("tuning max power to 200") {
      rrdInterface->tune({"tune", filename, "--maximum", "power:200"s});
      THEN("max power is 200") {
        auto limit{0.0};
        for (const auto& entry : rrdInterface->info(filename)) {
          DS_NOTICE(entry);
          if (entry.key == "ds[power].max"s) {
            CHECK(entry.type == RrdInfoType::value);
            limit = std::get<rrd_value_t>(entry.value);
          }
        }
        CHECK(limit == 200);
      }
    }
  }
}

SCENARIO("DERIVE -> DDERIVE", TAGS) {
  auto filename = "test.rrd";
  DS_DEFER(boost::filesystem::remove(filename));

  auto rrdInterface = makeRRDInterface();

  GIVEN("DERIVE") {
    // https://oss.oetiker.ch/rrdtool/tut/rrdtutorial.en.html
    rrdInterface->create(filename, 300, 920804400,
                         {"DS:speed:COUNTER:600:U:U"s, "RRA:AVERAGE:0.5:1:24"s, "RRA:AVERAGE:0.5:6:10"s});

    auto values = {"920804700:12345"s, "920805000:12357"s, "920805300:12363"s, "920805600:12363"s, "920805900:12363"s,
                   "920806200:12373"s, "920806500:12383"s, "920806800:12393"s, "920807100:12399"s, "920807400:12405"s,
                   "920807700:12411"s, "920808000:12415"s, "920808300:12420"s, "920808600:12422"s, "920808900:12423"s};
    auto cmd = std::vector<std::string>{"update"s, filename};
    cmd.insert(std::end(cmd), std::cbegin(values), std::cend(values));
    rrdInterface->update(cmd);

    THEN("last value is present") {
      auto [lastChanged, data] = rrdInterface->lastValue(filename);
      CHECK(lastChanged == 920808900);
      CHECK(!data.empty());
      CHECK(data[0] == std::make_pair("speed"s, 12423.0));
    }

    THEN("xport values are present") {
      // rrdtool fetch test.rrd AVERAGE --start 920804400 --end 920809200
      auto lines = {"xport"s,      "--start"s,   "920804400"s,
                    "--end"s,      "920809200"s, ds::str("DEF:speed=", filename, ":speed:AVERAGE"),
                    "XPORT:speed"s};
      auto [start, end, step, values] = rrdInterface->xport(lines);
      CHECK(step == 300);                                        // 300s is resolution of data source
      CHECK(values.size() == (920809200 - 920804400) / 300 - 1); // 15
      CHECK(values[4].value == 0.0);
      CHECK(!std::isnan(values[7].value));
      CHECK(values[8].value == 0.02);
      CHECK(start == 920804400);
      CHECK(values[0].timeStamp == 920804400 + step);
      CHECK(end == 920809200);
      CHECK(values[14].timeStamp == 920809200 - step);
    }

    WHEN("setting upper limit to xport values ") {
      auto lines = {"xport"s,
                    "--start"s,
                    "920804400"s,
                    "--end"s,
                    "920809200"s,
                    ds::str("DEF:speed=", filename, ":speed:AVERAGE"),
                    ds::str("CDEF:limit=speed,0,0.02,LIMIT"),
                    "XPORT:limit"s};
      auto [start, end, step, values] = rrdInterface->xport(lines);
      THEN("values exceedding the limit become NaN") {
        CHECK(values[4].value == 0.0);
        CHECK(std::isnan(values[7].value));
        CHECK(values[8].value == 0.02);
      }
    }

    WHEN("converting DERIVE -> DDRERIVE") {
      std::vector<std::string> lines;
      lines.push_back("tune");
      lines.push_back(filename);
      lines.push_back("-d");
      lines.push_back("speed:DDERIVE"s);
      rrdInterface->tune(lines);

      THEN("last value is reset") { CHECK_THROWS(rrdInterface->lastValue(filename)); }

      THEN("remaining value are present") {
        auto lines = {"xport"s,      "--start"s,   "920804400"s,
                      "--end"s,      "920809200"s, ds::str("DEF:speed=", filename, ":speed:AVERAGE"),
                      "XPORT:speed"s};
        auto [start, end, step, values] = rrdInterface->xport(lines);
        CHECK(values.size() == 15);
        CHECK(values[4].value == 0.0);
        CHECK(values[8].value == 0.02);
        CHECK(start == 920804400);
        CHECK(end == 920809200);
        CHECK(step == 300);
      }
    }
  }
}

SCENARIO("getSeries", TAGS) {
  auto filename = "test.rrd"s;
  DS_DEFER(boost::filesystem::remove(filename));

  auto rrdInterface = makeRRDInterface();

  auto firstUpdate = time_t{920804400};

  // https://oss.oetiker.ch/rrdtool/tut/rrdtutorial.en.html
  rrdInterface->create(filename, 1, 920804400,
                       {"DS:speed:DERIVE:1:U:U"s,
                        "RRA:AVERAGE:0.5:1:60"s,    // 1s:600 samples
                        "RRA:AVERAGE:0.5:60:72"s,   // 1min:72min
                        "RRA:AVERAGE:0.5:900:8"s}); // 15min:2h

  for (int i = 1; i < 7200; i++) {
    rrdInterface->update(std::vector{"update"s, filename, ds::str(firstUpdate + i, ":", i)});
  }

  auto end = firstUpdate + 7200;

  GIVEN("start/end/step") {
    WHEN("10 samples@1s") {
      auto [start_, end_, step, values] =
          rrdInterface->xport(std::vector{"xport"s, "--start"s, ds::str(end - 11), "--end"s, ds::str(end), "--step"s,
                                          ds::str(1), ds::str("DEF:raw=", filename, ":speed:AVERAGE"), //
                                          ds::str("XPORT:raw")});

      CHECK(step == 1);
      CHECK(end_ == end);
      CHECK(start_ == end - 10 - step); // align to steps
      CHECK(values.size() == 10);
    }

    WHEN("10 samples@1min") {
      auto [start_, end_, step, values] =
          rrdInterface->xport(std::vector{"xport"s, "--start"s, ds::str(end - 601), "--end"s, ds::str(end), "--step"s,
                                          ds::str(60), ds::str("DEF:raw=", filename, ":speed:AVERAGE"), //
                                          ds::str("XPORT:raw")});

      CHECK(step == 60);
      CHECK(end_ == end + end % step);
      CHECK(start_ == end - 600 - step);
      CHECK(values.size() == 10);
    }

    WHEN("non-matching step size") {
      auto [start_, end_, step, values] =
          rrdInterface->xport(std::vector{"xport"s, "--start"s, ds::str(end - 11), "--end"s, ds::str(end), "--step"s,
                                          ds::str(13), ds::str("DEF:raw=", filename, ":speed:AVERAGE"), //
                                          ds::str("XPORT:raw")});

      DS_NOTICE(values.size(), start_, end_, step);
      CHECK(step == 13);
      CHECK((end_ - start_) % step == 0); // 26 s
      CHECK(values.size() == 1);
    }
  }

  GIVEN("start/end/step/values") {
    WHEN("10 samples@1s") {
      auto [start_, end_, step, values] = rrdInterface->xport(
          std::vector{"xport"s, "--start"s, ds::str(end - 601), "--end"s, ds::str(end), "--step"s, "1"s, "--maxrows"s,
                      ds::str(10), ds::str("DEF:raw=", filename, ":speed:AVERAGE"), ds::str("XPORT:raw")});

      DS_NOTICE(start_, end_, step, values.size());
      CHECK(values.size() == 10);
      CHECK(step == 60); // max(step, (end - start) / maxrows)
      CHECK(end_ == end);
      CHECK(start_ == end - 600 - step);
    }
  }

  GIVEN("no start end") {
    THEN("10 samples with unpredictable step") {
      auto [start_, end_, step, values] =
          rrdInterface->xport(std::vector{"xport"s, "--step"s, ds::str(1), "--maxrows"s, ds::str(10),
                                          ds::str("DEF:raw=", filename, ":speed:AVERAGE"), ds::str("XPORT:raw")});

      DS_NOTICE(start_, end_, step);
      CHECK(values.size() <= 10);
      CHECK(step != 1); // max(step, (end - start) / maxrows)
      // -> if start/end not given, now is taken, making number of values, step size, etc unpredictable
    }
  }
}

} // namespace
