/*
    Copyright (c) 2015 digitalSTROM.org, Zurich, Switzerland

    Author: Remy Mahler, digitalSTROM AG <remy.mahler@digitalstrom.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ds/log.h>
#include <ds/str.h>

#include "autoclustermaintenance.h"
#include "dss.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/modelconst.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "model/zone.h"
#include "structuremanipulator.h"

static ds::LogChannel dsLogChannel("modelAutoClusterMaintenance");

namespace dss {

AutoClusterMaintenance::AutoClusterMaintenance(Apartment *_apartment) : m_pApartment(_apartment) {
  DS_ASSERT(m_pApartment, "");
} /* AutoClusterMaintenance */

AutoClusterMaintenance::~AutoClusterMaintenance() {
  m_pApartment = NULL;
} /* ~AutoClusterMaintenance */

void AutoClusterMaintenance::globalConsistencyCheck() {
  static const int SleepTimeBetweenChecks = 1000;
  DS_ASSERT(m_pApartment, "");
  Set deviceRefs = m_pApartment->getDevices();
  Set greyDevices = deviceRefs.getByColor(DeviceClasses_t::DEVICE_CLASS_GR);

  for (int i = 0; i < greyDevices.length(); ++i) {
    const DeviceReference devRef = greyDevices.get(i);
    Device *device = const_cast<Device *>(devRef.getDevice().get());
    consistencyCheck(*device);
    sleepMS(SleepTimeBetweenChecks);
  }
}

void AutoClusterMaintenance::consistencyCheck(Device &_device) {
  DS_ASSERT(m_pApartment, "");
  boost::recursive_mutex::scoped_lock scoped_lock(m_pApartment->getMutex());

  removeInvalidAssignments(_device);

  // find assignment to locked and unlocked clusters
  int deviceAssignedInLockedCluster = getFirstLockedClusterAssignment(_device);
  std::vector<std::shared_ptr<Cluster> > validClusters = getUnlockedClusterAssignment(_device);

  if (deviceAssignedInLockedCluster != 0) {
    // remove from all other automatic and unlocked clusters
    for (std::shared_ptr<Cluster> cluster : validClusters) {
      removeDeviceFromCluster(_device, cluster);
    }
  } else if (!validClusters.empty()) {
    // check type assignment:
    // if cd_none and wpc_none => no assignment to cluster
    if ((_device.getCardinalDirection() == cd_none) && (_device.getWindProtectionClass() == wpc_none)) {
      for (std::shared_ptr<Cluster> cluster : validClusters) {
        removeDeviceFromCluster(_device, cluster);
      }
      return;
    }

    // keep first assignment, remove the rest
    int deviceAssignedInCluster = validClusters[0]->getID();
    for (std::shared_ptr<Cluster> cluster : validClusters) {
      if (cluster->getID() == deviceAssignedInCluster) {
        continue;
      }
      removeDeviceFromCluster(_device, cluster);
    }
  } else {
    // No assignment.
    if ((_device.getCardinalDirection() == cd_none) && (_device.getWindProtectionClass() == wpc_none)) {
      DS_INFO(_device.getDSID(), "is not configured. No assignment to cluster");
      return;
    }

    std::shared_ptr<Cluster> cluster =
        findOrCreateCluster(_device.getCardinalDirection(), _device.getWindProtectionClass());
    if (cluster == NULL) {
      DS_WARNING("Cluster can not be created:", _device.getWindProtectionClass(), _device.getCardinalDirection());
      return;
    }

    if (cluster->isConfigurationLocked()) {
      DS_WARNING("The targeted cluster is locked.", cluster->getID(), _device.getDSID(), "can not be added");
      return;
    }

    assignDeviceToCluster(_device, cluster);
  }
} /* consistencyCheck */

void AutoClusterMaintenance::joinIdenticalClusters() {
  DS_ASSERT(m_pApartment, "");
  boost::recursive_mutex::scoped_lock scoped_lock(m_pApartment->getMutex());

  std::vector<std::shared_ptr<Cluster> > clusters = m_pApartment->getClusters();
  for (unsigned ctr = 0; ctr < clusters.size() - 1; ++ctr) {
    std::shared_ptr<Cluster> cluster = clusters[ctr];
    if (!cluster->isAutomatic() || cluster->isConfigurationLocked()) {
      continue;
    }
    for (unsigned index = ctr + 1; index < clusters.size(); ++index) {
      if (!clusters[index]->isAutomatic() || clusters[index]->isConfigurationLocked()) {
        continue;
      }
      if (cluster->getLocation() == clusters[index]->getLocation() &&
          cluster->getProtectionClass() == clusters[index]->getProtectionClass()) {
        DS_WARNING("The clusters with ids", cluster->getApplicationType(), "and id",
                   clusters[index]->getApplicationType(), "are identical");
        moveClusterDevices(clusters[index], cluster);
      }
    }
  }
} /* joinIdenticalClusters */

void AutoClusterMaintenance::cleanupEmptyCluster() {
  DS_ASSERT(m_pApartment, "");
  boost::recursive_mutex::scoped_lock scoped_lock(m_pApartment->getMutex());

  removeEmptyAutomaticCluster();

} /* removeEmptyAutomaticCluster */

void AutoClusterMaintenance::moveClusterDevices(std::shared_ptr<Cluster> _clusterSource,
                                                std::shared_ptr<Cluster> _clusterDestination) {
  DS_ASSERT(m_pApartment, "");
  Set clusterDevices = _clusterSource->getDevices();

  for (int iDevice = 0; iDevice < clusterDevices.length(); ++iDevice) {
    DeviceReference &ref = clusterDevices.get(iDevice);
    std::shared_ptr<Device> device = ref.getDevice();

    removeDeviceFromCluster(*device.get(), _clusterSource);
    assignDeviceToCluster(*device.get(), _clusterDestination);
  }
} /* moveClusterDevices */

void AutoClusterMaintenance::removeInvalidAssignments(Device &_device) {
  // remove invalid assignments from unlocked and automatic clusters
  for (std::shared_ptr<Cluster> cluster : m_pApartment->getClusters()) {
    if (cluster->isAutomatic() && !cluster->isConfigurationLocked()) {
      if ((cluster->getLocation() != _device.getCardinalDirection()) ||
          (cluster->getProtectionClass() != _device.getWindProtectionClass())) {
        // device must not be in cluster
        removeDeviceFromCluster(_device, cluster);
      }

      // remove assignment to cluster with cd_none and wpc_none
      if ((_device.getCardinalDirection() == cd_none) && (_device.getWindProtectionClass() == wpc_none) &&
          (cluster->getLocation() == _device.getCardinalDirection()) &&
          (cluster->getProtectionClass() == _device.getWindProtectionClass())) {
        // device must not be in cluster
        removeDeviceFromCluster(_device, cluster);
      }
    }
  }
} /* removeInvalidAssignments */

void AutoClusterMaintenance::removeDeviceFromCluster(Device &_device, std::shared_ptr<Cluster> _cluster) {
  StructureManipulator structureManipulator(*m_pApartment);
  structureManipulator.deviceRemoveFromGroup(_device.sharedFromThis(), _cluster, false);

  removeEmptyAutomaticCluster();
} /* removeDeviceFromCluster */

void AutoClusterMaintenance::removeEmptyAutomaticCluster() {
  // remove empty, automatic and unlocked clusters
  for (std::shared_ptr<Cluster> cluster : m_pApartment->getClusters()) {
    if (cluster->isAutomatic() && !cluster->isConfigurationLocked()) {
      if (cluster->releaseCluster()) {
        busUpdateCluster(cluster);
      }
    }
  }
} /* removeEmptyAutomaticCluster */

void AutoClusterMaintenance::assignDeviceToCluster(Device &_device, std::shared_ptr<Cluster> _cluster) {
  StructureManipulator structureManipulator(*m_pApartment);
  structureManipulator.deviceAddToGroup(_device.sharedFromThis(), _cluster);
} /* assignDeviceToCluster */

std::shared_ptr<Cluster> AutoClusterMaintenance::findOrCreateCluster(CardinalDirection_t _cardinalDirection,
                                                                     WindProtectionClass_t _protection) {
  // find or create cluster
  DS_ASSERT(m_pApartment, "");

  std::vector<std::shared_ptr<Cluster> > clusters = m_pApartment->getClusters();
  for (std::shared_ptr<Cluster> cluster : clusters) {
    if ((cluster->getApplicationType() != ApplicationType::none) && (cluster->getProtectionClass() == _protection) &&
        (cluster->getLocation() == _cardinalDirection) && (cluster->isAutomatic())) {
      return cluster;
    }
  }

  // cluster does not exist. Get new one
  std::shared_ptr<Cluster> cluster = m_pApartment->getEmptyCluster();
  if (cluster == NULL) {
    return cluster;
  }

  cluster->setLocation(_cardinalDirection);
  cluster->setProtectionClass(_protection);
  // TODO: why do we create gray cluster?
  cluster->setApplicationType(ApplicationType::blinds);

  // Naming scheme: "<orientation> - Class <class> - <speed>m/s" (e.g. "South-West – Class 1 -9.8 m/s")
  // if no orientation is defined: only "Class z - x.y m/s" (no "none"-Orientation)
  if (cluster->getLocation() == cd_none) {
    cluster->setName(toUIString(_protection));
  } else {
    cluster->setName(toUIString(_cardinalDirection) + " - " + toUIString(_protection));
  }
  cluster->setAutomatic(true);
  busUpdateCluster(cluster);
  return cluster;
} /* findOrCreateCluster */

std::vector<std::shared_ptr<Cluster> > AutoClusterMaintenance::getUnlockedClusterAssignment(Device &_device) {
  std::vector<std::shared_ptr<Cluster> > assignedClusters;
  for (std::shared_ptr<Cluster> cluster : m_pApartment->getClusters()) {
    if (!cluster->isConfigurationLocked() && cluster->isAutomatic() &&
        (cluster->getLocation() == _device.getCardinalDirection()) &&
        (cluster->getProtectionClass() == _device.getWindProtectionClass()) && (_device.isInGroup(cluster->getID()))) {
      assignedClusters.push_back(cluster);
    }
  }
  return assignedClusters;
} /* getUnlockedClusterAssignment */

int AutoClusterMaintenance::getFirstLockedClusterAssignment(Device &_device) {
  for (std::shared_ptr<Cluster> cluster : m_pApartment->getClusters()) {
    if (cluster->isConfigurationLocked() && cluster->isAutomatic() &&
        (cluster->getLocation() == _device.getCardinalDirection()) &&
        (cluster->getProtectionClass() == _device.getWindProtectionClass()) && (_device.isInGroup(cluster->getID()))) {
      return cluster->getID();
    }
  }
  return 0;
} /* getFirstLockedClusterAssignment */

void AutoClusterMaintenance::busUpdateCluster(std::shared_ptr<Cluster> _cluster) {
  try {
    // operation can be overridden for testing.
    StructureModifyingBusInterface *itf = m_pApartment->getBusInterface()->getStructureModifyingBusInterface();
    itf->clusterSetName(_cluster->getID(), _cluster->getName());
    itf->clusterSetApplication(_cluster->getID(), _cluster->getApplicationType(),
                               _cluster->getApplicationConfiguration());
    itf->clusterSetProperties(_cluster->getID(), _cluster->getLocation(), _cluster->getFloor(),
                              _cluster->getProtectionClass());
    itf->clusterSetLockedScenes(_cluster->getID(), _cluster->getLockedScenes());
  } catch (std::runtime_error &err) {
    DS_WARNING(_cluster->getID(), "can not be updated:", err.what());
  }
} /* busUpdateCluster */

} // namespace dss
