/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#pragma once

#include "businterface.h"
#include "ds485/dsbusinterface.h"
#include "ds485types.h"
#include "logger.h"
#include "model/device.h"
#include "model/set.h"
#include "structuremanipulator.h"
#include "taskprocessor.h"

namespace dss {

class StructureQueryBusInterface;
class DSDeviceBusInterface;
class Apartment;
class DSMeter;
class Device;
class ModelMaintenance;
class Zone;
class Set;

void setupBaseDevice(std::shared_ptr<Device> device, const DeviceSpec& spec);

class BusScanner {
public:
  BusScanner(StructureQueryBusInterface& interface, Apartment& apartment, ModelMaintenance& maintenance);
  bool scanDSMeter(std::shared_ptr<DSMeter> dsMeter);
  bool scanDeviceOnBus(std::shared_ptr<DSMeter> dsMeter, devid_t shortAddress);
  void syncBinaryInputStates(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Device> device);
  void synchronizeDSMeterData(std::shared_ptr<DSMeter> dsMeter, DSMeterSpec_t& spec);

  // for unit-tests
  bool scanZone(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Zone> zone);
  bool scanClusters(std::shared_ptr<DSMeter> dsMeter);
  bool scanPowerStates(std::shared_ptr<DSMeter> dsMeter);
  void scanDevicesOfZoneQuick(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Zone> zone);
  bool scanGroupsOfZone(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Zone> zone);
  bool scanStatusOfZone(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Zone> zone);
  bool scanTemperatureControl(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Zone> zone);
  void log(const std::string& line, aLogSeverity severity = lsDebug);
  bool initializeDeviceFromSpec(std::shared_ptr<DSMeter> dsm, std::shared_ptr<Zone> zone, const DeviceSpec& spec);

private:
  bool initializeDeviceFromSpecQuick(std::shared_ptr<DSMeter> dsMeter, std::shared_ptr<Zone> zone,
                                     const DeviceSpec& spec);
  void scheduleDeviceReadout(const std::shared_ptr<Device> pDevice);
  void setMeterCapability(std::shared_ptr<DSMeter> dsMeter);
  bool applyMeterSpec(std::shared_ptr<DSMeter> dsMeter);
  bool getMeterHash(std::shared_ptr<DSMeter> dsMeter, DSMeterHash_t& hash);
  bool getMeterState(std::shared_ptr<DSMeter> dsMeter, uint8_t* state);

private:
  Apartment& m_Apartment;
  StructureQueryBusInterface& m_Interface;
  ModelMaintenance& m_Maintenance;
  StructureManipulator m_structureManipulator;
};

class BinaryInputOrStateDeviceFilter : public IDeviceAction {
private:
  std::vector<std::shared_ptr<Device> > m_devs;

public:
  virtual bool perform(std::shared_ptr<Device> _device) {
    if (_device->isPresent() && (_device->getBinaryInputCount() > 0 || !_device->getStates().empty())) {
      m_devs.push_back(_device);
    }
    return true;
  }
  std::vector<std::shared_ptr<Device> > getDeviceList() { return m_devs; }
};

class BinaryInputScanner : public Task {
public:
  BinaryInputScanner(Apartment* _papartment, const std::string& _busConnection);
  virtual ~BinaryInputScanner();

  virtual void run();
  virtual void setup(std::shared_ptr<DSMeter> _dsMeter, std::shared_ptr<Device> _device);

  uint16_t getDeviceSensorValue(const dsuid_t& _dsm, dev_t _device, uint8_t _sensorIndex) const;
  uint8_t getDeviceConfig(const dsuid_t& _dsm, dev_t _device, uint8_t _configClass, uint8_t _configIndex) const;

private:
  Apartment* m_pApartment;
  std::string m_busConnection;
  DsmApiHandle_t m_dsmApiHandle;
  std::weak_ptr<DSMeter> m_dsm;
  std::weak_ptr<Device> m_device;
};

} // namespace dss
