/*
    Copyright (c) 2021 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "buttonmanipulator.h"

#include <ds/log.h>

#include <string>

using namespace dss;
using namespace dss::model;
using namespace std::literals;

static ds::LogChannel dsLogChannel("modelButtonManipulator");

namespace dss::model {

bool isAkmDevice(const Device& device) {
  // TODO same as getBinaryInputCount() > 0
  return (device.getDeviceType() == DEVICE_TYPE_AKM) || (device.getDeviceType() == DEVICE_TYPE_UMR) ||
         (device.getDeviceType() == DEVICE_TYPE_SKM && device.getDeviceNumber() == 300 &&
         (device.getButtonInputMode() >= ButtonInputMode::AKM_STANDARD && 
         device.getButtonInputMode() <= ButtonInputMode::AKM_FALLING_EDGE)) || // DSD-8498
         (device.getDeviceType() == DEVICE_TYPE_TKM && device.getDeviceNumber() >= 300) ||
         device.hardwareType() == HardwareType::GR_KL300;
}

namespace {

struct ButtonManipulatorImpl : public ButtonManipulator {
  Apartment& m_apartment;
  StructureManipulator& m_structureManipulator;

  ButtonManipulatorImpl(Apartment& apartment, StructureManipulator& structureManipulator)
      : m_apartment(apartment), m_structureManipulator(structureManipulator) {}

  ErrorCode setButtonInputMode(Device& device, std::string&& value) override {
    auto pDevice = device.sharedFromThis();
    DeviceFeatures_t features = pDevice->getDeviceFeatures();
    if (features.pairing == false) {
      return {false, "This device does not support button pairing"s};
    }

    if ((pDevice->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_SW) &&
        (pDevice->getActiveGroup() == GroupIDBlack) && (value != BUTTONINPUT_1WAY)) {
      return {false, "Joker devices must be set to a specific group for button pairing"};
    }

    auto pPartnerDevice = pDevice->getPartnerDevice();
    const auto& index = pDevice->getButtonInputIndex();
    bool wasPartnerSlave = pPartnerDevice->is2WaySlave();

    DS_NOTICE("Set button input mode.", *pDevice, value, index, wasPartnerSlave, pDevice->getZoneID(),
              pPartnerDevice->getZoneID());

    if (value == BUTTONINPUT_2WAY_DOWN) {
      if (index == 0) {
        pDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT2);
        pPartnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT1);
      } else {
        pDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT4);
        pPartnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT3);
      }
    } else if (value == BUTTONINPUT_2WAY_UP) {
      if (index == 0) {
        pDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2);
        pPartnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1);
      } else {
        pDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT4);
        pPartnerDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT3);
      }
    } else if (value == BUTTONINPUT_1WAY) {
      pDevice->setDeviceButtonInputMode(ButtonInputMode::STANDARD);
      if (pPartnerDevice->is2WaySlave()) {
        auto defaultInputMode = [](const auto& device) -> ButtonInputMode {
          return (device.hardwareType() == HardwareType::GR_KL300) ? ButtonInputMode::AKM_STANDARD
                                                                   : ButtonInputMode::STANDARD;
        };
        pPartnerDevice->setDeviceButtonInputMode(defaultInputMode(*pDevice));

        if (pPartnerDevice->isBinaryInput() || !pPartnerDevice->hasOutput()) {
          auto group = pPartnerDevice->getBinaryInput(0)->m_targetGroupId;
          if (device.hardwareType() == HardwareType::GR_KL300) {
            group = ColorIDBlack;
            pPartnerDevice->setDeviceButtonID((group << 4) | ButtonId_Generic);
          }
          // TODO this does too much, the input already has the right group set
          // would be sufficient to removeFromGroup, but there is no such method yet
          m_structureManipulator.setJokerGroup(pPartnerDevice, group);
        }
      }
    } else if (value == BUTTONINPUT_2WAY) {
      pDevice->setDeviceButtonInputMode(ButtonInputMode::TWO_WAY);
      pPartnerDevice->setDeviceButtonInputMode(ButtonInputMode::DEACTIVATED);
    } else if (value == BUTTONINPUT_1WAY_COMBINED) {
      pDevice->setDeviceButtonInputMode(ButtonInputMode::ONE_WAY);
      pPartnerDevice->setDeviceButtonInputMode(ButtonInputMode::DEACTIVATED);
    } else {
      return {false, "Invalid mode specified"};
    }

    if (value == BUTTONINPUT_1WAY) {
      return {true, ""s};
    }

    if (pDevice->getZoneID() != pPartnerDevice->getZoneID()) {
      std::shared_ptr<Zone> zone = m_apartment.getZone(pDevice->getZoneID());
      m_structureManipulator.addDeviceToZone(pPartnerDevice, zone);
    }

    // #3450 - remove slave devices from clusters
    m_structureManipulator.deviceRemoveFromGroups(pPartnerDevice);

    if (features.syncButtonID == true) {
      if (pDevice->getButtonID() != pPartnerDevice->getButtonID()) {
        pPartnerDevice->setDeviceButtonID(pDevice->getButtonID());
      }
    }

    if ((pDevice->getActiveGroup() != GroupIDNotApplicable) &&
        (pDevice->getActiveGroup() != pPartnerDevice->getActiveGroup())) {
      m_structureManipulator.setJokerGroup(pPartnerDevice, pDevice->getActiveGroup());
    }
    return {true, ""s};
  }

  ErrorCode setAKMInputProperty(Device& device, const AutomationInputMode automationInputMode) override {
    if (auto partnerDevice = device.tryGetPartnerDevice(); partnerDevice) {
      DS_REQUIRE(!partnerDevice->is2WaySlave(), "partner device must be unpaired first");
    }

    if (!isAkmDevice(device)) {
      return {false, "This device does not support AKM properties"s};
    }

    DS_NOTICE("Set akm input property.", device, automationInputMode);

    if (auto buttonInputMode = ds::tryToEnum<ButtonInputMode>(automationInputMode); buttonInputMode) {
      device.setDeviceButtonInputMode(*buttonInputMode);
      return {true, ""s};
    }

    return {false, ds::str("Unsupported mode: '", automationInputMode, "'")};
  }
};

} // namespace

} // namespace dss::model

std::unique_ptr<ButtonManipulator> dss::model::makeButtonManipulator(Apartment& apartment,
                                                                     StructureManipulator& structureManipulator) {
  return std::make_unique<ButtonManipulatorImpl>(apartment, structureManipulator);
}
