/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "floor.h"

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif
#include <ds-model/model-fwd.h>

#include <algorithm>

#include "apartment.h"
#include "base.h"
#include "propertysystem.h"
#include "zone.h"

namespace dss {

Floor::Floor(int id, Apartment *apartment) : m_id(id), m_order(0),
                                                      m_apartment(apartment) {
  m_name = "Floor #" + intToString(id);
}

int Floor::getID() const {
  return m_id;
}

dsModel::FloorId Floor::id() const {
  return dsModel::FloorId{ds::str(m_id)};
}

void Floor::setOrder(int position) {
  m_order = position;
}

int Floor::getOrder() const {
  return m_order;
}

const std::string& Floor::getName() const {
  return m_name;
}

void Floor::setName(std::string name) {
  m_name = name;
  if (m_apartment) {
    if (auto&& modelMaintenance = m_apartment->getModelMaintenance()) {
      modelMaintenance->addModelEvent(new ModelEvent(ModelEvent::etModelDirty));
    }
  }
}

std::vector<dsModel::ZoneId> Floor::zones() const {
  std::vector<dsModel::ZoneId> result;
  for (const auto& zone : m_zones) {
    result.push_back(dsModel::ZoneId{ds::str(zone)});
  }
  return result;
}

void Floor::addZone(int id) {
  std::shared_ptr<Zone> zone = m_apartment->getZone(id); // will throw if ID is not found
  std::shared_ptr<Floor> oldFloor;
  if (zone->getFloorID() != FLOOR_ID_NONE) {
    try {
      oldFloor = m_apartment->getFloor(zone->getFloorID());
    } catch (std::runtime_error& e) {}
  }

  if (std::find(m_zones.begin(), m_zones.end(), id) == m_zones.end()) {
    m_zones.push_back(id);

    if ((oldFloor != NULL) && (oldFloor->getID() != getID())) {
      oldFloor->removeZone(zone->getID());
    }
    zone->setFloorID(m_id);
    if (m_propertyNode != NULL) {
      m_propertyNode->createProperty("zones")->createProperty("zone" + intToString(id) + "/id")->setIntegerValue(id);
    }
  }
}

void Floor::removeZone(int id) {
  std::vector<int>::iterator it = std::find(m_zones.begin(), m_zones.end(), id);
  if (it != m_zones.end()) {
    m_zones.erase(it);
    if (m_propertyNode != NULL) {
      PropertyNodePtr zone = m_propertyNode->getProperty("zones/zone" + intToString(id));
      if (zone != NULL) {
        zone->getParentNode()->removeChild(zone);
      }
    }
  }
}

void Floor::publishToPropertyTree() {
  if (m_propertyNode == NULL) {
    if(m_apartment->getPropertyNode() != NULL) {
      m_propertyNode = m_apartment->getPropertyNode()->createProperty("floors/floor" + intToString(m_id));
      m_propertyNode->createProperty("id")->setIntegerValue(m_id);
      m_propertyNode->createProperty("name")->linkToProxy(PropertyProxyReference<std::string>(m_name, false));
      m_propertyNode->createProperty("order")->linkToProxy(PropertyProxyReference<int>(m_order, false));
      PropertyNodePtr zones = m_propertyNode->createProperty("zones");
      for (size_t i = 0; i < m_zones.size(); i++) {
        zones->createProperty("zone" + intToString(m_zones.at(i)) + "/id")->setIntegerValue(m_zones.at(i));
      }
    }
  }
}

void Floor::removeFromPropertyTree() {
  if (m_propertyNode != NULL) {
    m_propertyNode->getParentNode()->removeChild(m_propertyNode);
    m_propertyNode.reset();
  }
}

//TODO: +update proptree when adding/removing zones
//      +integrate floor class into apartment
//      add json calls
} // namespace
