/*
    Copyright (c) 2014 digitalSTROM AG, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>
            Michael Tross <michael.tross@digitalstrom.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

extern "C" {
    #include <SolTrack.h>
}

#include "event.h"
#include "event/event_create.h"
#include "logger.h"
#include "model/autoclustermaintenance.h"
#include "model/device.h"
#include "model/group.h"
#include "model/cluster.h"
#include "model/modelconst.h"
#include "model/modelmaintenance.h"
#include "model/dsmeter.h"
#include "model/set.h"
#include "model/state.h"
#include "model/zone.h"
#include "monitor_tasks.h"
#include "handler/system_states.h"
#include "security/security.h"
#include "structuremanipulator.h"
#include "time.h"

static ds::LogChannel dsLogChannel("monitorTasks");

namespace dss {

bool SensorMonitorTask::checkValueDueTime(SensorType _sensorType, DateTime _ts) {
  DateTime now;
  if (_ts != DateTime::NullDate) {
    int age = now.difference(_ts);
    return (age > SensorMaxLifeTime);
  }
  return false;
}

bool SensorMonitorTask::checkZoneValue(std::shared_ptr<Group> _group, SensorType _sensorType, DateTime _ts) {
  bool zoneTimeDue = checkValueDueTime(_sensorType, _ts);
  if (zoneTimeDue) {
    DateTime now;
    Logger::getInstance()->log(std::string("Sensor value (type: ") +
        sensorTypeName(_sensorType) + ") for zone #" +
        intToString(_group->getZoneID()) +
        " is too old: " + _ts.toISO8601_ms() +
        ", age in seconds is " + intToString(now.difference(_ts)), lsInfo);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().
          pushEvent(createZoneSensorErrorEvent(_group, _sensorType, _ts));
    }
  }
  return zoneTimeDue;
}

void SensorMonitorTask::run() {
  DS_CONTEXT("sensorMonitor");

  DSS::getInstance()->getSecurity().loginAsSystemUser("SensorMonitorTask needs system rights");

  try {
    std::vector<std::shared_ptr<Device> > devices = m_Apartment->getDevicesVector();
    for (size_t i = 0; i < devices.size(); i++) {
      std::shared_ptr<Device>& device = devices.at(i);
      if (!device) {
        continue;
      }

      if (!device->isValid() || !device->isPresent()) {
        continue;
      }

      uint8_t sensorCount = device->getSensorCount();
      for (uint8_t s = 0; s < sensorCount; s++) {
        const auto& sensor = device->getSensor(s);
        // as specced in #6371
        if (sensor.m_sensorPollInterval == 0) {
          continue;
        }

        int age = 0;
        if (sensor.m_sensorValueTS == DateTime::NullDate) {
          if (DSS::hasInstance()) {
            age = DSS::getInstance()->getUptime();
          }
        } else {
          DateTime now = DateTime();
          age = now.difference(sensor.m_sensorValueTS);
        }

        Logger::getInstance()->log(
            std::string("Sensor #") + intToString(s) + " of device " + dsuid2str(device->getDSID()) +
            " value is from: " + sensor.m_sensorValueTS.toISO8601_ms() + ", and age in seconds is " + intToString(age));

        // value is invalid because its older than its allowed lifetime
        if ((age > SensorMaxLifeTime) && device->isSensorDataValid(s)) {
          Logger::getInstance()->log(std::string("Sensor #") + intToString(s) + " of device " +
                                         dsuid2str(device->getDSID()) + " value is too old: " +
                                         sensor.m_sensorValueTS.toISO8601_ms() + ", age in seconds is " +
                                         intToString(age),
              lsInfo);
          device->setSensorDataValidity(s, false);

          if (DSS::hasInstance()) {
            std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(device, m_Apartment);
            std::shared_ptr<Event> pEvent =
                createDeviceInvalidSensorEvent(pDevRef, s, sensor.m_sensorType, sensor.m_sensorValueTS);
            DSS::getInstance()->getEventQueue().pushEvent(pEvent);
          }
        }
      }
    }
  } catch (const std::exception& e) {
    Logger::getInstance()->log("SensorMonitorTask: device sensor timeout exception: " + std::string(e.what()), lsWarning);
  }

  try {
    std::vector<std::shared_ptr<Zone> > zones = m_Apartment->getZones();
    for (std::vector<std::shared_ptr<Zone> >::iterator it = zones.begin(); it != zones.end(); it ++) {
      std::shared_ptr<Zone> pZone = *it;
      if (pZone->getID() > 0) {
        ZoneSensorStatus hSensors = pZone->getSensorStatus();

        checkZoneSensor(pZone, SensorType::TemperatureIndoors, hSensors);
        checkZoneSensor(pZone, SensorType::HumidityIndoors,    hSensors);
        checkZoneSensor(pZone, SensorType::CO2Concentration,   hSensors);
        checkZoneSensor(pZone, SensorType::BrightnessIndoors,  hSensors);

        ZoneHeatingStatus hStatus = pZone->getHeatingStatus();
        if (checkZoneValue(pZone->getGroup(GroupIDControlTemperature), SensorType::RoomTemperatureControlVariable,
                           hStatus.controlValueTS)) {
          pZone->resetControlValue();
        }
      }
    }
  } catch (const std::exception& e) {
    Logger::getInstance()->log("SensorMonitorTask: zone sensor timeout exception: " + std::string(e.what()), lsWarning);
  }

  try {
    ApartmentSensorStatus hSensors = m_Apartment->getSensorStatus();
    checkOutdoorSensor(SensorType::TemperatureOutdoors, hSensors);
    checkOutdoorSensor(SensorType::BrightnessOutdoors, hSensors);
    checkOutdoorSensor(SensorType::HumidityOutdoors, hSensors);
    checkOutdoorSensor(SensorType::AirPressure, hSensors);
    checkOutdoorSensor(SensorType::GustSpeed, hSensors);
    checkOutdoorSensor(SensorType::WindSpeed, hSensors);
    checkOutdoorSensor(SensorType::WindDirection, hSensors);
    checkOutdoorSensor(SensorType::Precipitation, hSensors);
  } catch (const std::exception& e) {
    Logger::getInstance()->log("SensorMonitorTask: sensor timeout exception: " + std::string(e.what()), lsWarning);
  }

  std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::CheckSensorValues);
  pEvent->setProperty("time", "+600");
  if (DSS::hasInstance()) {
    Logger::getInstance()->log("queued check_sensor_values event");
    DSS::getInstance()->getEventQueue().pushEvent(pEvent);
  }
}

DateTime SensorMonitorTask::getDateTimeForZoneSensor(const ZoneSensorStatus& _hSensors, SensorType _sensorType) {
  switch (_sensorType) {
    case SensorType::TemperatureIndoors: {
      return _hSensors.temperatureValueTS;
    }
    case SensorType::HumidityIndoors: {
      return _hSensors.humidityValueTS;
    }
    case SensorType::BrightnessIndoors: {
      return _hSensors.brightnessValueTS;
    }
    case SensorType::CO2Concentration: {
      return _hSensors.co2ConcentrationValueTS;
    }
    case SensorType::ActivePower: //no break
    case SensorType::OutputCurrent: //no break
    case SensorType::ElectricMeter: //no break
    case SensorType::TemperatureOutdoors: //no break
    case SensorType::BrightnessOutdoors: //no break
    case SensorType::HumidityOutdoors: //no break
    case SensorType::AirPressure: //no break
    case SensorType::GustSpeed: //no break
    case SensorType::GustDirection: //no break
    case SensorType::WindSpeed: //no break
    case SensorType::WindDirection: //no break
    case SensorType::Precipitation: //no break
    case SensorType::COConcentration: //no break
    case SensorType::SoundPressureLevel: //no break
    case SensorType::RoomTemperatureSetpoint: //no break
    case SensorType::RoomTemperatureControlVariable: //no break
    case SensorType::Status: ///< Status sensor value. Value bits are defined by \re StatusFieldType
    case SensorType::Reserved1: //no break
    case SensorType::Reserved2: //no break
    case SensorType::OutputCurrent16A: //no break
    case SensorType::ActivePowerVA: //no break
    case SensorType::Temperature: //no break
    case SensorType::Brightness: //no break
    case SensorType::Humidity: //no break
    case SensorType::WaterQuantity: //no break
    case SensorType::WaterFlowRate: //no break
    case SensorType::NotUsed: //no break
    case SensorType::UnknownType: //no break
    case SensorType::SunAzimuth:
    case SensorType::SunElevation:
    default:
      throw std::runtime_error("Unexpected SensorType: " + sensorTypeName(_sensorType));
  }
}

void SensorMonitorTask::checkZoneSensor(std::shared_ptr<Zone> _zone, SensorType _sensorType,
                                        const ZoneSensorStatus& _hSensors) {
  std::shared_ptr<Device> sensorDevice = _zone->getAssignedSensorDevice(_sensorType);

  bool sensorFault = (sensorDevice && !sensorDevice->isPresent());

  DateTime sensorTime = getDateTimeForZoneSensor(_hSensors, _sensorType);
  if (checkValueDueTime(_sensorType, sensorTime)) {
    sensorFault = true;
    switch (_sensorType) {
      case SensorType::TemperatureIndoors: {
        _zone->resetTemperature();
        break;
      }
      case SensorType::HumidityIndoors: {
        _zone->resetHumidityValue();
        break;
      }
      case SensorType::BrightnessIndoors: {
        _zone->resetBrightnessValue();
        break;
      }
      case SensorType::CO2Concentration: {
        _zone->resetCO2ConcentrationValue();
        break;
      }
      case SensorType::ActivePower: //no break
      case SensorType::OutputCurrent: //no break
      case SensorType::ElectricMeter: //no break
      case SensorType::TemperatureOutdoors: //no break
      case SensorType::BrightnessOutdoors: //no break
      case SensorType::HumidityOutdoors: //no break
      case SensorType::AirPressure: //no break
      case SensorType::GustSpeed: //no break
      case SensorType::GustDirection: //no break
      case SensorType::WindSpeed: //no break
      case SensorType::WindDirection: //no break
      case SensorType::Precipitation: //no break
      case SensorType::COConcentration: //no break
      case SensorType::SoundPressureLevel: //no break
      case SensorType::RoomTemperatureSetpoint: //no break
      case SensorType::RoomTemperatureControlVariable: //no break
      case SensorType::Status: ///< Status sensor value. Value bits are defined by \re StatusFieldType
      case SensorType::Reserved1: //no break
      case SensorType::Reserved2: //no break
      case SensorType::OutputCurrent16A: //no break
      case SensorType::ActivePowerVA: //no break
      case SensorType::Temperature: //no break
      case SensorType::Brightness: //no break
      case SensorType::Humidity: //no break
      case SensorType::WaterQuantity: //no break
      case SensorType::WaterFlowRate: //no break
      case SensorType::NotUsed: //no break
      case SensorType::UnknownType: //no break
      case SensorType::SunAzimuth:
      case SensorType::SunElevation:
      default:
        throw std::runtime_error("Unexpected SensorType: " + sensorTypeName(_sensorType));
    }
  }

  if (sensorFault) {
    DateTime now;
    Logger::getInstance()->log(std::string("Sensor not available, or value (type: ") +
        sensorTypeName(_sensorType) + ") for zone #" +
        intToString(_zone->getID()) +
        " is too old: " + sensorTime.toISO8601_ms() +
        ", age in seconds is " + intToString(now.difference(sensorTime)), lsWarning);
    _zone->getGroup(GroupIDBroadcast)->sensorInvalid(_sensorType);

    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().
        pushEvent(createZoneSensorErrorEvent(_zone->getGroup(GroupIDBroadcast), _sensorType, DateTime::NullDate));
    }
  }
}

DateTime SensorMonitorTask::getDateTimeForOutdoorSensor(const ApartmentSensorStatus& _hSensors, SensorType _sensorType) {
  switch (_sensorType) {
    case SensorType::TemperatureOutdoors: return _hSensors.m_TemperatureTS;
    case SensorType::BrightnessOutdoors: return _hSensors.m_BrightnessTS;
    case SensorType::HumidityOutdoors: return _hSensors.m_HumidityTS;
    case SensorType::AirPressure: return _hSensors.m_AirPressureTS;
    case SensorType::GustSpeed: return _hSensors.m_WindGustTS;
    case SensorType::WindSpeed: return _hSensors.m_WindSpeedTS;
    case SensorType::WindDirection: return _hSensors.m_WindDirectionTS;
    case SensorType::Precipitation: return _hSensors.m_PrecipitationTS;
    case SensorType::SunAzimuth: return _hSensors.m_SunAzimuthTS;
    case SensorType::SunElevation: return _hSensors.m_SunElevationTS;
    case SensorType::GustDirection: [[fallthrough]]; ///< outdoor sensor removed from apartment datamodel
    case SensorType::TemperatureIndoors: [[fallthrough]];
    case SensorType::HumidityIndoors: [[fallthrough]];
    case SensorType::BrightnessIndoors: [[fallthrough]];
    case SensorType::CO2Concentration: [[fallthrough]];
    case SensorType::ActivePower: [[fallthrough]];
    case SensorType::OutputCurrent: [[fallthrough]];
    case SensorType::ElectricMeter: [[fallthrough]];
    case SensorType::COConcentration: [[fallthrough]];
    case SensorType::SoundPressureLevel: [[fallthrough]];
    case SensorType::RoomTemperatureSetpoint: [[fallthrough]];
    case SensorType::RoomTemperatureControlVariable: [[fallthrough]];
    case SensorType::Status: [[fallthrough]]; ///< Status sensor value. Value bits are defined by \re StatusFieldType
    case SensorType::Reserved1: [[fallthrough]];
    case SensorType::Reserved2: [[fallthrough]];
    case SensorType::OutputCurrent16A: [[fallthrough]];
    case SensorType::ActivePowerVA: [[fallthrough]];
    case SensorType::Temperature: [[fallthrough]];
    case SensorType::Brightness: [[fallthrough]];
    case SensorType::Humidity: [[fallthrough]];
    case SensorType::WaterQuantity: [[fallthrough]];
    case SensorType::WaterFlowRate: [[fallthrough]];
    case SensorType::NotUsed: [[fallthrough]];
    case SensorType::UnknownType: [[fallthrough]];
    default:
      throw std::runtime_error("Unexpected SensorType: " + sensorTypeName(_sensorType));
  }
}

void SensorMonitorTask::checkOutdoorSensor(SensorType _sensorType, const ApartmentSensorStatus& _hSensors) {
  DateTime sensorTime = getDateTimeForOutdoorSensor(_hSensors, _sensorType);

  if (!checkValueDueTime(_sensorType, sensorTime)) {
    return;
  }

  switch (_sensorType) {
    case SensorType::TemperatureOutdoors:
      m_Apartment->setTemperature(0.0, DateTime::NullDate);
      break;
    case SensorType::BrightnessOutdoors:
      m_Apartment->setBrightnessValue(0.0, DateTime::NullDate);
      break;
    case SensorType::HumidityOutdoors:
      m_Apartment->setHumidityValue(0.0, DateTime::NullDate);
      break;
    case SensorType::AirPressure:
      m_Apartment->setAirPressure(0.0, DateTime::NullDate);
      break;
    case SensorType::GustSpeed:
      m_Apartment->setWindGust(0.0, DateTime::NullDate);
      break;
    case SensorType::WindSpeed:
      m_Apartment->setWindSpeed(0.0, DateTime::NullDate);
      break;
    case SensorType::WindDirection:
      m_Apartment->setWindDirection(0.0, DateTime::NullDate);
      break;
    case SensorType::Precipitation:
      m_Apartment->setPrecipitation(0.0, DateTime::NullDate);
      break;
    case SensorType::GustDirection: [[fallthrough]]; ///< outdoor sensor ignored by the system
    case SensorType::HumidityIndoors: [[fallthrough]];
    case SensorType::TemperatureIndoors: [[fallthrough]];
    case SensorType::BrightnessIndoors: [[fallthrough]];
    case SensorType::CO2Concentration: [[fallthrough]];
    case SensorType::ActivePower: [[fallthrough]];
    case SensorType::OutputCurrent: [[fallthrough]];
    case SensorType::ElectricMeter: [[fallthrough]];
    case SensorType::COConcentration: [[fallthrough]];
    case SensorType::SoundPressureLevel: [[fallthrough]];
    case SensorType::RoomTemperatureSetpoint: [[fallthrough]];
    case SensorType::RoomTemperatureControlVariable: [[fallthrough]];
    case SensorType::Status: [[fallthrough]]; ///< Status sensor value. Value bits are defined by \re StatusFieldType
    case SensorType::Reserved1: [[fallthrough]];
    case SensorType::Reserved2: [[fallthrough]];
    case SensorType::OutputCurrent16A: [[fallthrough]];
    case SensorType::ActivePowerVA: [[fallthrough]];
    case SensorType::Temperature: [[fallthrough]];
    case SensorType::Brightness: [[fallthrough]];
    case SensorType::Humidity: [[fallthrough]];
    case SensorType::WaterQuantity: [[fallthrough]];
    case SensorType::WaterFlowRate: [[fallthrough]];
    case SensorType::NotUsed: [[fallthrough]];
    case SensorType::UnknownType: [[fallthrough]];
    case SensorType::SunAzimuth: [[fallthrough]];
    case SensorType::SunElevation: [[fallthrough]];
    default:
        throw std::runtime_error("Unexpected SensorType: " + sensorTypeName(_sensorType));
  }

  DateTime now;
  DS_WARNING("Sensor value is too old.", sensorTypeName(_sensorType), intToString(now.difference(sensorTime)));
  m_Apartment->getGroup(GroupIDBroadcast)->sensorInvalid(_sensorType);
}

void HeatingMonitorTask::syncZone(int _zoneID) {
  try {
    std::shared_ptr<Zone> pZone = m_Apartment->getZone(_zoneID);
    std::shared_ptr<Group> pGroup = pZone->getGroup(GroupIDControlTemperature);
    ZoneSensorStatus hSensors = pZone->getSensorStatus();
    const TemperatureConfig& hProp = pZone->temperatureConfig();

    switch (hProp.m_mode) {
      case HeatingControlMode::PID:
        if (HeatingOperationModeInvalid != pZone->getHeatingOperationMode()) {
          pGroup->callScene(CallOrigin::system, SceneAccessCategory::manual, pZone->getHeatingOperationMode(), "", false);
          usleep(1000 * 1000);
        }
        if (hSensors.temperatureValueTS != DateTime::NullDate && hSensors.temperatureValue) {
          pZone->pushSensor(CallOrigin::system, SceneAccessCategory::manual, DSUID_NULL, SensorType::TemperatureIndoors,
                            *hSensors.temperatureValue, "");
          usleep(1000 * 1000);
        }

        break;
      case HeatingControlMode::FIXED:
        if (HeatingOperationModeInvalid != pZone->getHeatingOperationMode()) {
          pGroup->callScene(CallOrigin::system, SceneAccessCategory::manual, pZone->getHeatingOperationMode(), "", false);
          usleep(1000 * 1000);
        }
        break;
      case HeatingControlMode::MANUAL:
        if (HeatingOperationModeInvalid != pZone->getHeatingOperationMode()) {
          ZoneHeatingStatus hStatus = pZone->getHeatingStatus();
          if (hStatus.controlValueTS != DateTime::NullDate && hStatus.controlValue) {
            pGroup->pushSensor(CallOrigin::system, SceneAccessCategory::manual, DSUID_NULL, SensorType::RoomTemperatureControlVariable,
                               *hStatus.controlValue, "");
            usleep(1000 * 1000);
          }
        }
        break;
      case HeatingControlMode::ZONE_FOLLOWER:
      case HeatingControlMode::OFF:
        break;
    }
  } catch (const std::exception& e) {
    Logger::getInstance()->log("HeatingMonitorTask: sync controller exception: " + std::string(e.what()), lsWarning);
  }
}

void HeatingMonitorTask::run() {
  const auto& event = m_event->getName();
  DS_CONTEXT(event, "heatingMonitor");

  if (m_event->getName() == EventName::ModelReady) {
    auto event = std::make_shared<Event>(EventName::TemperatureControllersElection);
    event->setProperty("time", "+60");
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(event);
    }

    return;
  }

  DSS::getInstance()->getSecurity().loginAsSystemUser("HeatingMonitorTask needs system rights");

  if (m_event->getName() == EventName::TemperatureControllersElection) {
    try {
      StructureManipulator manipulator(*m_Apartment);
      manipulator.electTemperatureControllers();
    } catch (const std::exception& e) {
      // be sure that timer goes on
      DS_ERROR("UpdateZoneTemperatureMeters failed.", e);
    }

    auto event = std::make_shared<Event>(EventName::TemperatureControllersElection);
    event->setProperty("time", "+60");
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(event);
    }
    return;
  }

  if (m_event->getName() == EventName::ZoneSensorValue && m_event->getRaiseLocation() == dss::erlGroup) {
    if (m_event->getPropertyByName<SensorType>("sensorType") == SensorType::RoomTemperatureControlVariable) {
      // Only elected controllers are allowed to send heating control values.
      // Values from non-elected controllers imply active non-elected controller.
      // Recover by disabling it.
      // https://jira.digitalstrom.org/browse/DSD-6149
      const auto origin = Dsuid{m_event->getPropertyByName("originDSID")};
      auto zone = m_Apartment->zone(m_event->getRaisedAtGroup()->zoneId());
      const auto electedTemperatureController = zone->temperatureController();
      DS_INFO("HeatingMonitor - roomTemperatureSetpoint", electedTemperatureController, origin, zone);
      if (electedTemperatureController && origin != electedTemperatureController->getDSID()) {
        DS_WARNING("HeatingMonitor - control value from non-elected controller -> disabling.", origin, zone,
                   electedTemperatureController);
        auto& modifying = *m_Apartment->getBusInterface()->getStructureModifyingBusInterface();
        modifying.disableZoneHeatingConfig(origin, zone->getID());
      }
    }
  }

  int zoneID;
  if (m_event->getName() == EventName::HeatingControllerState) {
    if (m_event->getPropertyByName("ControlState") == intToString(HeatingControlStateIDEmergency)) {
      zoneID = strToInt(m_event->getPropertyByName("ZoneID"));
      dsuid_t dsmdsuid = str2dsuid(m_event->getPropertyByName("ControlDSUID"));

      Logger::getInstance()->log("HeatingMonitorTask: emergency state in zone " +
          intToString(zoneID) + " and controller " + dsuid2str(dsmdsuid),
          lsWarning);
      syncZone(zoneID);
    }
  } else if (m_event->getName() == EventName::DSMeterReady) {
    dsuid_t dsmdsuid = str2dsuid(m_event->getPropertyByName("dsMeter"));
    std::shared_ptr<DSMeter> pMeter = m_Apartment->getDSMeterByDSID(dsmdsuid);
    Set devList = pMeter->getDevices();
    std::vector<std::shared_ptr<Zone> > zones = m_Apartment->getZones();
    for (size_t i = 0; i < zones.size(); i++) {
      if (zones[i]->getID() == 0) {
        continue;
      }

      Set devices = devList.getByZone(zones[i]->getID());
      if (!devices.isEmpty() || zones[i]->registeredOnDSMeter(pMeter)) {
        syncZone(zones[i]->getID());
      }
    }
  }
}

size_t HeatingValveProtectionTask::m_zoneIndex = 0;

/**
 * Every Thursday around 15:00 send a HeatingOperationModeIDValveProtection scene call
 * to all rooms that contain devices in the temperature control group. The calls into
 * the zones are delayed by 15 minutes each.
 */
void HeatingValveProtectionTask::run() {
  const auto& event = m_event->getName();
  DS_CONTEXT("heatingValveProtection", event);

  if (m_event->getName() == EventName::ModelReady) {
    std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::HeatingValveProtection);
    // randomize the valve protection over two hours
    int randomizeStartMinutes = rand() % 120;
    DateTime start = DateTime::parseRFC2445("19700101T140000");
    start = start.addMinute(randomizeStartMinutes);
    pEvent->setProperty(EventProperty::ICalStartTime, start.toRFC2445IcalDataTime());
    pEvent->setProperty(EventProperty::ICalRRule, "FREQ=WEEKLY;BYDAY=TH");
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    }
    return;
  }

  DSS::getInstance()->getSecurity().loginAsSystemUser("HeatingValveProtectionTask needs system rights");

  if (m_event->getName() == EventName::HeatingValveProtection) {
    try {
      std::vector<std::shared_ptr<Zone> > zones = m_Apartment->getZones();
      for ( ; m_zoneIndex < zones.size(); m_zoneIndex++) {
        if (zones[m_zoneIndex]->getID() == 0) {
          continue;
        }
        std::shared_ptr<Group> tempControlGroup = zones[m_zoneIndex]->getGroup(GroupIDControlTemperature);
        Set devices = tempControlGroup->getDevices();
        if (devices.isEmpty()) {
          continue;
        }
        tempControlGroup->callScene(CallOrigin::system, SceneAccessCategory::manual, HeatingOperationModeIDValveProtection, "", false);
        std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::HeatingValveProtection);
        pEvent->setProperty("time", "+900");
        if (DSS::hasInstance()) {
          DSS::getInstance()->getEventQueue().pushEvent(pEvent);
        }
        m_zoneIndex++;
        return;
      }
      m_zoneIndex = 0;

    } catch (const std::exception& e) {
      Logger::getInstance()->log("HeatingValveProtectionTask: exception: " + std::string(e.what()), lsWarning);
    }
  }
}

const std::string AutoClusterTask::propAutoConfigDelay = "/config/subsystems/EventInterpreter/autoconfigDelay";

void AutoClusterTask::initPropertyTree() {
    static const int defaultTimeOffset = 1800; // 30 min
    DSS::getInstance()->getPropertySystem().setIntValue(propAutoConfigDelay, defaultTimeOffset, true, false);
}

void AutoClusterTask::run() {
  const auto& event = m_event->getName();
  DS_CONTEXT("autoCluster", event);

  if (m_event->getName() == EventName::ModelReady) {
    std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::AutoClusterUpdate);
    int timeOffset = DSS::getInstance()->getPropertySystem().getIntValue(propAutoConfigDelay);
    std::string data = "+" + intToString(timeOffset);
    pEvent->setProperty("time", data);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    }
    return;
  }
  if (m_event->getName() == EventName::AutoClusterUpdate) {
    AutoClusterMaintenance maintenance(m_Apartment);
    maintenance.globalConsistencyCheck();
  }
}

__DEFINE_LOG_CHANNEL__(LockClusterTask, lsInfo);

const std::string LockClusterTask::propStartupDelay = "/config/subsystems/EventInterpreter/lockClusterStartupDelay";
const std::string LockClusterTask::propRefreshDelay = "/config/subsystems/EventInterpreter/lockClusterRefreshDelay";

void LockClusterTask::initDefaults() {
  static const int defaultStartupDelay = 60 * 5; // 5 min
  DSS::getInstance()->getPropertySystem().setIntValue(propStartupDelay, defaultStartupDelay, true, false);
  static const int defaultRefreshDelay = 60 * 60; // 60 min
  DSS::getInstance()->getPropertySystem().setIntValue(propRefreshDelay, defaultRefreshDelay, true, false);
}

std::string LockClusterTask::createLockStateName(const int& _clusterId) {
  return std::string("cluster." + intToString(_clusterId) + ".user_lock");
}

bool LockClusterTask::evalLockState(std::shared_ptr<State> _lockState) {
  if (nullptr == _lockState || _lockState->getType() != StateType_Group) {
    return false;
  }

  std::shared_ptr<Cluster> cl = std::dynamic_pointer_cast<Cluster> (_lockState->getProviderGroup().lock());
  std::vector<std::string> lockEvents = cl->getLockEvents();
  std::shared_ptr<State> systemState;
  int newLockState = State_Inactive;
  for (auto& sName : lockEvents) {
    // wind and rain can exist in cluster or apartment
    std::string clusterStateName = sName + ".group" + intToString(cl->getID());
    systemState = m_Apartment->tryGetNonScriptState(clusterStateName);
    if (nullptr == systemState) {
      systemState = m_Apartment->tryGetNonScriptState(sName);
    }
    if (nullptr != systemState && systemState->getState() == State_Active) {
      newLockState = State_Active;
      break;
    }
  }

  bool stateChanged = false;
  if (_lockState->getState() != newLockState) {
    stateChanged = true;
  }
  _lockState->setState(CallOrigin::system, newLockState);
  return stateChanged;
}

void LockClusterTask::scheduleCommand(const int& _clusterId, bool _immediate) {
  std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::LockClusterRefresh);
  if (!_immediate) {
    int delay = DSS::getInstance()->getPropertySystem().getIntValue(propRefreshDelay);
    pEvent->setProperty("time", std::string("+") + intToString(delay));
  }
  pEvent->setProperty("clusterId", intToString(_clusterId));
  if (DSS::hasInstance()) {
    DSS::getInstance()->getEventQueue().pushEvent(pEvent);
  }
  DS_INFO("Schedule refresh for cluster " + intToString(_clusterId));
}

void LockClusterTask::handleClusterUpdate() {
  std::list<std::shared_ptr<State> > stateList;

  // iterate over clusters and find lockedEvents configurations
  auto clusters = m_Apartment->getClusters();
  for (auto&& cl : clusters) {
    std::shared_ptr<State> lockRequestState;
    std::string sName = createLockStateName(cl->getID());
    std::vector<std::string> lockEvents = cl->getLockEvents();
    if (lockEvents.size() > 0) {
      DS_INFO("Update " + sName, lockEvents);
      try {
        lockRequestState = m_Apartment->getNonScriptState(sName);
      } catch (ItemNotFoundException& e) {
        lockRequestState = m_Apartment->allocateGroupState(cl, sName);
        DS_DEBUG("Created new locking state", lockRequestState);
      }
      stateList.push_back(lockRequestState);
    } else {
      try {
        lockRequestState = m_Apartment->getNonScriptState(sName);
        m_Apartment->removeState(lockRequestState);
        DS_DEBUG("Removed locking state", lockRequestState);
      } catch (ItemNotFoundException& e) {
      }
    }
  }
  // update the lock request states for changed states
  for (auto&& lState : stateList) {
    if (evalLockState(lState)) {
      scheduleCommand(lState->getProviderGroup().lock()->getID(), true);
    }
  }
}

void LockClusterTask::handleStateUpdate(const std::string& _stateName) {
  if (beginsWith(_stateName, StateName::Hail) || beginsWith(_stateName, StateName::Frost) ||
      beginsWith(_stateName, StateName::Rain) || beginsWith(_stateName, StateName::Wind)) {
    DS_DEBUG("State update " + _stateName);

    std::shared_ptr<State> lockState;
    size_t dot = _stateName.find(".group");

    if (std::string::npos != dot) {
      // is there a user lock in this cluster?
      int groupId = -1;
      try {
        groupId = strToInt(_stateName.substr(dot + 6));
      } catch(std::exception& e) {
      }
      if (isCluster(groupId)) {
        std::shared_ptr<Cluster> cl = m_Apartment->getCluster(groupId);
        std::string lockId = createLockStateName(cl->getID());
        try {
          lockState = m_Apartment->getNonScriptState(lockId);
          DS_INFO("Evaluate new lock for " + lockId);
          if (evalLockState(lockState)) {
            scheduleCommand(cl->getID(), true);
          }
        } catch(ItemNotFoundException& e) {}
      }
    } else {
      // iterate over all clusters and get those with user locks
      for (auto&& cl : m_Apartment->getClusters()) {
        std::string lockId = createLockStateName(cl->getID());
        try {
          lockState = m_Apartment->getNonScriptState(lockId);
          DS_INFO("Evaluate new lock for " + lockId);
          if (evalLockState(lockState)) {
            scheduleCommand(cl->getID(), true);
          }
        } catch(ItemNotFoundException& e) {}
      }
    }
  }
}

void LockClusterTask::run() {
  std::list<std::shared_ptr<State> > stateList;
  const auto& event = m_event->getName();
  DS_CONTEXT("lockCluster", event);

  try {

    if (m_event->getName() == EventName::ModelReady) {
      // re-evaluate the user lock states a few after system startup finished
      int delay = DSS::getInstance()->getPropertySystem().getIntValue(propStartupDelay);
      std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::LockClusterUpdate);
      pEvent->setProperty("time", std::string("+") + intToString(delay));
      if (DSS::hasInstance()) {
        DSS::getInstance()->getEventQueue().pushEvent(pEvent);
      }
      return;
    }

    if (m_event->getName() == EventName::LockClusterUpdate) {
      handleClusterUpdate();
      return;
    }

    //  send the command again in locked cluster after at least 60 minutes
    if (m_event->getName() == EventName::LockClusterRefresh) {
      std::string lockId;
      try {
        std::string clusterId = m_event->getPropertyByName("clusterId");
        int clNumber = strToInt(clusterId);
        std::string lockId = createLockStateName(clNumber);
        std::shared_ptr<State> lockState = m_Apartment->getNonScriptState(lockId);
        DS_INFO("Refresh lock: " + lockId, lockState);

        if (lockState->getState() == State_Active) {
          m_Apartment->getBusInterface()->getStructureModifyingBusInterface()->clusterSetOperationLock(clNumber, true);
          scheduleCommand(clNumber, false);
        } else if (lockState->getState() == State_Inactive) {
          m_Apartment->getBusInterface()->getStructureModifyingBusInterface()->clusterSetOperationLock(clNumber, false);
        }
      } catch(std::invalid_argument& e) {
        DS_NOTICE(std::string("Invalid clusterId ") + lockId + ": " + std::string(e.what()));
      } catch(ItemNotFoundException& e) {
        DS_NOTICE(std::string("Data model error: ") + std::string(e.what()));
      }
      return;
    }

    // update the lock status in the affected clusters of the alarm scene call
    if (m_event->getName() == EventName::StateChange) {
        handleStateUpdate(m_event->getPropertyByName("statename"));
        return;
    }

  } catch (const std::exception& e) {
    DS_ERROR("Exception: " + std::string(e.what()));
  }
}

void UpdateShadeStopTimingsTask::run() {
  KLDeviceFilter filterKL;
  int i;

  Set devices =
      m_Apartment->getDevices().getByColor(DeviceClasses_t::DEVICE_CLASS_GR).getSubset(filterKL).getByPresence(true);

  for (i = 0; i < devices.length(); i++) {
    std::shared_ptr<Device> device = devices[i].getDevice();
    try {
      // position time has not yet been read out
      if (device->getMaxPosTime() == 0) {
          double val = round(std::max(device->getDevicePosTimeUp(),
                                      device->getDevicePosTimeDown()));
          if (val > UINT8_MAX) {
            val = UINT8_MAX;
            Logger::getInstance()->log("UpdateShadeStopTimingsTask: maximum "
                "position time for device " + dsuid2str(device->getDSID()) +
                "exceeded dSM range and was truncated.", lsNotice);
          }
          device->setMaxPosTime((uint8_t)val);
          sleep(10); // do not block the bus by repeated readouts
      }
    } catch (const std::exception &ex) {
    Logger::getInstance()->log("UpdateShadeStopTimingsTask: could not read "
        "position information from device " +
        dsuid2str(device->getDSID()) + ", " +
        std::string(ex.what()), lsWarning);
    }
  }

  std::vector<std::shared_ptr<Cluster> > clusters = m_Apartment->getClusters();
  for (size_t s = 0; s < clusters.size(); s++) {
    std::shared_ptr<Cluster> cluster = clusters.at(s);
    if (cluster->getColor() != ColorIDGray) {
      continue;
    }

    Set clusterDevices = devices.getByGroup(cluster);
    if (clusterDevices.isEmpty()) {
      continue;
    }

    uint8_t maxPosTime = 0;
    for (i = 0; i < clusterDevices.length(); i++) {
      maxPosTime = std::max(maxPosTime, clusterDevices[i].getDevice()->
              getMaxPosTime());
    }

    if (maxPosTime > 0) {
      try {
        m_Apartment->getBusInterface()->getStructureModifyingBusInterface()->
            clusterSetStopTimeout(cluster->getID(), maxPosTime);
      } catch (const std::exception &ex) {
        Logger::getInstance()->log("UpdateShadeStopTimingsTask: could not set "
        "stop timeout on cluster " + intToString(cluster->getID()) + ", " +
        std::string(ex.what()), lsError);
      }
    }
  }

  std::vector<std::shared_ptr<Zone> > zones = m_Apartment->getZones();
  for (size_t s = 0; s < zones.size(); s++) {
    std::shared_ptr<Zone> zone = zones.at(s);
    if (zone->getID() == 0) {
      continue;
    }

    Set zoneDevices = devices.getByZone(zone->getID());
    if (zoneDevices.isEmpty()) {
      continue;
    }

    uint8_t maxPosTime = 0;
    for (i = 0; i < zoneDevices.length(); i++) {
      maxPosTime = std::max(maxPosTime, zoneDevices[i].getDevice()->
              getMaxPosTime());
    }

    if (maxPosTime > 0) {
      try {
        m_Apartment->getBusInterface()->getStructureModifyingBusInterface()->
            zoneSetStopTimeout(zone->getID(), GroupIDGray, maxPosTime);
      } catch (const std::exception &ex) {
        Logger::getInstance()->log("UpdateShadeStopTimingsTask: could not set "
        "stop timeout on group " + intToString(zone->getID()) + ", " +
        std::string(ex.what()), lsError);
      }
    }
  }
}

void SunPositionSensorTask::run() {
  // geographic position and output are in degrees
  int useDegrees = 1;

  // Azimuth: 0 = South, pi/2 (90deg) = West  ->  0 = North, pi/2 (90deg) = East
  int useNorthEqualsZero = 1;

  // Compute the distance to the Sun in AU
  int computeDistance = 0;

  time_t now = time(NULL);
  struct tm utc;
  gmtime_r(&now, &utc);

  struct Time sol_time;
  sol_time.year = utc.tm_year + 1900;
  sol_time.month = utc.tm_mon + 1;
  sol_time.day = utc.tm_mday + 1;
  sol_time.hour = utc.tm_hour;
  sol_time.minute = utc.tm_min;
  sol_time.second = utc.tm_sec;

  struct Location loc;

  try {
    PropertySystem* ps = m_Apartment->getPropertySystem();
    PropertyNodePtr longnode = ps->getProperty("/config/geodata/longitude");
    PropertyNodePtr latnode = ps->getProperty("/config/geodata/latitude");

    if (!longnode || !latnode) {
      throw std::runtime_error("missing longitude or latitude data");
    }

    loc.longitude = strToDouble(longnode->getAsString());
    loc.latitude = strToDouble(latnode->getAsString());

    // compure refraction-corrected equatorial coordinates (Hour angle, declination)
    int computeRefrEquatorial = 0;

    auto sensorStatus = m_Apartment->getSensorStatus();
    if ((sensorStatus.m_AirPressureTS != DateTime::NullDate) && (sensorStatus.m_TemperatureTS != DateTime::NullDate)) {
      computeRefrEquatorial = 1;
      loc.pressure = sensorStatus.m_AirPressureValue / 10.0;      // hPa -> kPa
      loc.temperature = sensorStatus.m_TemperatureValue + 273.15; // C -> K
    }

    struct Position pos;
    SolTrack(sol_time, loc, &pos, useDegrees, useNorthEqualsZero,
             computeRefrEquatorial, computeDistance);

    DateTime ts = DateTime(now);

    m_Apartment->setSunAzimuth(pos.azimuthRefract, ts);
    m_Apartment->setSunElevation(pos.altitude, ts);

    std::shared_ptr<Zone> pZone = m_Apartment->getZone(0);
    pZone->pushSensor(CallOrigin::system, SceneAccessCategory::manual, DSUID_NULL,
                      SensorType::SunAzimuth, pos.azimuthRefract, "");

    pZone->pushSensor(CallOrigin::system, SceneAccessCategory::manual, DSUID_NULL,
                      SensorType::SunElevation, pos.altitude, "");
  } catch (std::runtime_error &ex) {
    // dSS is being restarted each time the above config nodes are modified,
    // so listeners for the proptree are not needed
    Logger::getInstance()->log("SunPositionSensorTask: location coordinates "
                               "not set, azimuth/altitude sensor values will "
                               "not be available: " + std::string(ex.what()), lsError);
  }
}

}// namespace
