/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "structuremanipulator.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>


#include "businterface-mock.h"
#include "model/apartment-mock.h"
#include "model/cluster.h"
#include "model/device-mock.h"
#include "model/modelmaintenance-mock.h"
#include "model/zone.h"

static ds::LogChannel dsLogChannel("structureManipulatorTest");

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const char* TAGS = "[structure][manipulator]";

SCENARIO("Clusters", TAGS) {
  PropertySystem propertySystem;
  auto busInterface = BusInterfaceMock{};
  auto pApartment = makeApartment(nullptr, &propertySystem, &busInterface);
  auto& apartment = *pApartment;
  auto zone = apartment.allocateZone(0);
  auto modelMaintenance = ModelMaintenanceMock{};
  auto cluster = apartment.getCluster(16);

  StructureManipulator structManipulator(apartment);

  SECTION("removeCluster") {
    apartment.allocateState(StateType_Apartment, " wind.group16", "");
    apartment.allocateState(StateType_Apartment, " cluster.16.operation_lock", "");
    WHEN("Cluster is removed") {
      REQUIRE_CALL(busInterface.structureModify, removeCluster(_));
      structManipulator.removeGroup(zone, 16);
      THEN("All related states are removed") {
        REQUIRE_FALSE(apartment.tryGetNonScriptState("wind.group16"));
        REQUIRE_FALSE(apartment.tryGetNonScriptState("cluster.16.operation_lock"));
      }
    }
  }
}

SCENARIO("Sensors", TAGS) {
  PropertySystem propertySystem;
  auto busInterface = BusInterfaceMock{};
  ApartmentMock apartment{nullptr, &propertySystem, &busInterface};

  auto modelMaintenance = ModelMaintenanceMock{};

  auto structure = StructureManipulator(apartment);

  GIVEN("zone with a sensor in it") {
    auto& zone = *apartment.allocateZone(7);
    auto& sensor = *createDevice(apartment, {.hardwareType = HardwareType::SW_AKM200});

    sensor.setSensors({DeviceSensorSpec_t{.sensorType = SensorType::TemperatureIndoors}});
    markOnline(sensor);
    zone.addDevice({sensor.sharedFromThis(), &apartment});

    GIVEN("sensor is default sensor in zone") {
      zone.setSensor({sensor.getDSID(), SensorType::TemperatureIndoors, 1});
      REQUIRE(zone.isSensorAssigned(SensorType::TemperatureIndoors));

      WHEN("resetZoneSensor") {
        REQUIRE_CALL(busInterface.structureModify, resetZoneSensor(zone.getID(), SensorType::TemperatureIndoors));
        structure.resetZoneSensor(zone, SensorType::TemperatureIndoors);
        THEN("no assigned sensor in zone") { REQUIRE(!zone.isSensorAssigned(SensorType::TemperatureIndoors)); }
      }
    }

    GIVEN("no sensor is default sensor in zone") {
      REQUIRE(!zone.isSensorAssigned(SensorType::TemperatureIndoors));

      WHEN("autoAssignZoneSensors") {
        REQUIRE_CALL(busInterface.structureModify,
                     setZoneSensor(zone.getID(), SensorType::TemperatureIndoors, sensor.getDSID()));
        structure.autoAssignZoneSensors(zone);
        THEN("sensor becomes default zone sensor") {
          REQUIRE(*zone.getAssignedSensorDevice(SensorType::TemperatureIndoors) == sensor);
        }
      }
    }
  }

  GIVEN("zone with two sensors in it") {
    auto& zone = *apartment.allocateZone(7);

    auto& sensor1 = *createDevice(apartment, {.hardwareType = HardwareType::SW_AKM200});
    sensor1.setSensors({DeviceSensorSpec_t{.sensorType = SensorType::TemperatureIndoors}});
    markOnline(sensor1);
    zone.addDevice({sensor1.sharedFromThis(), &apartment});

    auto& sensor2 = *createDevice(apartment, {.hardwareType = HardwareType::SW_AKM200});
    sensor2.setSensors({DeviceSensorSpec_t{.sensorType = SensorType::TemperatureIndoors}});
    markOnline(sensor2);
    zone.addDevice({sensor2.sharedFromThis(), &apartment});

    GIVEN("no default sensor present") {
      REQUIRE(!zone.isSensorAssigned(SensorType::TemperatureIndoors));

      WHEN("synchronizeZoneSensorAssignment ") {
        // for all possible sensors in the zone
        ALLOW_CALL(busInterface.structureModify, resetZoneSensor(zone.getID(), _));
        structure.synchronizeZoneSensorAssignment(zone);

        THEN("no assigned sensor in zone") {
          // explicit autoAssignZoneSensors needed
          REQUIRE(!zone.isSensorAssigned(SensorType::TemperatureIndoors));
        }
      }
    }

    GIVEN("first sensor is default sensor") {
      zone.setSensor({sensor1.getDSID(), SensorType::TemperatureIndoors, 1});
      REQUIRE(zone.isSensorAssigned(SensorType::TemperatureIndoors));

      WHEN("first sensor is moved to another zone") {
        auto& zoneAlt = *apartment.allocateZone(11);
        zoneAlt.addDevice({sensor1.getDSID(), &apartment});

        THEN("sensor is still default sensor") {
          // TODO: shouldn't checkSensorsOnDeviceRemoval do that?
          // this is probably a bug, inconsistent state and exception if another thread queries current sensor
          REQUIRE(zone.isSensorAssigned(SensorType::TemperatureIndoors));
          REQUIRE_THROWS(*zone.getAssignedSensorDevice(SensorType::TemperatureIndoors));
        }

        WHEN("sensors in zone are resynced again") {
          ALLOW_CALL(busInterface.structureModify, resetZoneSensor(zone.getID(), _));
          structure.synchronizeZoneSensorAssignment(zone); // sensor is still removed

          THEN("no assigned sensor in fist zone") { REQUIRE(!zone.isSensorAssigned(SensorType::TemperatureIndoors)); }

          WHEN("autoAssignZoneSensors") {
            REQUIRE_CALL(busInterface.structureModify,
                         setZoneSensor(zone.getID(), SensorType::TemperatureIndoors, sensor2.getDSID()));
            structure.autoAssignZoneSensors(zone);

            THEN("sensor2 becomes default zone sensor") {
              REQUIRE(*zone.getAssignedSensorDevice(SensorType::TemperatureIndoors) == sensor2);
            }
          }
        }
      }
    }
  }
}

} // namespace
