/*
    Copyright (c) 2013 digitalSTROM.org, Zurich, Switzerland

    Authors: Christian Hitz, aizo AG <christian.hitz@aizo.com>
             Michael Tross, aizo GmbH <michael.tross@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include "base.h"
#include "dss.h"
#include "expatparser.h"
#include "logger.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/zone.h"
#include "util.h"

#include <math.h>
#include <vdcapi.pb.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <boost/algorithm/string/replace.hpp>
#include <boost/filesystem.hpp>
#include <regex>

namespace dss {

void synchronizeGroups(Apartment* _apartment, StructureModifyingBusInterface* _interface) {
  std::vector<std::shared_ptr<Cluster> > clusters = _apartment->getClusters();
  for (std::shared_ptr<Cluster> pCluster : clusters) {
    if (!pCluster->isSynchronized()) {
      Logger::getInstance()->log("Forced configuration update in cluster " + intToString(pCluster->getID()), lsInfo);

      try {
        _interface->clusterSetName(pCluster->getID(), pCluster->getName());
        _interface->clusterSetApplication(pCluster->getID(), pCluster->getApplicationType(),
                                          pCluster->getApplicationConfiguration());
        _interface->clusterSetProperties(pCluster->getID(), pCluster->getLocation(), pCluster->getFloor(),
                                         pCluster->getProtectionClass());
        _interface->clusterSetLockedScenes(pCluster->getID(), pCluster->getLockedScenes());
        pCluster->setIsSynchronized(true);
      } catch (BusApiError& e) {
        Logger::getInstance()->log(
            "Error updating configuration for cluster " + intToString(pCluster->getID()) + ": " + e.what(), lsWarning);
      }
    }

    std::vector<std::shared_ptr<Group> > globalApps = _apartment->getGlobalApps();
    for (std::shared_ptr<Group> pGlobalApp : globalApps) {
      if (!pGlobalApp->isSynchronized()) {
        Logger::getInstance()->log(
            "Forced configuration update in global application group " + intToString(pGlobalApp->getID()), lsInfo);

        try {
          _interface->groupSetApplication(0, pGlobalApp->getID(), pGlobalApp->getApplicationType(),
                                          pGlobalApp->getApplicationConfiguration());

          pGlobalApp->setIsSynchronized(true);
        } catch (BusApiError& e) {
          Logger::getInstance()->log(
              "Error updating global application group " + intToString(pGlobalApp->getID()) + ": " + e.what(),
              lsWarning);
        }
      }
    }

    std::vector<std::shared_ptr<Zone> > zones = _apartment->getZones();
    for (std::shared_ptr<Zone> pZone : zones) {
      if (pZone->getID() == 0 || !pZone->isConnected()) {
        continue;
      }
      std::vector<std::shared_ptr<Group> > groups = pZone->getGroups();
      for (std::shared_ptr<Group> pGroup : groups) {
        // Regular-User-Groups
        if (isZoneUserGroup(pGroup->getID())) {
          try {
            _interface->createGroup(pZone->getID(), pGroup->getID(), pGroup->getApplicationType(),
                                    pGroup->getApplicationConfiguration(), pGroup->getName());
            _interface->groupSetApplication(pZone->getID(), pGroup->getID(), pGroup->getApplicationType(),
                                            pGroup->getApplicationConfiguration());
            pGroup->setIsSynchronized(true);
          } catch (BusApiError& e) {
            Logger::getInstance()->log("Error updating user group configuration in zone " +
                                           intToString(pZone->getID()) + " and group " + intToString(pGroup->getID()) +
                                           ": " + e.what(),
                                       lsWarning);
          }
        }
      }
    }
  }
}

std::string escapeHTML(std::string _input) {
  boost::replace_all(_input, "&", "&amp;");
  boost::replace_all(_input, "\"", "&quot;");
  boost::replace_all(_input, "'", "&#039;");
  boost::replace_all(_input, "<", "&lt;");
  boost::replace_all(_input, ">", "&gt;");

  return _input;
}

std::string unescapeHTML(std::string _input) {
  boost::replace_all(_input, "&amp;", "&");
  boost::replace_all(_input, "&quot;", "\"");
  boost::replace_all(_input, "&#039;", "'");
  boost::replace_all(_input, "&lt;", "<");
  boost::replace_all(_input, "&gt;", ">");

  return _input;
}

// Due to issues with old browsers and unicode line feed(U+2028) and paragraph
// feed(U+2029), we need to replace them with a space, in case some user
// decides to use them in the user editable fields. Refixes Redmine #14863.
std::string sanitizeLineFeeds(const std::string& userString) {
  const char* raw = reinterpret_cast<const char*>(u8"(\u2028|\u2029)");
  const static std::regex whitespace(raw);

  return regex_replace(userString,whitespace," ");
}

typedef struct OutputChannelInfo {
  uint8_t size;
  uint8_t min;
  double max;
} OutputChannelInfo_t;

const OutputChannelInfo& outputChannelInfo(ChannelType type) {
  static const std::map<ChannelType, OutputChannelInfo> outputChannels {
    {ChannelType::none, {0, 0, 0}},                       // id 0
    {ChannelType::brightness, {8, 0, 100}},               // id 1  light brightness
    {ChannelType::hue, {8, 0, 360}},                      // id 2  colored light hue
    {ChannelType::saturation, {8, 0, 100}},               // id 3  colored light saturation
    {ChannelType::colortemp, {8, 100, 1000}},             // id 4, color temperature mired
    {ChannelType::x, {8, 0, 1.0}},                        // id 5  CIE color model x component
    {ChannelType::y, {8, 0, 1.0}},                        // id 6  CIE color model y component
    {ChannelType::shadePositionOutside, {16, 0, 100}},    // id 7  shade position opening percentage for e.g. blinds and roller shutters
    {ChannelType::shadePositionIndoor, {16, 0, 100}},     // id 8  shade position opening percentage for e.g. curtains
    {ChannelType::shadeOpeningAngleOutside, {8, 0, 100}}, // id 9  shade position opening angle for e.g. lamellars
    {ChannelType::shadeOpeningAngleIndoor, {8, 0, 100}},  // id 10  indoor shade position opening angle for e.g. lamellars
    {ChannelType::transparency, {8, 0, 100}},             // id 11  transparency of e.g. a smart window
    {ChannelType::airFlowIntensity, {8, 0, 100}},         // id 12  intensity of ventilation
    {ChannelType::airFlowDirection, {8, 0, 2}},           // id 13  direction of air flow
    {ChannelType::airFlapPosition, {8, 0, 100}},          // id 14  flap position
    {ChannelType::airLouverPosition, {8, 0, 100}},        // id 15  louver position
    {ChannelType::heatingPower, {8, 0, 100}},             // id 16  heating power and intensity
    {ChannelType::coolingCapacity, {8, 0, 100}},          // id 17  cooling capacity and intensity
    {ChannelType::audioVolume, {8, 0, 100}},              // id 18  audio loudness
    {ChannelType::powerState, {8, 0, 2}},                 // id 19  power status
    {ChannelType::airLouverAuto, {8, 0, 1}},              // id 20  autmatic louver position of ventilation
    {ChannelType::airFlowAuto, {8, 0, 1}},                // id 21  autmatic intensity of ventilation
    {ChannelType::waterTemperature, {8, 0, 150}},         // id 22  water temperature
    {ChannelType::waterFlowRate, {8, 0, 100}},            // id 23  water flow
    {ChannelType::powerLevel, {8, 0, 100}},               // id 24  power level
    {ChannelType::videoStation, {16, 0, 10000}},          // id 25  tv station number
    {ChannelType::videoInputSource, {8, 0, 100}},         // id 26  tv input source enum e.g. TV, HDMI
  };
  
  const auto& channel = outputChannels.find(type);
  if (channel != outputChannels.end()) {
    return channel->second;
  }

  DS_FAIL_REQUIRE("Invalid output type.", type);
}

std::pair<ChannelType, uint8_t> getOutputChannelIdAndSize(std::string name) {
  auto type = ds::tryToEnum<ChannelType>(name);
  if (!type.has_value()) {
    DS_FAIL_REQUIRE("Invalid output type name.", name);
  }
  const auto& info = outputChannelInfo(*type);
  return std::make_pair(*type, info.size);
}

std::string getOutputChannelName(ChannelType type) {
  if (auto name = ds::tryFromEnum<std::string>(type)) {
    return *name;
  }

  DS_FAIL_REQUIRE("Invalid output type.", type);
}

uint16_t convertToOutputChannelValue(ChannelType type, double value) {
  const auto& info = outputChannelInfo(type);
  if ((value < info.min) || (value > info.max)) {
    DS_FAIL_REQUIRE("Output value out of range.", value, type, info.min, info.max);
  }

  return (uint16_t)round(((pow(2, info.size) - 1) * (value - info.min) / (info.max - info.min)));
}

double convertValueToOtherChannelType(ChannelType inType, double inValue, ChannelType outType) {
  if (inType == outType) {
    return inValue;
  }

  const auto& inInfo = outputChannelInfo(inType);
  const auto& outInfo = outputChannelInfo(outType);
  double factor = (outInfo.max - outInfo.min) / (inInfo.max - inInfo.min);

  return ((inValue - inInfo.min) * factor) + outInfo.min;
}

double limitOutputChannelValue(ChannelType type, double value) {
  const auto& info = outputChannelInfo(type);
  if (value < info.min) {
    return info.min;
  } else if (value > info.max) {
    return info.max;
  }
  return value;
}

double convertFromOutputChannelValue(ChannelType type, uint16_t value) {
  const auto& info = outputChannelInfo(type);

  return (((info.max - info.min) * value) /
          (double)(pow(2, info.size) - 1)) +
          info.min;
}

bool saveValidatedXML(const std::string& _fileName, const std::string& _targetFile) {
  std::shared_ptr<XMLFileValidator> v = std::make_shared<XMLFileValidator>();
  int ret = v->validateFile(_fileName);
  if (ret) {
    // move it to the desired location
    ret = rename(_fileName.c_str(), _targetFile.c_str());
    if (ret != 0) {
      Logger::getInstance()->log(
          "Copying to final destination (" + _targetFile + ") failed: " + std::string(strerror(errno)), lsFatal);
    }
  } else {
    Logger::getInstance()->log(
        "XML not saved! Generated file '" + _fileName + "' is invalid, will not overwrite" + _targetFile, lsFatal);
    if (DSS::getInstance()->getPropertySystem().getBoolValue("/config/debug/storeInvalidXML")) {
      std::string invalidDir("/var/log/invalid_xml/");
      (void)mkdir(invalidDir.c_str(), 0777); // can fail if the directory already exists
      (void)rename(_fileName.c_str(), (invalidDir + boost::filesystem::path(_fileName).filename().string()).c_str());
    }
  }

  return ret;
}

/**
 * filters a bitarray for non-zero bits, and returns a vector
 * containing the indices of the non-zero bits
 *
 *            0         1         2      +0, +10
 * indices:   0123456789012345678901234  0-9 repeating
 * bitmap:    0001110011100000000100100
 * -> result: [ 3, 4, 5, 8, 9, 10, 19, 22 ]
 */
std::vector<int> parseBitfield(const uint8_t* _bitfield, int _bits) {
  std::vector<int> result;
  for (int i = 0; i < _bits; ++i) {
    if ((_bitfield[i / 8] & (1 << (i % 8))) != 0) {
      result.push_back(i);
    }
  }
  return result;
}

std::string propertyValue2String(const vdcapi::PropertyValue& pvalue) {
  std::string value;
  if (pvalue.has_v_bool()) {
    value = pvalue.v_bool() ? "active" : "inactive";
  }
  if (pvalue.has_v_double()) {
    value = doubleToString(pvalue.v_double());
  }
  if (pvalue.has_v_int64()) {
    value = intToString(pvalue.v_int64());
  }
  if (pvalue.has_v_uint64()) {
    value = uintToString(pvalue.v_uint64());
  }
  if (pvalue.has_v_string()) {
    value = pvalue.v_string();
  }
  return value;
}

double convertDimTimeFromDevice(uint8_t value) {
  uint8_t exp = (value >> 4) & 0x0F;
  uint8_t lin = value & 0x0F;
  return roundDigits(0.1 * pow(2, exp) * (1.0 - ((15.0 - lin) / 32.0)), 3);
}

uint8_t convertDimTimeToDevice(double seconds) {
  uint8_t exp = 0;
  for (; exp < 16; exp++) {
    if ((0.1 * pow(2, exp)) >= seconds) {
      break;
    }
  }
  auto lin = static_cast<uint8_t>(round(15.0 - 32.0 * (1.0 - seconds / (0.1 * pow(2, exp)))));
  return (exp << 4 | lin);
}

} // namespace
