/*
    Copyright (c) 2019 digitalSTROM AG, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "devicerequesthandler.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "cache/apartment-proxy-mock.h"
#include "model/device-mock.h"
#include "web/restful.h"


using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;
using namespace std::literals;

static ds::LogChannel dsLogChannel("webHandlerDeviceRequestHandlerTest");
static const char* TAGS = "[web][handler][devicerequesthandler]";

namespace dss {

struct DeviceConfigReply {
  std::string action;
  ds::JsonFragment device;
  ds::JsonFragment update;
};

} // namespace dss

DS_JSON_DECLARE_STRUCT((dss), DeviceConfigReply, (action)(device)(update));
DS_JSON_DEFINE_STRUCT((dss), DeviceConfigReply);

namespace {

TEST_CASE("dssWebHandlerDeviceRequestHandler", TAGS) {
  auto busInterface = BusInterfaceMock{};

  auto proxyPtr = std::make_unique<ApartmentProxyMock>();
  auto& proxy = *proxyPtr;

  auto pApartment = makeApartment(nullptr, nullptr, &busInterface);
  auto& apartment = *pApartment;
  apartment.setProxy(std::move(proxyPtr));

  auto deviceRequestHandler = DeviceRequestHandler(apartment);

  auto& devGRKL300 = *createDevice(
      apartment, DeviceMockConf{.hardwareType = HardwareType::GR_KL300, .buttonInputMode = ButtonInputMode::STANDARD});
  auto& devUMR200 = *createDevice(apartment, DeviceMockConf{.hardwareType = HardwareType::SW_UMR200});

  std::stringstream ss;

  SECTION("setConfig") {
    REQUIRE_CALL(busInterface.device, setDeviceConfig(_, _, _, _));
    ss << deviceRequestHandler.jsonHandleRequest(
        RestfulRequest{"foo/setConfig",
                       ds::str("dsuid=", devGRKL300.getDSID(), "&class=1", "&index=90", "&value=0xff")},
        nullptr, nullptr);
    CHECK(ss.str() == R"({"ok":true})");
  }

  SECTION("setButtonInputMode") {
    REQUIRE_CALL(busInterface.device,
                 setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2)));
    REQUIRE_CALL(busInterface.device,
                 setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1)));
    ss << deviceRequestHandler.jsonHandleRequest(
        RestfulRequest{"foo/setButtonInputMode", ds::str("dsuid=", devGRKL300.getDSID(), "&mode=", "2way_up")}, nullptr,
        nullptr);
    auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
    REQUIRE(reply.ok);

    auto deviceConfigReply = ds::fromJson<DeviceConfigReply>(reply.result);
    CHECK(deviceConfigReply.action == "remove"s); // ?? broke existing pairing ??

    // CHECK(deviceConfigReply.device == toJSON(devGRKL300));

    auto bar = ds::fromJson<std::map<std::string, ds::JsonFragment>>(deviceConfigReply.update);
    CHECK(ds::fromJson<std::string>(bar["dSUID"]) == ds::str(devGRKL300.getDSID()));
    CHECK(ds::fromJson<uint8_t>(bar["buttonInputMode"]) ==
          static_cast<uint8_t>(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2));
  }

  SECTION("setAKMInputProperty") {
    GIVEN("AKM device") {
      THEN("valid akm modes are mapped correctly") {
        struct Y {
          std::string modeArg;
          ButtonInputMode buttonInputMode;
        };

        const auto& y =
            GENERATE(values(std::initializer_list<Y>{{"standard"s, ButtonInputMode::AKM_STANDARD},
                                                     {"inverted", ButtonInputMode::AKM_INVERTED},
                                                     {"on_rising_edge", ButtonInputMode::AKM_ON_RISING_EDGE},
                                                     {"on_falling_edge", ButtonInputMode::AKM_ON_FALLING_EDGE},
                                                     {"off_rising_edge", ButtonInputMode::AKM_OFF_RISING_EDGE},
                                                     {"off_falling_edge", ButtonInputMode::AKM_OFF_FALLING_EDGE},
                                                     {"rising_edge", ButtonInputMode::AKM_RISING_EDGE},
                                                     {"falling_edge", ButtonInputMode::AKM_FALLING_EDGE}}));

        REQUIRE_CALL(busInterface.device, setDeviceConfig(_, _, _, static_cast<uint8_t>(y.buttonInputMode)));
        ss << deviceRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setAKMInputProperty", ds::str("dsuid=", devUMR200.getDSID(), "&mode=", y.modeArg)},
            nullptr, nullptr);

        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(reply.ok);
      }
      THEN("invalid akm modes are rejected") {
        ss << deviceRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setAKMInputProperty", ds::str("dsuid=", devUMR200.getDSID(), "&mode=", "invalid-mode")},
            nullptr, nullptr);

        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(!reply.ok);
        REQUIRE(reply.message == ds::str("Unsupported mode: 'invalid-mode'"));
      }
    }

    GIVEN("non-AKM device") {
      auto& device = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL200});

      THEN("setAKMInputProperty fails") {
        ss << deviceRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setAKMInputProperty", ds::str("dsuid=", device.getDSID(), "&mode=", "standard")},
            nullptr, nullptr);

        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(!reply.ok);
        REQUIRE(reply.message == "This device does not support AKM properties"s);
      }
    }

    GIVEN("paired devices") {
      ALLOW_CALL(busInterface.device,
                 setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2)));
      ALLOW_CALL(busInterface.device,
                 setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1)));
      auto& partnerDevice = *devUMR200.getPartnerDevice();
      devUMR200.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_UP_WITH_INPUT2);
      partnerDevice.setDeviceButtonInputMode(ButtonInputMode::TWO_WAY_DW_WITH_INPUT1);

      THEN("pairing is broken up first") {
        REQUIRE(partnerDevice.is2WaySlave());
        REQUIRE_CALL(busInterface.device,
                     setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::STANDARD)));
        REQUIRE_CALL(busInterface.device,
                     setDeviceConfig(_, _, _, static_cast<uint8_t>(ButtonInputMode::AKM_RISING_EDGE)));
        ss << deviceRequestHandler.jsonHandleRequest(
            RestfulRequest{"foo/setAKMInputProperty", ds::str("dsuid=", devUMR200.getDSID(), "&mode=", "rising_edge")},
            nullptr, nullptr);

        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        CHECK(reply.ok);

        auto deviceConfigReply = ds::fromJson<DeviceConfigReply>(reply.result);
        CHECK(deviceConfigReply.action == "add"s); // ?? broke existing pairing ??

        auto visibleDevice = ds::fromJson<std::map<std::string, ds::JsonFragment>>(deviceConfigReply.device);
        CHECK(ds::fromJson<std::string>(visibleDevice["dSUID"]) == ds::str(partnerDevice.getDSID()));
        CHECK(ds::fromJson<uint8_t>(visibleDevice["buttonInputMode"]) ==
              static_cast<uint8_t>(ButtonInputMode::STANDARD));
      }
    }
  }

  SECTION("area membership") {
    auto addToAreaRequest = [&devGRKL300](int area) {
      return RestfulRequest{"foo/addToArea", ds::str("dsuid=", devGRKL300.getDSID(), "&areaScene=", area)};
    };
    auto removeFromAreaRequest = [&devGRKL300](int area) {
      return RestfulRequest{"foo/removeFromArea", ds::str("dsuid=", devGRKL300.getDSID(), "&areaScene=", area)};
    };
    SECTION("area index is given as 1,2,3,4") {
      const auto& x = GENERATE(1, 2, 3, 4);
      REQUIRE_CALL(busInterface.action, saveScene(_, CallOrigin::system, x - SceneOffA1 + SceneA11, _));
      WHEN("addToArea") {
        REQUIRE_CALL(busInterface.action,
                     callScene(_, CallOrigin::system, SceneAccessCategory::manual, SceneMax, ""s, true));
        ss << deviceRequestHandler.jsonHandleRequest(addToAreaRequest(x), nullptr, nullptr);
        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(reply.ok);
      }
      WHEN("removeFromArea") {
        REQUIRE_CALL(busInterface.action,
                     callScene(_, CallOrigin::system, SceneAccessCategory::manual, SceneMin, ""s, true));
        ss << deviceRequestHandler.jsonHandleRequest(removeFromAreaRequest(x), nullptr, nullptr);
        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(reply.ok);
      }
    }
    SECTION("area index is given as SceneA11,SceneA21") {
      const auto& y = GENERATE(values(std::initializer_list<int>{SceneA11, SceneA21, SceneA31, SceneA41}));
      REQUIRE_CALL(busInterface.action, saveScene(_, CallOrigin::system, y, _));
      WHEN("addToArea") {
        REQUIRE_CALL(busInterface.action,
                     callScene(_, CallOrigin::system, SceneAccessCategory::manual, SceneMax, ""s, true));
        ss << deviceRequestHandler.jsonHandleRequest(addToAreaRequest(y), nullptr, nullptr);
        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(reply.ok);
      }
      WHEN("removeFromArea") {
        REQUIRE_CALL(busInterface.action,
                     callScene(_, CallOrigin::system, SceneAccessCategory::manual, SceneMin, ""s, true));
        ss << deviceRequestHandler.jsonHandleRequest(removeFromAreaRequest(y), nullptr, nullptr);
        auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
        REQUIRE(reply.ok);
      }
    }
    SECTION("invalid area numbers") {
      const auto& x = GENERATE(0, 5, 0xa, SceneA41 + 1);
      CHECK_THROWS(deviceRequestHandler.jsonHandleRequest(addToAreaRequest(x), nullptr, nullptr));
      CHECK_THROWS(deviceRequestHandler.jsonHandleRequest(removeFromAreaRequest(x), nullptr, nullptr));
    }
  }

  SECTION("setOutputChannelDontCareFlag") {
    REQUIRE_CALL(proxy, updateDeviceSceneChannelProperty(_, _));
    ss << deviceRequestHandler.jsonHandleRequest(
        RestfulRequest{"foo/setOutputChannelDontCareFlag",
                       ds::str("dsuid=", devGRKL300.getDSID(), "&channels=shadeOpeningAngleOutside", "&dontCare=1",
                               "&sceneNumber=11")},
        nullptr, nullptr);
    DS_NOTICE(ss.str());
    auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
    REQUIRE(reply.ok);
  }
  SECTION("setOutputChannelSceneValue") {
    REQUIRE_CALL(proxy, updateSceneConfig(_, 15));
    ss << deviceRequestHandler.jsonHandleRequest(
        RestfulRequest{
            "foo/setOutputChannelSceneValue",
            ds::str("dsuid=", devGRKL300.getDSID(), "&channelvalues=shadeOpeningAngleOutside=15", "&sceneNumber=11")},
        nullptr, nullptr);
    DS_NOTICE(ss.str());
    auto reply = ds::fromJson<dss::web::JSONResponse>(ss.str());
    REQUIRE(reply.ok);
  }
}

} // namespace
